/**
 * VueLayers
 * Web map Vue components with the power of OpenLayers
 *
 * @package vuelayers
 * @author Vladimir Vershinin <ghettovoice@gmail.com>
 * @version 0.12.0-rc.26
 * @license MIT
 * @copyright (c) 2017-2021, Vladimir Vershinin <ghettovoice@gmail.com>
 */
import _defineProperty from '@babel/runtime/helpers/esm/defineProperty';
import _typeof from '@babel/runtime/helpers/esm/typeof';
import { cleanSourceParams, EPSG_3857, transformPoint, transformLine, transformPolygon, transformMultiPoint, transformMultiLine, transformMultiPolygon, transformExtent as transformExtent$1, writeGeoJsonGeometry, readGeoJsonGeometry, writeGeoJsonFeature, readGeoJsonFeature, COORD_PRECISION, roundExtent, initializeLayer, getLayerId, setLayerId, initializeControl, getControlId, isGeoJSONGeometry, findPointOnSurface, dumpStyle, initializeFeature, getFeatureId, setFeatureId, getFeatureProperties, setFeatureProperties, CIRCLE_SERIALIZE_PROP, STYLE_SERIALIZE_PROP, roundPointCoords, initializeGeometry, getGeometryId, setGeometryId, getSourceId, initializeSource, setSourceId, initializeStyle, getStyleId, setStyleId, initializeInteraction, getInteractionId, setInteractionId, getInteractionPriority, setInteractionPriority, initializeOverlay, getOverlayId, dumpStrokeStyle, dumpFillStyle, roundCoords, transforms, isEqualCoord, expandUrl, isGeoJSONFeature, getGeoJsonFmt, transform, extentFromProjection } from './ol-ext';
import { hasProp, stubObject, identity, keys, IdentityMap, noop, newLogger, mergeDescriptors, kebabCase, coalesce, assert, makeWatchers, isFunction, isEqual, isArray, addPrefix, lowerFirst, upperFirst, isObjectLike, clonePlainObject, reduce, extractChildren, waitFor, stubTrue, isNumber, isPlainObject, instanceOf, forEach, find, isString, isEmpty, get as get$1, or, negate, sealFactory, replaceTokens, pick, and, stubArray, every, map as map$1, constant, difference } from './utils';
import _classCallCheck from '@babel/runtime/helpers/esm/classCallCheck';
import _assertThisInitialized from '@babel/runtime/helpers/esm/assertThisInitialized';
import _inherits from '@babel/runtime/helpers/esm/inherits';
import _possibleConstructorReturn from '@babel/runtime/helpers/esm/possibleConstructorReturn';
import _getPrototypeOf from '@babel/runtime/helpers/esm/getPrototypeOf';
import _wrapNativeSuper from '@babel/runtime/helpers/esm/wrapNativeSuper';
import _toConsumableArray from '@babel/runtime/helpers/esm/toConsumableArray';
import _asyncToGenerator from '@babel/runtime/helpers/esm/asyncToGenerator';
import _regeneratorRuntime from '@babel/runtime/regenerator';
import debounce from 'debounce-promise';
import { Observable, Object as Object$1, Collection, Feature, Overlay, getUid as getUid$1 } from 'ol';
import EventType from 'ol/events/EventType';
import ObjectEventType from 'ol/ObjectEventType';
import { merge, race, of, throwError, interval } from 'rxjs';
import { retry, filter, publishReplay, refCount, first, mergeMap, delayWhen, map, mapTo, tap, distinctUntilChanged } from 'rxjs/operators';
import { v4 } from 'uuid';
import VueQuery from 'vuequery';
import { fromVueEvent, fromOlChangeEvent, fromOlEvent, fromVueWatcher, bufferDebounceTime } from './rx-ext';
import Vue from 'vue';
import { get } from 'ol/proj';
import _slicedToArray from '@babel/runtime/helpers/esm/slicedToArray';
import CollectionEventType from 'ol/CollectionEventType';
import { defaults, Control } from 'ol/control';
import { getUid } from 'ol/util';
import { Geometry } from 'ol/geom';
import { Style, Fill, Image, Stroke, Text } from 'ol/style';
import RenderEventType from 'ol/render/EventType';
import { Source, Vector, XYZ } from 'ol/source';
import SourceState from 'ol/source/State';
import { defaults as defaults$1, Interaction } from 'ol/interaction';
import BaseLayer from 'ol/layer/Base';
import { boundingExtent as boundingExtent$1 } from 'ol/extent';
import GeometryLayout from 'ol/geom/GeometryLayout';
import GeometryType from 'ol/geom/GeometryType';
import TileEventType from 'ol/source/TileEventType';
import { all } from 'ol/loadingstrategy';
import WMSServerType from 'ol/source/WMSServerType';
import { createTileUrlFunctionFromTemplates } from 'ol-tilecache';
import { createXYZ } from 'ol/tilegrid';

var EVENT_BUS_PROP = '$vlEventBus';
var eventBus = {
  beforeCreate: function beforeCreate() {
    var _this = this;

    initEventBus(); // define local getter

    Object.defineProperties(this, {
      /**
       * @type {Vue}
       */
      $eventBus: {
        enumerable: true,
        get: function get() {
          return _this[EVENT_BUS_PROP];
        }
      }
    });
  }
};

function initEventBus() {
  if (hasProp(Vue, EVENT_BUS_PROP) || hasProp(Vue.prototype, EVENT_BUS_PROP)) return;
  var bus = new Vue();

  if (!hasProp(Vue, EVENT_BUS_PROP)) {
    Object.defineProperties(Vue, _defineProperty({}, EVENT_BUS_PROP, {
      enumerable: true,
      get: function get() {
        return bus;
      }
    }));
  }

  if (!hasProp(Vue.prototype, EVENT_BUS_PROP)) {
    Object.defineProperties(Vue.prototype, _defineProperty({}, EVENT_BUS_PROP, {
      enumerable: true,
      get: function get() {
        return bus;
      }
    }));
  }
}

var IDENTITY_MAP_PROP = '$vlIdentityMap';
var INSTANCES_POOL = 'instances';
var identMap = {
  data: function data() {
    return {
      idents: stubObject()
    };
  },
  beforeCreate: function beforeCreate() {
    var _this = this;

    initIdentityMap(); // define local getter

    Object.defineProperties(this, {
      $identityMap: {
        enumerable: true,
        get: function get() {
          return _this[IDENTITY_MAP_PROP];
        }
      }
    });
  },
  destroyed: function destroyed() {
    this.unsetInstances();
  },
  methods: {
    /**
     * @param parts
     * @return {string}
     */
    makeIdent: function makeIdent() {
      for (var _len = arguments.length, parts = new Array(_len), _key = 0; _key < _len; _key++) {
        parts[_key] = arguments[_key];
      }

      return parts.filter(identity).join('.');
    },

    /**
     * Caches or reuse factory result in the identity map
     * and returns result.
     *
     * @param {string|undefined} ident
     * @param {function} factory
     * @returns {*}
     */
    instanceFactoryCall: function instanceFactoryCall(ident, factory) {
      if (ident && this.hasInstance(ident)) {
        return this.getInstance(ident);
      }

      var inst = factory();

      if (ident) {
        this.setInstance(ident, inst);
      }

      return inst;
    },

    /**
     * @param {string|undefined} ident
     * @returns {*}
     */
    getInstance: function getInstance(ident) {
      if (!ident || !this.hasInstance(ident)) return;
      this.idents[ident] = true;
      return this.$identityMap.get(ident, INSTANCES_POOL);
    },

    /**
     * @param ident
     * @param inst
     */
    setInstance: function setInstance(ident, inst) {
      if (!ident) return;
      this.idents[ident] = true;
      this.$identityMap.set(ident, inst, INSTANCES_POOL);
    },

    /**
     * @param {string|undefined} ident
     * @returns {*}
     */
    hasInstance: function hasInstance(ident) {
      if (!ident) return;
      return this.$identityMap.has(ident, INSTANCES_POOL);
    },

    /**
     * @param {string|undefined} ident
     * @return {void}
     */
    unsetInstance: function unsetInstance(ident) {
      if (!ident) return;
      delete this.idents[ident];
      this.$identityMap.unset(ident, INSTANCES_POOL);
    },

    /**
     * Unsets all self indets
     * @return {void}
     */
    unsetInstances: function unsetInstances() {
      keys(this.idents).forEach(this.unsetInstance.bind(this));
    },

    /**
     * @param {string|undefined} fromIdent
     * @param {string|undefined} toIdent
     * @returns {boolean}
     */
    moveInstance: function moveInstance(fromIdent, toIdent) {
      return this.$identityMap.move(fromIdent, toIdent, INSTANCES_POOL);
    }
  }
};

function initIdentityMap() {
  if (hasProp(Vue, IDENTITY_MAP_PROP) || hasProp(Vue.prototype, IDENTITY_MAP_PROP)) return;
  var imap = new IdentityMap();

  if (!hasProp(Vue, IDENTITY_MAP_PROP)) {
    Object.defineProperties(Vue, _defineProperty({}, IDENTITY_MAP_PROP, {
      enumerable: true,
      get: function get() {
        return imap;
      }
    }));
  }

  if (!hasProp(Vue.prototype, IDENTITY_MAP_PROP)) {
    Object.defineProperties(Vue.prototype, _defineProperty({}, IDENTITY_MAP_PROP, {
      enumerable: true,
      get: function get() {
        return imap;
      }
    }));
  }
}

/**
 * RxJS subscriptions manager.
 */

var rxSubs = {
  beforeCreate: function beforeCreate() {
    /**
     * @type {Subscription[]}
     * @private
     */
    this._rxSubs = [];
  },
  destroyed: function destroyed() {
    this.unsubscribeAll();
  },
  methods: {
    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {},

    /**
     * @return {void}
     * @protected
     */
    unsubscribeAll: function unsubscribeAll() {
      this._rxSubs.forEach(function (subs) {
        return subs.unsubscribe();
      });

      this._rxSubs = [];
    },

    /**
     * @param {Observable<T>} observable
     * @param {function} [next] Next handler or Observer object.
     * @param {function} [error] Error handler.
     * @param {function} [complete] Complete handler.
     * @return {Subscription}
     * @protected
     */
    subscribeTo: function subscribeTo(observable) {
      var _this = this;

      var next = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : noop;
      var error = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : noop;
      var complete = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : noop;

      var errorWrap = function errorWrap(err) {
        if (process.env.NODE_ENV !== 'production') {
          if (_this.$logger) {
            _this.$logger.error(err.stack);
          } else {
            newLogger(_this.vmName).error(err.stack);
          }
        }

        error(err);
      };

      var subs = observable.pipe(retry(3)).subscribe(next, errorWrap, complete);

      this._rxSubs.push(subs);

      return subs;
    },
    unsubscribe: function unsubscribe(subs) {
      var idx = this._rxSubs.indexOf(subs);

      if (idx === -1) return;
      subs.unsubscribe();

      this._rxSubs.splice(idx, 1);
    }
  }
};

// const SERVICES_PROP = Symbol('services')

var SERVICES_PROP = 'services';
/**
 * Service container mixin
 */

var services = {
  inject: {
    $services: SERVICES_PROP // todo works only in Vue 2.5.x
    // $services: {from: SERVICES_PROP, default: Object.create(null)},

  },
  provide: function provide() {
    return _defineProperty({}, SERVICES_PROP, this.getServices());
  },
  methods: {
    /**
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      return mergeDescriptors({}, this.$services || {});
    }
  },
  beforeCreate: function beforeCreate() {
    var source = this.$parent;

    while (source) {
      if (source._provided != null && source._provided[SERVICES_PROP] != null) {
        break;
      }

      source = source.$parent;
    }

    if (source == null || source._provided[SERVICES_PROP] == null) {
      delete this.$options.inject.$services;
    }
  }
};

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = _getPrototypeOf(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = _getPrototypeOf(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return _possibleConstructorReturn(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {})); return true; } catch (e) { return false; } }

function ownKeys$o(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$o(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$o(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$o(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var VM_PROP = 'vm';
var FRAME_TIME = 1000 / 60;
var OlObjectState = {
  UNDEF: 'undef',
  CREATING: 'creating',
  CREATED: 'created',
  MOUNTING: 'mounting',
  MOUNTED: 'mounted',
  UNMOUNTING: 'unmounting',
  UNMOUNTED: 'unmounted',
  DESTROYING: 'destroying',
  DESTROYED: 'destroyed'
};
var OlObjectEvent = {
  CREATED: OlObjectState.CREATED,
  MOUNTED: OlObjectState.MOUNTED,
  UNMOUNTED: OlObjectState.UNMOUNTED,
  DESTROYED: OlObjectState.DESTROYED,
  ERROR: 'error'
};
var OlObjectAction = {
  CREATE: 'create',
  MOUNT: 'mount',
  UNMOUNT: 'unmount',
  DESTROY: 'destroy'
};
/**
 * Basic ol component mixin.
 * todo try to subscribe to generic change event here and update rev according to internal ol counter
 */

var olCmp = {
  mixins: [identMap, rxSubs, services, eventBus],
  props: {
    /**
     * @type {string|number}
     */
    id: {
      type: [String, Number],
      default: v4,
      validator: function validator(value) {
        return value != null && value !== '';
      }
    },

    /**
     * Unique key for saving to identity map
     * @type {string|number|undefined}
     */
    ident: [String, Number]
  },
  data: function data() {
    return {
      rev: 0,
      currentId: this.id
    };
  },
  computed: {
    /**
     * @type {string}
     */
    vmClass: function vmClass() {
      return kebabCase(this.$options.name);
    },

    /**
     * @type {string}
     */
    vmId: function vmId() {
      return [this.vmClass, this.currentId].filter(identity).join('.');
    },

    /**
     * @type {string}
     */
    vmName: function vmName() {
      return [this.$options.name, this.currentId].filter(identity).join('.');
    },

    /**
     * @type {string|undefined}
     */
    olObjIdent: function olObjIdent() {
      if (!this.ident) return;
      return this.makeIdent(this.ident);
    }
  },
  watch: _objectSpread$o({
    rev: function rev() {
      if (!this.$olObject) return;

      if (this.currentId !== this.getIdInternal()) {
        this.currentId = this.getIdInternal();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['id', 'currentId', 'olObjIdent'])),
  created: function created() {
    var _this = this;

    return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              /**
               * @type {{warn: (function(...[*]): void), log: (function(...[*]): void), error: (function(...[*]): void)}}
               * @private
               */
              _this._logger = newLogger(_this.vmName);
              /**
               * @type {number}
               * @private
               */

              _this._olObjectState = OlObjectState.UNDEF;
              /**
               * @type {module:ol/Object~BaseObject}
               * @private
               */

              _this._olObject = undefined;
              /**
               * @type {Observable}
               * @private
               */

              _this._olObjectEvents = newOlObjectEventsObs.call(_this);
              defineDebounceMethods.call(_this);
              defineServices$i.call(_this);
              _context.next = 8;
              return execInit.call(_this, false);

            case 8:
            case "end":
              return _context.stop();
          }
        }
      }, _callee);
    }))();
  },
  mounted: function mounted() {
    var _this2 = this;

    return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
      return _regeneratorRuntime.wrap(function _callee2$(_context2) {
        while (1) {
          switch (_context2.prev = _context2.next) {
            case 0:
              _context2.next = 2;
              return execMount.call(_this2);

            case 2:
            case "end":
              return _context2.stop();
          }
        }
      }, _callee2);
    }))();
  },
  beforeDestroy: function beforeDestroy() {
    var _this3 = this;

    return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
      return _regeneratorRuntime.wrap(function _callee3$(_context3) {
        while (1) {
          switch (_context3.prev = _context3.next) {
            case 0:
              _context3.next = 2;
              return execUnmount.call(_this3, true);

            case 2:
            case "end":
              return _context3.stop();
          }
        }
      }, _callee3);
    }))();
  },
  destroyed: function destroyed() {
    var _this4 = this;

    return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
      return _regeneratorRuntime.wrap(function _callee4$(_context4) {
        while (1) {
          switch (_context4.prev = _context4.next) {
            case 0:
              _context4.next = 2;
              return execDeinit.call(_this4);

            case 2:
            case "end":
              return _context4.stop();
          }
        }
      }, _callee4);
    }))();
  },
  methods: {
    /**
     * @return {Promise<void>} Resolves when initialization completes
     * @protected
     */
    beforeInit: function beforeInit() {},

    /**
     * @return {Promise<void>} Resolves when initialization completes
     * @protected
     */
    init: function init() {
      var _this5 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return _this5.instanceFactoryCall(_this5.olObjIdent, _this5.createOlObject.bind(_this5));

              case 2:
                _this5._olObject = _context5.sent;
                _this5._olObject[VM_PROP] || (_this5._olObject[VM_PROP] = []); // for loaded from IdentityMap

                if (!_this5._olObject[VM_PROP].includes(_this5)) {
                  _this5._olObject[VM_PROP].push(_this5);
                }

                _context5.next = 7;
                return _this5.changed();

              case 7:
                _this5.subscribeAll();

              case 8:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5);
      }))();
    },

    /**
     * @return {module:ol/Object~BaseObject|Promise<module:ol/Object~BaseObject>}
     * @protected
     * @abstract
     */
    createOlObject: function createOlObject() {
      throw new Error("".concat(this.vmName, " not implemented method: createOlObject()"));
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    deinit: function deinit() {
      var _this6 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee6() {
        return _regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _this6.unsubscribeAll();

                _this6.unsetInstances();

                if (_this6._olObject) {
                  _this6._olObject[VM_PROP] = _this6._olObject[VM_PROP].filter(function (vm) {
                    return vm !== _this6;
                  });
                  _this6._olObject = undefined;
                }

              case 3:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeMount: function beforeMount() {},

    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {},

    /**
     * @return {void|Promise<void>}
     * @protected
     */
    unmount: function unmount() {},

    /**
     * Refresh internal ol objects
     * @return {Promise<void>}
     */
    refresh: function refresh() {
      var _this7 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee7() {
        var olObj;
        return _regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.prev = 0;
                _context7.next = 3;
                return _this7.resolveOlObject();

              case 3:
                olObj = _context7.sent;

                if (!(olObj instanceof Observable)) {
                  _context7.next = 6;
                  break;
                }

                return _context7.abrupt("return", new Promise(function (resolve) {
                  olObj.once('change', function () {
                    return resolve();
                  });
                  olObj.changed();
                }));

              case 6:
                _context7.next = 10;
                break;

              case 8:
                _context7.prev = 8;
                _context7.t0 = _context7["catch"](0);

              case 10:
                return _context7.abrupt("return", _this7.changed());

              case 11:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, null, [[0, 8]]);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    scheduleRefresh: function scheduleRefresh() {
      var _this8 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee8() {
        return _regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                _context8.prev = 0;
                _context8.next = 3;
                return _this8.debounceRefresh();

              case 3:
                _context8.next = 7;
                break;

              case 5:
                _context8.prev = 5;
                _context8.t0 = _context8["catch"](0);

              case 7:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, null, [[0, 5]]);
      }))();
    },

    /**
     * Internal usage only in components that doesn't support refreshing.
     * @return {Promise<void>}
     * @protected
     */
    remount: function remount() {
      var _this9 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee9() {
        return _regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                if (!(_this9.$olObjectState === OlObjectState.MOUNTED)) {
                  _context9.next = 8;
                  break;
                }

                if (process.env.VUELAYERS_DEBUG) {
                  _this9.$logger.log('remounting...');
                }

                _context9.next = 4;
                return execUnmount.call(_this9);

              case 4:
                _context9.next = 6;
                return execMount.call(_this9);

              case 6:
                _context9.next = 9;
                break;

              case 8:
                if (process.env.VUELAYERS_DEBUG) {
                  _this9.$logger.log('remount discarded');
                }

              case 9:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    scheduleRemount: function scheduleRemount() {
      var _this10 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee10() {
        return _regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                if (![OlObjectState.MOUNTING, OlObjectState.MOUNTED].includes(_this10.$olObjectState)) {
                  _context10.next = 6;
                  break;
                }

                if (process.env.VUELAYERS_DEBUG) {
                  _this10.$logger.log('remount scheduled');
                }

                _context10.next = 4;
                return _this10.debounceRemount();

              case 4:
                _context10.next = 7;
                break;

              case 6:
                if (process.env.VUELAYERS_DEBUG) {
                  _this10.$logger.log('remount discarded');
                }

              case 7:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10);
      }))();
    },

    /**
     * Only for internal purpose to support watching for properties
     * for which OpenLayers doesn't provide setters.
     * @return {Promise}
     * @protected
     */
    recreate: function recreate() {
      var _this11 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee11() {
        var mounted;
        return _regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                if (![OlObjectState.CREATED, OlObjectState.MOUNTING, OlObjectState.MOUNTED, OlObjectState.UNMOUNTING, OlObjectState.UNMOUNTED].includes(_this11.$olObjectState)) {
                  _context11.next = 15;
                  break;
                }

                if (process.env.VUELAYERS_DEBUG) {
                  _this11.$logger.log('recreating...');
                }

                mounted = [OlObjectState.MOUNTING, OlObjectState.MOUNTED].includes(_this11.$olObjectState);

                if (!mounted) {
                  _context11.next = 6;
                  break;
                }

                _context11.next = 6;
                return execUnmount.call(_this11);

              case 6:
                _context11.next = 8;
                return execDeinit.call(_this11);

              case 8:
                _context11.next = 10;
                return execInit.call(_this11);

              case 10:
                if (!mounted) {
                  _context11.next = 13;
                  break;
                }

                _context11.next = 13;
                return execMount.call(_this11);

              case 13:
                _context11.next = 16;
                break;

              case 15:
                if (process.env.VUELAYERS_DEBUG) {
                  _this11.$logger.log('recreate discarded');
                }

              case 16:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    scheduleRecreate: function scheduleRecreate() {
      var _this12 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee12() {
        return _regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                if (![OlObjectState.CREATING, OlObjectState.CREATED, OlObjectState.MOUNTING, OlObjectState.MOUNTED, OlObjectState.UNMOUNTING, OlObjectState.UNMOUNTED].includes(_this12.$olObjectState)) {
                  _context12.next = 6;
                  break;
                }

                if (process.env.VUELAYERS_DEBUG) {
                  _this12.$logger.log('recreate scheduled');
                }

                _context12.next = 4;
                return _this12.debounceRecreate();

              case 4:
                _context12.next = 7;
                break;

              case 6:
                if (process.env.VUELAYERS_DEBUG) {
                  _this12.$logger.log('recreate discarded');
                }

              case 7:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12);
      }))();
    },

    /**
     * Redefine for easy call in child components
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      return services.methods.getServices.call(this);
    },

    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      subscribeToOlObjectEvents.call(this);
    },

    /**
     * @returns {Promise<Object>}
     * @throws {Error} If underlying OpenLayers object not initialized (incorrect initialization, already destroy).
     */
    resolveOlObject: function resolveOlObject() {
      var _this13 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee13() {
        return _regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                _context13.next = 2;
                return _this13.$createPromise;

              case 2:
                return _context13.abrupt("return", _this13.$olObject || function (e) {
                  throw e;
                }(new OLObjectNotInitializedError("".concat(_this13.vmName, " OpenLayers object is undefined"))));

              case 3:
              case "end":
                return _context13.stop();
            }
          }
        }, _callee13);
      }))();
    },

    /**
     * @return {Promise}
     */
    $nextTickPromise: function $nextTickPromise() {
      var _this14 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee14() {
        return _regeneratorRuntime.wrap(function _callee14$(_context14) {
          while (1) {
            switch (_context14.prev = _context14.next) {
              case 0:
                return _context14.abrupt("return", new Promise(_this14.$nextTick.bind(_this14)));

              case 1:
              case "end":
                return _context14.stop();
            }
          }
        }, _callee14);
      }))();
    },

    /**
     * @return {string|number}
     */
    getId: function getId() {
      return coalesce(this.$olObject && this.getIdInternal(), this.currentId);
    },

    /**
     * @return {*}
     * @protected
     */
    getIdInternal: function getIdInternal() {
      if (this.$olObject instanceof Object$1) {
        return this.$olObject.get('id');
      }

      return this.$olObject.id;
    },

    /**
     * @param {string|number} id
     */
    setId: function setId(id) {
      assert(id != null && id !== '', 'Invalid id');

      if (this.currentId !== id) {
        this.currentId = id;
      }

      this.$olObject && this.setIdInternal(id);
    },

    /**
     * @param {*} id
     * @protected
     */
    setIdInternal: function setIdInternal(id) {
      if (this.$olObject instanceof Object$1) {
        if (this.$olObject.get('id') !== id) {
          this.$olObject.set('id', id);
        }

        return;
      }

      if (this.$olObject.id !== id) {
        this.$olObject.id = id;
      }
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    changed: function changed() {
      var _this15 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee15() {
        return _regeneratorRuntime.wrap(function _callee15$(_context15) {
          while (1) {
            switch (_context15.prev = _context15.next) {
              case 0:
                _this15.rev++;
                _context15.next = 3;
                return _this15.$nextTickPromise();

              case 3:
              case "end":
                return _context15.stop();
            }
          }
        }, _callee15);
      }))();
    },

    /**
     * @param {string|number} value
     * @protected
     */
    idChanged: function idChanged(value) {
      this.setId(value);
    },

    /**
     * @param {string|number} value
     * @protected
     */
    currentIdChanged: function currentIdChanged(value) {
      if (value === this.id) return;
      this.$emit('update:id', value);
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    olObjIdentChanged: function olObjIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.$olObject) {
        this.setInstance(value, this.$olObject);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    },

    /**
     * @protected
     */
    syncNonObservable: function syncNonObservable() {
      this.setId(this.getId());
    }
  }
};

function defineDebounceMethods() {
  this.debounceRefresh = debounce(this.refresh.bind(this), FRAME_TIME);
  this.debounceRemount = debounce(this.remount.bind(this), FRAME_TIME);
  this.debounceRecreate = debounce(this.recreate.bind(this), FRAME_TIME);
  this.debounceChanged = debounce(this.changed.bind(this), FRAME_TIME);
}

function defineServices$i() {
  var _this16 = this;

  Object.defineProperties(this, {
    $vq: {
      enumerable: true,
      get: function get() {
        return VueQuery(_this16);
      }
    },

    /**
     * @type {{warn: (function(...[*]): void), log: (function(...[*]): void), error: (function(...[*]): void)}}
     */
    $logger: {
      enumerable: true,
      get: function get() {
        return _this16._logger;
      }
    },

    /**
     * @type {string}
     */
    $olObjectState: {
      enumerable: true,
      get: function get() {
        return _this16._olObjectState;
      }
    },

    /**
     * @type {module:ol/Object~BaseObject|undefined}
     */
    $olObject: {
      enumerable: true,
      get: function get() {
        return _this16._olObject;
      }
    },
    $olObjectEvents: {
      enumerable: true,
      get: function get() {
        return _this16._olObjectEvents;
      }
    },

    /**
     * @type {Promise<void>}
     */
    $createPromise: {
      enumerable: true,
      get: function get() {
        return newLifecyclePromise.call(_this16, OlObjectAction.CREATE, OlObjectEvent.CREATED, _defineProperty({}, OlObjectState.CREATING, [OlObjectAction.CREATE, OlObjectEvent.CREATED]));
      }
    },

    /**
     * @type {Promise<void>}
     */
    $mountPromise: {
      enumerable: true,
      get: function get() {
        return _this16.$createPromise.then(function () {
          var _newLifecyclePromise$2;

          return newLifecyclePromise.call(_this16, OlObjectAction.MOUNT, OlObjectEvent.MOUNTED, (_newLifecyclePromise$2 = {}, _defineProperty(_newLifecyclePromise$2, OlObjectState.CREATING, [OlObjectAction.CREATE, OlObjectEvent.CREATED]), _defineProperty(_newLifecyclePromise$2, OlObjectState.MOUNTING, [OlObjectAction.MOUNT, OlObjectEvent.MOUNTED]), _newLifecyclePromise$2));
        });
      }
    },

    /**
     * @type {Promise<void>}
     */
    $unmountPromise: {
      enumerable: true,
      get: function get() {
        return _this16.$mountPromise.then(function () {
          var _newLifecyclePromise$3;

          return newLifecyclePromise.call(_this16, OlObjectAction.UNMOUNT, OlObjectEvent.UNMOUNTED, (_newLifecyclePromise$3 = {}, _defineProperty(_newLifecyclePromise$3, OlObjectState.CREATING, [OlObjectAction.CREATE, OlObjectEvent.CREATED]), _defineProperty(_newLifecyclePromise$3, OlObjectState.MOUNTING, [OlObjectAction.MOUNT, OlObjectEvent.MOUNTED]), _defineProperty(_newLifecyclePromise$3, OlObjectState.UNMOUNTING, [OlObjectAction.UNMOUNT, OlObjectEvent.UNMOUNTED]), _newLifecyclePromise$3));
        });
      }
    },

    /**
     * @type {Promise<void>}
     */
    $destroyPromise: {
      enumerable: true,
      get: function get() {
        return _this16.$unmountPromise.then(function () {
          var _newLifecyclePromise$4;

          return newLifecyclePromise.call(_this16, OlObjectAction.DESTROY, OlObjectEvent.DESTROYED, (_newLifecyclePromise$4 = {}, _defineProperty(_newLifecyclePromise$4, OlObjectState.CREATING, [OlObjectAction.CREATE, OlObjectEvent.CREATED]), _defineProperty(_newLifecyclePromise$4, OlObjectState.MOUNTING, [OlObjectAction.MOUNT, OlObjectEvent.MOUNTED]), _defineProperty(_newLifecyclePromise$4, OlObjectState.UNMOUNTING, [OlObjectAction.UNMOUNT, OlObjectEvent.UNMOUNTED]), _defineProperty(_newLifecyclePromise$4, OlObjectState.DESTROYING, [OlObjectAction.DESTROY, OlObjectEvent.DESTROYED]), _newLifecyclePromise$4));
        });
      }
    }
  });
}

function newOlObjectEventsObs() {
  var _this17 = this;

  return merge.apply(void 0, _toConsumableArray(Object.values(OlObjectEvent).map(function (name) {
    return fromVueEvent(_this17.$eventBus, name, function (args) {
      return {
        name: name,
        args: isArray(args) ? args : [args]
      };
    }).pipe(filter(function (_ref) {
      var args = _ref.args;
      return args[0] === _this17 || args[1] === _this17;
    }));
  }))).pipe(publishReplay(), refCount());
}

function newLifecyclePromise(action, endEvent, startStates) {
  var _this18 = this;

  var newFinishObs = function newFinishObs(action, endEvent) {
    return race(_this18.$olObjectEvents.pipe(filter(function (_ref2) {
      var name = _ref2.name;
      return name === endEvent;
    })), _this18.$olObjectEvents.pipe(filter(function (_ref3) {
      var name = _ref3.name,
          args = _ref3.args;
      return name === OlObjectEvent.ERROR && args[0] instanceof LifecycleError && args[0].action === action;
    })));
  };

  return merge(newFinishObs(action, endEvent), this.$olObjectEvents.pipe(filter(function (_ref4) {
    var name = _ref4.name,
        args = _ref4.args;
    return name === OlObjectEvent.ERROR && args[0] instanceof CanceledError;
  }))).pipe(first(), mergeMap(function (evt) {
    var obs = of(evt); // If at the time of CancelError the ol object stay in some intermediate state
    // we need to delay observable emit until current hook complete

    if (evt.name === OlObjectEvent.ERROR && evt.args[0] instanceof CanceledError && startStates[_this18.$olObjectState] != null) {
      obs = obs.pipe(delayWhen(function () {
        return newFinishObs.apply(void 0, _toConsumableArray(startStates[_this18.$olObjectState]));
      }));
    }

    return obs;
  }), mergeMap(function (evt) {
    if (evt.name === OlObjectEvent.ERROR) {
      return throwError(evt.args[0]);
    }

    return of(evt);
  }), map(function (_ref5) {
    var args = _ref5.args;
    return args.length === 1 ? args[0] : args;
  })).toPromise();
}
/**
 * @param {boolean} [resetEventsObs=true]
 * @returns {Promise<void>}
 * @private
 */


function execInit() {
  return _execInit.apply(this, arguments);
}
/**
 * @return {Promise<void>}
 * @private
 */


function _execInit() {
  _execInit = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee16() {
    var resetEventsObs,
        prevState,
        canceled,
        lcErr,
        _args16 = arguments;
    return _regeneratorRuntime.wrap(function _callee16$(_context16) {
      while (1) {
        switch (_context16.prev = _context16.next) {
          case 0:
            resetEventsObs = _args16.length > 0 && _args16[0] !== undefined ? _args16[0] : true;
            prevState = this._olObjectState;

            if (resetEventsObs) {
              this._olObjectEvents = newOlObjectEventsObs.call(this);
            }

            this._olObjectEvents.subscribe();

            _context16.prev = 4;
            _context16.next = 7;
            return Promise.race([Promise.resolve(this.beforeInit()).then(function () {
              return false;
            }), this.$olObjectEvents.pipe(filter(function (_ref7) {
              var name = _ref7.name,
                  args = _ref7.args;
              return name === OlObjectEvent.ERROR && args[0] instanceof CanceledError;
            }), first()).toPromise().then(function () {
              return true;
            })]);

          case 7:
            canceled = _context16.sent;

            if (!canceled) {
              _context16.next = 11;
              break;
            }

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectAction.CREATE, " canceled"));
            }

            return _context16.abrupt("return");

          case 11:
            this._olObjectState = OlObjectState.CREATING;
            _context16.next = 14;
            return this.init();

          case 14:
            this._olObjectState = OlObjectState.CREATED;

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectEvent.CREATED));
            }

            this.$emit(OlObjectEvent.CREATED, this);
            this.$eventBus.$emit(OlObjectEvent.CREATED, this);
            _context16.next = 27;
            break;

          case 20:
            _context16.prev = 20;
            _context16.t0 = _context16["catch"](4);
            lcErr = new LifecycleError(_context16.t0, this.vmName, OlObjectAction.CREATE);
            this._olObjectState = prevState;
            this.$emit(OlObjectEvent.ERROR, lcErr, this);
            this.$eventBus.$emit(OlObjectEvent.ERROR, lcErr, this);
            throw lcErr;

          case 27:
          case "end":
            return _context16.stop();
        }
      }
    }, _callee16, this, [[4, 20]]);
  }));
  return _execInit.apply(this, arguments);
}

function execMount() {
  return _execMount.apply(this, arguments);
}
/**
 * @param {boolean} [fireCancel=false]
 * @return {void|Promise<void>}
 * @private
 */


function _execMount() {
  _execMount = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee17() {
    var prevState, canceled, lcErr;
    return _regeneratorRuntime.wrap(function _callee17$(_context17) {
      while (1) {
        switch (_context17.prev = _context17.next) {
          case 0:
            prevState = this._olObjectState;
            _context17.prev = 1;
            _context17.next = 4;
            return this.$createPromise;

          case 4:
            _context17.next = 6;
            return Promise.race([Promise.resolve(this.beforeMount()).then(function () {
              return false;
            }), this.$olObjectEvents.pipe(filter(function (_ref8) {
              var name = _ref8.name,
                  args = _ref8.args;
              return name === OlObjectEvent.ERROR && args[0] instanceof CanceledError;
            }), first()).toPromise().then(function () {
              return true;
            })]);

          case 6:
            canceled = _context17.sent;

            if (!canceled) {
              _context17.next = 10;
              break;
            }

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectAction.MOUNT, " canceled"));
            }

            return _context17.abrupt("return");

          case 10:
            this._olObjectState = OlObjectState.MOUNTING;
            _context17.next = 13;
            return this.mount();

          case 13:
            this._olObjectState = OlObjectState.MOUNTED;

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectEvent.MOUNTED));
            }

            this.$emit(OlObjectEvent.MOUNTED, this);
            this.$eventBus.$emit(OlObjectEvent.MOUNTED, this);
            _context17.next = 29;
            break;

          case 19:
            _context17.prev = 19;
            _context17.t0 = _context17["catch"](1);

            if (!(_context17.t0 instanceof CanceledError)) {
              _context17.next = 24;
              break;
            }

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectAction.MOUNT, " canceled"));
            }

            return _context17.abrupt("return");

          case 24:
            lcErr = new LifecycleError(_context17.t0, this.vmName, OlObjectAction.MOUNT);
            this._olObjectState = prevState;
            this.$emit(OlObjectEvent.ERROR, lcErr, this);
            this.$eventBus.$emit(OlObjectEvent.ERROR, lcErr, this);
            throw lcErr;

          case 29:
          case "end":
            return _context17.stop();
        }
      }
    }, _callee17, this, [[1, 19]]);
  }));
  return _execMount.apply(this, arguments);
}

function execUnmount() {
  return _execUnmount.apply(this, arguments);
}
/**
 * @returns {Promise<void>}
 * @private
 */


function _execUnmount() {
  _execUnmount = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee18() {
    var fireCancel,
        prevState,
        err,
        lcErr,
        _args18 = arguments;
    return _regeneratorRuntime.wrap(function _callee18$(_context18) {
      while (1) {
        switch (_context18.prev = _context18.next) {
          case 0:
            fireCancel = _args18.length > 0 && _args18[0] !== undefined ? _args18[0] : false;
            prevState = this._olObjectState;

            if (fireCancel) {
              if ([OlObjectState.UNDEF, OlObjectState.CREATING, OlObjectState.CREATED, OlObjectState.MOUNTING].includes(this.$olObjectState)) {
                err = new CanceledError("".concat(this.vmName, " ol object lifecycle canceled"));
                this.$emit(OlObjectEvent.ERROR, err, this);
                this.$eventBus.$emit(OlObjectEvent.ERROR, err, this);
              }
            }

            _context18.prev = 3;
            _context18.prev = 4;
            _context18.next = 7;
            return this.$mountPromise;

          case 7:
            _context18.next = 18;
            break;

          case 9:
            _context18.prev = 9;
            _context18.t0 = _context18["catch"](4);

            if (!(_context18.t0 instanceof CanceledError)) {
              _context18.next = 17;
              break;
            }

            if (!(this.$olObjectState !== OlObjectState.MOUNTED)) {
              _context18.next = 15;
              break;
            }

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectAction.UNMOUNT, " canceled"));
            }

            return _context18.abrupt("return");

          case 15:
            _context18.next = 18;
            break;

          case 17:
            throw _context18.t0;

          case 18:
            this._olObjectState = OlObjectState.UNMOUNTING;
            _context18.next = 21;
            return this.unmount();

          case 21:
            this._olObjectState = OlObjectState.UNMOUNTED;

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectEvent.UNMOUNTED));
            }

            this.$emit(OlObjectEvent.UNMOUNTED, this);
            this.$eventBus.$emit(OlObjectEvent.UNMOUNTED, this);
            _context18.next = 34;
            break;

          case 27:
            _context18.prev = 27;
            _context18.t1 = _context18["catch"](3);
            lcErr = new LifecycleError(_context18.t1, this.vmName, OlObjectAction.UNMOUNT);
            this._olObjectState = prevState;
            this.$emit(OlObjectEvent.ERROR, this, lcErr);
            this.$eventBus.$emit(OlObjectEvent.ERROR, this, lcErr);
            throw lcErr;

          case 34:
          case "end":
            return _context18.stop();
        }
      }
    }, _callee18, this, [[3, 27], [4, 9]]);
  }));
  return _execUnmount.apply(this, arguments);
}

function execDeinit() {
  return _execDeinit.apply(this, arguments);
}

function _execDeinit() {
  _execDeinit = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee19() {
    var prevState, lcErr;
    return _regeneratorRuntime.wrap(function _callee19$(_context19) {
      while (1) {
        switch (_context19.prev = _context19.next) {
          case 0:
            prevState = this._olObjectState;
            _context19.prev = 1;
            _context19.prev = 2;
            _context19.next = 5;
            return this.$unmountPromise;

          case 5:
            _context19.next = 16;
            break;

          case 7:
            _context19.prev = 7;
            _context19.t0 = _context19["catch"](2);

            if (!(_context19.t0 instanceof CanceledError)) {
              _context19.next = 15;
              break;
            }

            if (!(this.$olObjectState !== OlObjectState.CREATED)) {
              _context19.next = 13;
              break;
            }

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectAction.DESTROY, " canceled"));
            }

            return _context19.abrupt("return");

          case 13:
            _context19.next = 16;
            break;

          case 15:
            throw _context19.t0;

          case 16:
            this._olObjectState = OlObjectState.DESTROYING;
            _context19.next = 19;
            return this.deinit();

          case 19:
            this._olObjectState = OlObjectState.DESTROYED;

            if (process.env.VUELAYERS_DEBUG) {
              this.$logger.log("ol object ".concat(OlObjectEvent.DESTROYED));
            }

            this.$emit(OlObjectEvent.DESTROYED, this);
            this.$eventBus.$emit(OlObjectEvent.DESTROYED, this);
            _context19.next = 32;
            break;

          case 25:
            _context19.prev = 25;
            _context19.t1 = _context19["catch"](1);
            lcErr = new LifecycleError(_context19.t1, this.vmName, OlObjectAction.DESTROY);
            this._olObjectState = prevState;
            this.$emit(OlObjectEvent.ERROR, this, lcErr);
            this.$eventBus.$emit(OlObjectEvent.ERROR, this, lcErr);
            throw lcErr;

          case 32:
          case "end":
            return _context19.stop();
        }
      }
    }, _callee19, this, [[1, 25], [2, 7]]);
  }));
  return _execDeinit.apply(this, arguments);
}

function subscribeToOlObjectEvents() {
  var _this19 = this;

  if (this.$olObject instanceof Observable) {
    var setterKey = addPrefix('set');
    var propChanges = fromOlChangeEvent(this.$olObject, ['id'], true, function (evt) {
      return _objectSpread$o(_objectSpread$o({}, evt), {}, {
        setter: _this19[setterKey(evt.prop)]
      });
    });
    this.subscribeTo(propChanges, function (_ref6) {
      var setter = _ref6.setter,
          value = _ref6.value;
      return setter(value);
    });
    this.subscribeTo(merge(fromOlEvent(this.$olObject, EventType.CHANGE), fromOlEvent(this.$olObject, ObjectEventType.PROPERTYCHANGE)), this.debounceChanged.bind(this));
    this.subscribeTo(fromOlEvent(this.$olObject, EventType.ERROR), function (evt) {
      return _this19.$emit(OlObjectEvent.ERROR, evt);
    });
  } else {
    this.subscribeTo(interval(FRAME_TIME), this.syncNonObservable.bind(this));
  }
}

var LifecycleError = /*#__PURE__*/function (_Error) {
  _inherits(LifecycleError, _Error);

  var _super = _createSuper(LifecycleError);

  function LifecycleError(err, vmName, action) {
    var _this20;

    _classCallCheck(this, LifecycleError);

    if (err instanceof LifecycleError) {
      err = err.err;
    }

    var baseMessage = "".concat(vmName, " ").concat(action, " ol object failed");
    var message = "".concat(baseMessage, ": ").concat(err.message);
    _this20 = _super.call(this, message);

    _defineProperty(_assertThisInitialized(_this20), "name", 'LifecycleError');

    _this20.vmName = vmName;
    _this20.action = action;
    _this20.err = err;

    if (err.stack) {
      _this20.stack = "".concat(baseMessage, ": ").concat(err.stack);
    } else {
      if (isFunction(Error.captureStackTrace)) {
        Error.captureStackTrace(_assertThisInitialized(_this20), _this20.constructor);
      } else {
        _this20.stack = new Error(message).stack;
      }
    }

    return _this20;
  }

  return LifecycleError;
}( /*#__PURE__*/_wrapNativeSuper(Error));
var CanceledError = /*#__PURE__*/function (_Error2) {
  _inherits(CanceledError, _Error2);

  var _super2 = _createSuper(CanceledError);

  function CanceledError(message) {
    var _this21;

    _classCallCheck(this, CanceledError);

    _this21 = _super2.call(this, message);

    _defineProperty(_assertThisInitialized(_this21), "name", 'CanceledError');

    if (isFunction(Error.captureStackTrace)) {
      Error.captureStackTrace(_assertThisInitialized(_this21), _this21.constructor);
    } else {
      _this21.stack = new Error(message).stack;
    }

    return _this21;
  }

  return CanceledError;
}( /*#__PURE__*/_wrapNativeSuper(Error));
function isCreateError(err) {
  return err instanceof LifecycleError && err.action === OlObjectAction.CREATE;
}
function isMountError(err) {
  return err instanceof LifecycleError && err.action === OlObjectAction.MOUNT;
}
function isUnmountError(err) {
  return err instanceof LifecycleError && err.action === OlObjectAction.UNMOUNT;
}
function isDestroyError(err) {
  return err instanceof LifecycleError && err.action === OlObjectAction.DESTROY;
}
var OLObjectNotInitializedError = /*#__PURE__*/function (_Error3) {
  _inherits(OLObjectNotInitializedError, _Error3);

  var _super3 = _createSuper(OLObjectNotInitializedError);

  function OLObjectNotInitializedError(message) {
    var _this22;

    _classCallCheck(this, OLObjectNotInitializedError);

    _this22 = _super3.call(this, message);

    _defineProperty(_assertThisInitialized(_this22), "name", 'OLObjectNotInitializedError');

    if (isFunction(Error.captureStackTrace)) {
      Error.captureStackTrace(_assertThisInitialized(_this22), _this22.constructor);
    } else {
      _this22.stack = new Error(message).stack;
    }

    return _this22;
  }

  return OLObjectNotInitializedError;
}( /*#__PURE__*/_wrapNativeSuper(Error));
function makeChangeOrRecreateWatchers(props) {
  var deepProps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  return makeWatchers(props, function (prop) {
    return {
      deep: deepProps.includes(prop),
      handler: function handler() {
        var handler = this["".concat(prop, "Changed")];

        if (isFunction(handler)) {
          return handler.apply(void 0, arguments);
        }

        if (isEqual(arguments.length <= 0 ? undefined : arguments[0], arguments.length <= 1 ? undefined : arguments[1])) return;

        if (process.env.VUELAYERS_DEBUG) {
          this.$logger.log("".concat(prop, " changed, scheduling recreate... %O ===> %O"), arguments.length <= 1 ? undefined : arguments[1], arguments.length <= 0 ? undefined : arguments[0]);
        }

        return this.scheduleRecreate();
      }
    };
  });
}

function ownKeys$n(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$n(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$n(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$n(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

var serialize = function serialize(value) {
  if (value == null) return;
  return _typeof(value) === 'object' ? JSON.stringify(value) : value;
};

var cleanArcGisSourceParams = function cleanArcGisSourceParams(params) {
  return cleanSourceParams(params, ['FORMAT', 'F', 'LAYERS', 'LAYERDEFS', 'DYNAMICLAYERS', 'DPI', 'TRANSPARENT', 'TIME', 'LAYERTIMEOPTIONS', 'GDBVERSION', 'MAPSCALE', 'ROTATION', 'DATUMTRANSFORMATIONS', 'MAPRANGEVALUES', 'LAYERRANGEVALUES', 'LAYERPARAMETERVALUES', 'HISTORICMOMENT']);
};
/**
 * Shared ArcGIS source params and methods.
 */


var arcgisSource = {
  props: {
    // ArcGIS params from https://developers.arcgis.com/rest/services-reference/export-map.htm

    /**
     * @type {boolean}
     */
    hidpi: {
      type: Boolean,
      default: true
    },

    /**
     * @type {string}
     */
    format: {
      type: String,
      default: 'PNG32'
    },

    /**
     * @type {string|string[]|undefined}
     */
    layers: [String, Array],

    /**
     * @type {Object|string|undefined}
     */
    layerDefs: [Object, String],

    /**
     * @type {Object[]|string|undefined}
     */
    dynamicLayers: [Array, String],

    /**
     * @type {number|undefined}
     */
    dpi: Number,

    /**
     * @type {boolean}
     */
    transparent: {
      type: Boolean,
      default: true
    },

    /**
     * @type {string|string[]|undefined}
     */
    time: [String, Array],

    /**
     * @type {Object|string|undefined}
     */
    layerTimeOptions: [Object, String],

    /**
     * @type {string|undefined}
     */
    gdbVersion: String,

    /**
     * @type {string|undefined}
     */
    mapScale: String,

    /**
     * @type {number|undefined}
     */
    rotation: Number,

    /**
     * @type {Array|string|undefined}
     */
    datumTransformations: [Array, String],

    /**
     * @type {Array|string|undefined}
     */
    mapRangeValues: [Array, String],

    /**
     * @type {Array|string|undefined}
     */
    layerRangeValues: [Array, String],

    /**
     * @type {Array|string|undefined}
     */
    layerParameterValues: [Array, String],

    /**
     * @type {number|undefined}
     */
    historicMoment: Number,

    /**
     * Additional ArcGIS request parameters
     * @params {Object|undefined}
     */
    params: Object
  },
  data: function data() {
    return {
      currentParams: undefined
    };
  },
  computed: _objectSpread$n({
    /**
     * @returns {string|undefined}
     */
    inputLayers: function inputLayers() {
      return isArray(this.layers) ? this.layers.join(',') : this.layers;
    },

    /**
     * @type {string|undefined}
     */
    inputTime: function inputTime() {
      return isArray(this.time) ? this.time.join(',') : this.time;
    },

    /**
     * @return {undefined|string}
     */
    inputLayerDefs: function inputLayerDefs() {
      return serialize(this.layerDefs);
    },

    /**
     * @return {undefined|string}
     */
    inputDynamicLayers: function inputDynamicLayers() {
      return serialize(this.dynamicLayers);
    },

    /**
     * @return {undefined|string}
     */
    inputLayerTimeOptions: function inputLayerTimeOptions() {
      return serialize(this.layerTimeOptions);
    },

    /**
     * @return {undefined|string}
     */
    inputDatumTransformations: function inputDatumTransformations() {
      return serialize(this.datumTransformations);
    },

    /**
     * @return {undefined|string}
     */
    inputMapRangeValues: function inputMapRangeValues() {
      return serialize(this.mapRangeValues);
    },

    /**
     * @return {undefined|string}
     */
    inputLayerRangeValues: function inputLayerRangeValues() {
      return serialize(this.layerRangeValues);
    },

    /**
     * @return {undefined|string}
     */
    inputLayerParameterValues: function inputLayerParameterValues() {
      return serialize(this.layerParameterValues);
    },

    /**
     * @return {undefined|string}
     */
    inputHistoricMoment: function inputHistoricMoment() {
      return serialize(this.historicMoment);
    },

    /**
     * @returns {Object|null}
     */
    customParams: function customParams() {
      return this.params ? cleanArcGisSourceParams(this.params) : undefined;
    },

    /**
     * @returns {Object}
     */
    inputParams: function inputParams() {
      return _objectSpread$n(_objectSpread$n({}, this.customParams || {}), {}, {
        LAYERS: this.inputLayers,
        FORMAT: this.format,
        LAYERDEFS: this.inputLayerDefs,
        DYNAMICLAYERS: this.inputDynamicLayers,
        DPI: this.dpi,
        TRANSPARENT: this.transparent,
        TIME: this.inputTime,
        LAYERTIMEOPTIONS: this.inputLayerTimeOptions,
        GDBVERSION: this.gdbVersion,
        MAPSCALE: this.mapScale,
        ROTATION: this.rotation,
        DATUMTRANSFORMATIONS: this.inputDatumTransformations,
        MAPRANGEVALUES: this.inputMapRangeValues,
        LAYERRANGEVALUES: this.inputLayerRangeValues,
        LAYERPARAMETERVALUES: this.inputLayerParameterValues,
        HISTORICMOMENT: this.inputHistoricMoment
      });
    }
  }, /*#__PURE__*/reduce(['format', 'inputLayers', 'inputLayerDefs', 'inputDynamicLayers', 'dpi', 'transparent', 'inputTime', 'inputLayerTimeOptions', 'gdbVersion', 'mapScale', 'rotation', 'inputDatumTransformations', 'inputMapRangeValues', 'inputLayerRangeValues', 'inputLayerParameterValues', 'inputHistoricMoment'], function (props, inProp) {
    var prop = inProp.slice(0, 5) === 'input' ? lowerFirst(inProp.slice(5)) : inProp;
    var curProp = 'current' + upperFirst(prop);

    props[curProp] = function () {
      return coalesce((this.currentParams || {})[prop.toUpperCase()], (this.inputParams || {})[prop.toUpperCase()]);
    };

    return props;
  }, {})),
  watch: _objectSpread$n(_objectSpread$n({
    rev: function rev() {
      if (!this.$source) return;

      if (!isEqual(this.currentParams, this.$source.getParams())) {
        this.currentParams = this.$source.getParams();
      }
    }
  }, /*#__PURE__*/makeWatchers(['currentFormat', 'currentLayers', 'currentLayerDefs', 'currentDynamicLayers', 'currentDpi', 'currentTransparent', 'currentTime', 'currentLayerTimeOptions', 'currentGdbVersion', 'currentMapScale', 'currentRotation', 'currentDatumTransformations', 'currentMapRangeValues', 'currentLayerRangeValues', 'currentLayerParameterValues', 'currentHistoricMoment'], function (curProp) {
    return function (value) {
      var prop = lowerFirst(curProp.slice(7));
      var inProp = hasProp(this, 'input' + upperFirst(prop)) ? 'input' + upperFirst(prop) : prop;
      if (isEqual(value, this[inProp])) return;
      this.$emit('update:' + prop, isObjectLike(value) ? clonePlainObject(value) : value);
    };
  })), /*#__PURE__*/makeChangeOrRecreateWatchers(['hidpi', 'inputParams', 'currentParams'], ['inputParams', 'currentParams'])),
  created: function created() {
    this.currentParams = this.inputParams && clonePlainObject(this.inputParams);
  },
  methods: {
    /**
     * @returns {Object}
     */
    getParams: function getParams() {
      var _this$$source;

      return coalesce((_this$$source = this.$source) === null || _this$$source === void 0 ? void 0 : _this$$source.getParams(), this.currentParams);
    },

    /**
     * @param {Object} params
     */
    updateParams: function updateParams(params) {
      params = reduce(_objectSpread$n(_objectSpread$n({}, this.currentParams), params), function (params, value, name) {
        return _objectSpread$n(_objectSpread$n({}, params), {}, _defineProperty({}, name.toUpperCase(), value));
      }, {});

      if (!isEqual(params, this.currentParams)) {
        this.currentParams = params;
      }

      if (this.$source && !isEqual(params, this.$source.getParams())) {
        this.$source.updateParams(params);
      }
    },

    /**
     * @param {string} param
     * @param {*} value
     */
    updateParam: function updateParam(param, value) {
      this.updateParams(_defineProperty({}, param.toUpperCase(), value));
    },

    /**
     * @param {Object|undefined} value
     * @protected
     */
    inputParamsChanged: function inputParamsChanged(value) {
      this.updateParams(value);
    },

    /**
     * @param {Object|undefined} value
     * @protected
     */
    currentParamsChanged: function currentParamsChanged(value) {
      value = value ? cleanArcGisSourceParams(value) : undefined;
      if (isEqual(value, this.customParams)) return;
      this.$emit('update:params', value && clonePlainObject(value));
    }
  }
};

/**
 * Mixin with helpers for projection transforms between current view projection and global defined projection.
 */

var projTransforms = {
  computed: {
    /**
     * @return {module:ol/proj~ProjectionLike}
     */
    resolvedViewProjection: function resolvedViewProjection() {
      return coalesce(this.viewProjection, // may or may not be present
      EPSG_3857);
    },

    /**
     * @return {module:ol/proj~ProjectionLike}
     */
    resolvedDataProjection: function resolvedDataProjection() {
      var _this$$options;

      return coalesce(this.dataProjection, // may or may not be present
      (_this$$options = this.$options) === null || _this$$options === void 0 ? void 0 : _this$$options.dataProjection, // may or may not be present
      this.resolvedViewProjection);
    }
  },
  watch: {
    resolvedViewProjection: function resolvedViewProjection() {
      return this.resolvedViewProjectionChanged.apply(this, arguments);
    },
    resolvedDataProjection: function resolvedDataProjection() {
      return this.resolvedDataProjectionChanged.apply(this, arguments);
    }
  },
  methods: {
    pointToViewProj: function pointToViewProj(point) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformPoint(point, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    pointToDataProj: function pointToDataProj(point) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformPoint(point, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    lineToViewProj: function lineToViewProj(line) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformLine(line, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    lineToDataProj: function lineToDataProj(line) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformLine(line, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    polygonToViewProj: function polygonToViewProj(polygon) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformPolygon(polygon, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    polygonToDataProj: function polygonToDataProj(polygon) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformPolygon(polygon, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    multiPointToViewProj: function multiPointToViewProj(multiPoint) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformMultiPoint(multiPoint, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    multiPointToDataProj: function multiPointToDataProj(multiPoint) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformMultiPoint(multiPoint, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    multiLineToViewProj: function multiLineToViewProj(multiLine) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformMultiLine(multiLine, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    multiLineToDataProj: function multiLineToDataProj(multiLine) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformMultiLine(multiLine, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    multiPolygonToViewProj: function multiPolygonToViewProj(multiPolygon) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformMultiPolygon(multiPolygon, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    multiPolygonToDataProj: function multiPolygonToDataProj(multiPolygon) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformMultiPolygon(multiPolygon, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    extentToViewProj: function extentToViewProj(extent) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformExtent$1(extent, this.resolvedDataProjection, this.resolvedViewProjection, precision);
    },
    extentToDataProj: function extentToDataProj(extent) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return transformExtent$1(extent, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    writeGeometryInDataProj: function writeGeometryInDataProj(geometry) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return writeGeoJsonGeometry(geometry, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    writeGeometryInViewProj: function writeGeometryInViewProj(geometry) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return writeGeoJsonGeometry(geometry, this.resolvedViewProjection, this.resolvedViewProjection, precision);
    },
    readGeometryInDataProj: function readGeometryInDataProj(geometry) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return readGeoJsonGeometry(geometry, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    readGeometryInViewProj: function readGeometryInViewProj(geometry) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return readGeoJsonGeometry(geometry, this.resolvedViewProjection, this.resolvedViewProjection, precision);
    },
    writeFeatureInDataProj: function writeFeatureInDataProj(feature) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return writeGeoJsonFeature(feature, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    writeFeatureInViewProj: function writeFeatureInViewProj(feature) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return writeGeoJsonFeature(feature, this.resolvedViewProjection, this.resolvedViewProjection, precision);
    },
    readFeatureInDataProj: function readFeatureInDataProj(feature) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return readGeoJsonFeature(feature, this.resolvedViewProjection, this.resolvedDataProjection, precision);
    },
    readFeatureInViewProj: function readFeatureInViewProj(feature) {
      var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : COORD_PRECISION;
      return readGeoJsonFeature(feature, this.resolvedViewProjection, this.resolvedViewProjection, precision);
    },

    /**
     * @param {string} value
     * @param {string} prev
     * @protected
     */
    resolvedViewProjectionChanged: function resolvedViewProjectionChanged(value, prev) {
      if (value === prev) return;
      this.$emit('update:viewProjection', value);
    },

    /**
     * @param {string} value
     * @param {string} prev
     * @protected
     */
    resolvedDataProjectionChanged: function resolvedDataProjectionChanged(value, prev) {
      if (value === prev) return;
      this.$emit('update:dataProjection', value);
    }
  }
};

/**
 * Renders stub VNode for component.
 */

var stubVNode = {
  /**
   * @param {function} h
   * @returns {VNode}
   */
  render: function render(h) {
    var _this$$options;

    var options = ((_this$$options = this.$options) === null || _this$$options === void 0 ? void 0 : _this$$options.stubVNode) || {}; // render as HTML comment

    if (options.empty) {
      var vnode = h();

      if (typeof options.empty === 'string') {
        vnode.text = options.empty;
      } else if (typeof options.empty === 'function') {
        vnode.text = options.empty.call(this);
      }

      return vnode;
    }

    var children;

    if (options.slots === false) {
      children = undefined;
    } else {
      children = extractChildren(this.$slots, options.slots);
    }

    var attrs = typeof options.attrs === 'function' ? options.attrs.call(this) : options.attrs;
    var data = {
      attrs: attrs,
      style: {
        display: 'none !important'
      }
    };
    return h(options.tag || 'i', data, children);
  }
};

var waitForMap = {
  methods: {
    beforeInit: function beforeInit() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.prev = 0;
                _context.next = 3;
                return waitFor(function () {
                  return _this.$mapVm != null;
                }, race(_this.$olObjectEvents.pipe(filter(function (_ref) {
                  var name = _ref.name,
                      args = _ref.args;
                  return name === OlObjectEvent.ERROR && args[0] instanceof CanceledError;
                })), fromVueEvent(_this.$eventBus, OlObjectEvent.ERROR).pipe(filter(function (_ref2) {
                  var _ref3 = _slicedToArray(_ref2, 2),
                      err = _ref3[0],
                      vm = _ref3[1];

                  return isCreateError(err) && hasProp(vm, '$map') && _this.$vq.closest(vm);
                }))).pipe(mapTo(stubTrue())));

              case 3:
                if (_this.$mapVm.resolvedViewProjection !== _this.viewProjection) {
                  _this.viewProjection = _this.$mapVm.resolvedViewProjection;
                }

                if (_this.$mapVm.resolvedDataProjection !== _this.dataProjection) {
                  _this.dataProjection = _this.$mapVm.resolvedDataProjection;
                }

                _this.subscribeTo(fromVueWatcher(_this, function () {
                  return _this.$mapVm.resolvedViewProjection;
                }), function (_ref4) {
                  var value = _ref4.value;

                  if (value !== _this.viewProjection) {
                    _this.viewProjection = value;
                  }
                });

                _this.subscribeTo(fromVueWatcher(_this, function () {
                  return _this.$mapVm.resolvedDataProjection;
                }), function (_ref5) {
                  var value = _ref5.value;

                  if (value !== _this.dataProjection) {
                    _this.dataProjection = value;
                  }
                });

                _context.next = 9;
                return _this.$nextTickPromise();

              case 9:
                return _context.abrupt("return", olCmp.methods.beforeInit.call(_this));

              case 12:
                _context.prev = 12;
                _context.t0 = _context["catch"](0);
                _context.t0.message = "".concat(_this.vmName, " wait for $mapVm injection: ").concat(_context.t0.message);
                throw _context.t0;

              case 16:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, null, [[0, 12]]);
      }))();
    }
  }
};

function ownKeys$m(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$m(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$m(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$m(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base layer mixin.
 */

var baseLayer = {
  mixins: [stubVNode, projTransforms, olCmp, waitForMap],
  stubVNode: {
    attrs: function attrs() {
      return {
        id: this.vmId,
        class: this.vmClass
      };
    }
  },
  props: {
    // ol/layer/Base

    /**
     * A CSS class name to set to the layer element.
     * @type {string}
     */
    className: {
      type: String,
      default: 'ol-layer'
    },

    /**
     * @type {number}
     */
    opacity: {
      type: Number,
      default: 1
    },

    /**
     * @type {boolean}
     */
    visible: {
      type: Boolean,
      default: true
    },

    /**
     * @type {number[]|undefined}
     */
    extent: {
      type: Array,
      validator: function validator(value) {
        return value.length === 4 && value.every(isNumber);
      }
    },

    /**
     * @type {string|undefined}
     */
    extentProjection: {
      type: String,
      validator: function validator(value) {
        return !!get(value);
      }
    },

    /**
     * @type {number|undefined}
     */
    zIndex: Number,

    /**
     * @type {number|undefined}
     */
    minResolution: Number,

    /**
     * @type {number|undefined}
     */
    maxResolution: Number,

    /**
     * @type {number|undefined}
     */
    minZoom: Number,

    /**
     * @type {number|undefined}
     */
    maxZoom: Number
  },
  data: function data() {
    return {
      viewProjection: EPSG_3857,
      dataProjection: EPSG_3857,
      currentExtentViewProj: roundExtent(this.extent),
      currentMaxResolution: this.maxResolution,
      currentMinResolution: this.minResolution,
      currentMaxZoom: this.maxZoom,
      currentMinZoom: this.minZoom,
      currentOpacity: this.opacity,
      currentVisible: this.visible,
      currentZIndex: this.zIndex
    };
  },
  computed: {
    resolvedExtentProjection: function resolvedExtentProjection() {
      return coalesce(this.extentProjection, this.resolvedDataProjection);
    },
    extentDataProj: function extentDataProj() {
      return roundExtent(this.extent);
    },
    extentViewProj: function extentViewProj() {
      return transformExtent$1(this.extent, this.resolvedExtentProjection, this.resolvedViewProjection);
    },
    currentExtentDataProj: function currentExtentDataProj() {
      return transformExtent$1(this.currentExtentViewProj, this.resolvedViewProjection, this.resolvedExtentProjection);
    }
  },
  watch: _objectSpread$m({
    rev: function rev() {
      if (!this.$layer) return;

      if (!isEqual(this.currentExtentViewProj, this.$layer.getExtent())) {
        this.currentExtentViewProj = this.$layer.getExtent();
      }

      if (this.currentMaxResolution !== this.$layer.getMaxResolution()) {
        this.currentMaxResolution = this.$layer.getMaxResolution();
      }

      if (this.currentMinResolution !== this.$layer.getMinResolution()) {
        this.currentMinResolution = this.$layer.getMinResolution();
      }

      if (this.currentMaxZoom !== this.$layer.getMaxZoom()) {
        this.currentMaxZoom = this.$layer.getMaxZoom();
      }

      if (this.currentMinZoom !== this.$layer.getMinZoom()) {
        this.currentMinZoom = this.$layer.getMinZoom();
      }

      if (this.currentOpacity !== this.$layer.getOpacity()) {
        this.currentOpacity = this.$layer.getOpacity();
      }

      if (this.currentVisible !== this.$layer.getVisible()) {
        this.currentVisible = this.$layer.getVisible();
      }

      if (this.currentZIndex !== this.$layer.getZIndex()) {
        this.currentZIndex = this.$layer.getZIndex();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['className', 'opacity', 'currentOpacity', 'visible', 'currentVisible', 'extentViewProj', 'currentExtentDataProj', 'zIndex', 'currentZIndex', 'minResolution', 'currentMinResolution', 'maxResolution', 'currentMaxResolution', 'minZoom', 'currentMinZoom', 'maxZoom', 'currentMaxZoom'], ['extentViewProj', 'currentExtentDataProj'])),
  created: function created() {
    var _this$extentViewProj;

    defineServices$h.call(this);
    this.currentExtentViewProj = (_this$extentViewProj = this.extentViewProj) === null || _this$extentViewProj === void 0 ? void 0 : _this$extentViewProj.slice();
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeInit: function beforeInit() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all([olCmp.methods.beforeInit.call(_this), waitForMap.methods.beforeInit.call(_this)]);

              case 2:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {Promise<module:ol/layer/Base~BaseLayer>}
     * @protected
     */
    createOlObject: function createOlObject() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.t0 = initializeLayer;
                _context2.next = 3;
                return _this2.createLayer();

              case 3:
                _context2.t1 = _context2.sent;
                _context2.t2 = _this2.currentId;
                return _context2.abrupt("return", (0, _context2.t0)(_context2.t1, _context2.t2));

              case 6:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @return {module:ol/layer/Base~BaseLayer|Promise<module:ol/layer/Base~BaseLayer>}
     * @protected
     * @abstract
     */
    createLayer: function createLayer() {
      throw new Error("".concat(this.vmName, " not implemented method: createLayer"));
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        var _this3$$layersContain;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                (_this3$$layersContain = _this3.$layersContainer) === null || _this3$$layersContain === void 0 ? void 0 : _this3$$layersContain.addLayer(_this3);
                return _context3.abrupt("return", olCmp.methods.mount.call(_this3));

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this4 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
        var _this4$$layersContain;

        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                (_this4$$layersContain = _this4.$layersContainer) === null || _this4$$layersContain === void 0 ? void 0 : _this4$$layersContain.removeLayer(_this4);
                return _context4.abrupt("return", olCmp.methods.unmount.call(_this4));

              case 2:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4);
      }))();
    },

    /**
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(olCmp.methods.getServices.call(this), {
        get layerVm() {
          return vm;
        }

      });
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      olCmp.methods.subscribeAll.call(this);
      subscribeToLayerEvents$3.call(this);
    },

    /**
     * @returns {string|number}
     * @protected
     */
    getIdInternal: function getIdInternal() {
      return getLayerId(this.$layer);
    },

    /**
     * @param {string|number} id
     * @protected
     */
    setIdInternal: function setIdInternal(id) {
      if (id === this.getIdInternal()) return;
      setLayerId(this.$layer, id);
    },

    /**
     * @return {Promise<module:ol/layer/Base~BaseLayer>}
     */
    resolveLayer: olCmp.methods.resolveOlObject,

    /**
     * @param {boolean} [viewProj=false]
     * @returns {number[]|undefined}
     */
    getExtent: function getExtent() {
      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      if (!this.$layer) {
        return viewProj ? this.currentExtentViewProj : this.currentExtentDataProj;
      }

      var extent = this.$layer.getExtent();
      if (viewProj) return roundExtent(extent);
      return transformExtent$1(extent, this.resolvedViewProjection, this.resolvedExtentProjection);
    },

    /**
     * @param {number[]} extent
     * @param {boolean} [viewProj=false]
     */
    setExtent: function setExtent(extent) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      assert(extent == null || isArray(extent) && extent.length === 4, 'Invalid extent');
      extent = viewProj ? roundExtent(extent) : transformExtent$1(extent, this.resolvedExtentProjection, this.resolvedViewProjection);

      if (!isEqual(extent, this.currentExtentViewProj)) {
        this.currentExtentViewProj = extent;
      }

      if (this.$layer && !isEqual(extent, this.$layer.getExtent())) {
        this.$layer.setExtent(extent);
      }
    },

    /**
     * @returns {number|undefined}
     */
    getMaxResolution: function getMaxResolution() {
      var _this$$layer;

      return coalesce((_this$$layer = this.$layer) === null || _this$$layer === void 0 ? void 0 : _this$$layer.getMaxResolution(), this.currentMaxResolution);
    },

    /**
     * @param {number} resolution
     */
    setMaxResolution: function setMaxResolution(resolution) {
      resolution = Number(resolution);
      assert(isNumber(resolution), 'Invalid max resolution');

      if (resolution !== this.currentMaxResolution) {
        this.currentMaxResolution = resolution;
      }

      if (this.$layer && resolution !== this.$layer.getMaxResolution()) {
        this.$layer.setMaxResolution(resolution);
      }
    },

    /**
     * @returns {number|undefined}
     */
    getMinResolution: function getMinResolution() {
      var _this$$layer2;

      return coalesce((_this$$layer2 = this.$layer) === null || _this$$layer2 === void 0 ? void 0 : _this$$layer2.getMinResolution(), this.currentMinResolution);
    },

    /**
     * @param {number} resolution
     */
    setMinResolution: function setMinResolution(resolution) {
      resolution = Number(resolution);
      assert(isNumber(resolution), 'Invalid min resolution');

      if (resolution !== this.currentMinResolution) {
        this.currentMinResolution = resolution;
      }

      if (this.$layer && resolution !== this.$layer.getMinResolution()) {
        this.$layer.getMinResolution(resolution);
      }
    },

    /**
     * @returns {number}
     */
    getMaxZoom: function getMaxZoom() {
      var _this$$layer3;

      return coalesce((_this$$layer3 = this.$layer) === null || _this$$layer3 === void 0 ? void 0 : _this$$layer3.getMaxZoom(), this.currentMaxZoom);
    },

    /**
     * @param {number} zoom
     */
    setMaxZoom: function setMaxZoom(zoom) {
      zoom = Number(zoom);
      assert(isNumber(zoom), 'Invalid max zoom');

      if (zoom !== this.currentMaxZoom) {
        this.currentMaxZoom = zoom;
      }

      if (this.$layer && zoom !== this.$layer.getMaxZoom()) {
        this.$layer.setMaxZoom(zoom);
      }
    },

    /**
     * @returns {number}
     */
    getMinZoom: function getMinZoom() {
      var _this$$layer4;

      return coalesce((_this$$layer4 = this.$layer) === null || _this$$layer4 === void 0 ? void 0 : _this$$layer4.getMinZoom(), this.currentMinZoom);
    },

    /**
     * @param {number} zoom
     */
    setMinZoom: function setMinZoom(zoom) {
      zoom = Number(zoom);
      assert(isNumber(zoom), 'Invalid min zoom');

      if (zoom !== this.currentMinZoom) {
        this.currentMinZoom = zoom;
      }

      if (this.$layer && zoom !== this.$layer.getMinZoom()) {
        this.$layer.setMinZoom(zoom);
      }
    },

    /**
     * @returns {number}
     */
    getOpacity: function getOpacity() {
      var _this$$layer5;

      return coalesce((_this$$layer5 = this.$layer) === null || _this$$layer5 === void 0 ? void 0 : _this$$layer5.getOpacity(), this.currentOpacity);
    },

    /**
     * @param {number} opacity
     */
    setOpacity: function setOpacity(opacity) {
      opacity = Number(opacity);
      assert(isNumber(opacity), 'Invalid opacity');

      if (opacity !== this.currentOpacity) {
        this.currentOpacity = opacity;
      }

      if (this.$layer && opacity !== this.$layer.getOpacity()) {
        this.$layer.setOpacity(opacity);
      }
    },

    /**
     * @returns {boolean}
     */
    getVisible: function getVisible() {
      var _this$$layer6;

      return coalesce((_this$$layer6 = this.$layer) === null || _this$$layer6 === void 0 ? void 0 : _this$$layer6.getVisible(), this.currentVisible);
    },

    /**
     * @param {boolean} visible
     */
    setVisible: function setVisible(visible) {
      visible = !!visible;

      if (visible !== this.currentVisible) {
        this.currentVisible = visible;
      }

      if (this.$layer && visible !== this.$layer.getVisible()) {
        this.$layer.setVisible(visible);
      }
    },

    /**
     * @returns {number}
     */
    getZIndex: function getZIndex() {
      var _this$$layer7;

      return coalesce((_this$$layer7 = this.$layer) === null || _this$$layer7 === void 0 ? void 0 : _this$$layer7.getZIndex(), this.currentZIndex);
    },

    /**
     * @param {number} zIndex
     */
    setZIndex: function setZIndex(zIndex) {
      zIndex = Number(zIndex);
      assert(isNumber(zIndex), 'Invalid zIndex');

      if (zIndex !== this.currentZIndex) {
        this.currentZIndex = zIndex;
      }

      if (this.$layer && zIndex !== this.$layer.getZIndex()) {
        this.$layer.setZIndex(zIndex);
      }
    },

    /**
     * @param {number[]} pixel
     * @return {boolean}
     */
    isAtPixel: function isAtPixel(pixel) {
      var _this5 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
        var layer;
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return _this5.resolveLayer();

              case 2:
                layer = _context5.sent;
                return _context5.abrupt("return", _this5.$mapVm.forEachLayerAtPixel(pixel, function (mapLayer) {
                  return mapLayer === layer;
                }));

              case 4:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5);
      }))();
    },

    /**
     * @param {number} value
     * @protected
     */
    opacityChanged: function opacityChanged(value) {
      this.setOpacity(value);
    },

    /**
     * @param {number} value
     * @protected
     */
    currentOpacityChanged: function currentOpacityChanged(value) {
      if (value === this.opacity) return;
      this.$emit('update:opacity', value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    visibleChanged: function visibleChanged(value) {
      this.setVisible(value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    currentVisibleChanged: function currentVisibleChanged(value) {
      if (value === this.visible) return;
      this.$emit('update:visible', value);
    },

    /**
     * @param {number[]|undefined} value
     * @protected
     */
    extentViewProjChanged: function extentViewProjChanged(value) {
      this.setExtent(value, true);
    },

    /**
     * @param {number[]|undefined} value
     * @protected
     */
    currentExtentDataProjChanged: function currentExtentDataProjChanged(value) {
      if (isEqual(value, this.extentDataProj)) return;
      this.$emit('update:extent', value.slice());
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    zIndexChanged: function zIndexChanged(value) {
      this.setZIndex(value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    currentZIndexChanged: function currentZIndexChanged(value) {
      if (value === this.zIndex) return;
      this.$emit('update:zIndex', value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    minResolutionChanged: function minResolutionChanged(value) {
      this.setMinResolution(value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    currentMinResolutionChanged: function currentMinResolutionChanged(value) {
      if (value === this.minResolution) return;
      this.$emit('update:minResolution', value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    maxResolutionChanged: function maxResolutionChanged(value) {
      this.setMaxResolution(value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    currentMaxResolutionChanged: function currentMaxResolutionChanged(value) {
      if (value === this.maxResolution) return;
      this.$emit('update:maxResolution', value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    minZoomChanged: function minZoomChanged(value) {
      this.setMinZoom(value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    currentMinZoomChanged: function currentMinZoomChanged(value) {
      if (value === this.minZoom) return;
      this.$emit('update:minZoom', value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    maxZoomChanged: function maxZoomChanged(value) {
      this.setMaxZoom(value);
    },

    /**
     * @param {number|undefined} value
     * @protected
     */
    currentMaxZoomChanged: function currentMaxZoomChanged(value) {
      if (value === this.maxZoom) return;
      this.$emit('update:maxZoom', value);
    }
  }
};

function defineServices$h() {
  var _this6 = this;

  Object.defineProperties(this, {
    /**
     * @type {module:ol/layer/Base~BaseLayer|undefined}
     */
    $layer: {
      enumerable: true,
      get: function get() {
        return _this6.$olObject;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $mapVm: {
      enumerable: true,
      get: function get() {
        var _this6$$services;

        return (_this6$$services = _this6.$services) === null || _this6$$services === void 0 ? void 0 : _this6$$services.mapVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $viewVm: {
      enumerable: true,
      get: function get() {
        var _this6$$services2;

        return (_this6$$services2 = _this6.$services) === null || _this6$$services2 === void 0 ? void 0 : _this6$$services2.viewVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $layersContainer: {
      enumerable: true,
      get: function get() {
        var _this6$$services3;

        return (_this6$$services3 = _this6.$services) === null || _this6$$services3 === void 0 ? void 0 : _this6$$services3.layersContainer;
      }
    }
  });
}

function subscribeToLayerEvents$3() {
  return _subscribeToLayerEvents$3.apply(this, arguments);
}

function _subscribeToLayerEvents$3() {
  _subscribeToLayerEvents$3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee6() {
    var _this7 = this;

    var setterKey, propChanges;
    return _regeneratorRuntime.wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            setterKey = addPrefix('set');
            propChanges = fromOlChangeEvent(this.$layer, ['opacity', 'visible', 'zIndex', 'minResolution', 'maxResolution', 'minZoom', 'maxZoom', 'extent'], true, function (evt) {
              return _objectSpread$m(_objectSpread$m({}, evt), {}, {
                setter: function setter(val) {
                  var args = [val];

                  if (evt.prop === 'extent') {
                    args.push(true);
                  }

                  _this7[setterKey(evt.prop)].apply(_this7, args);
                }
              });
            });
            this.subscribeTo(propChanges, function (_ref) {
              var setter = _ref.setter,
                  value = _ref.value;
              return setter(value);
            });

          case 3:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee6, this);
  }));
  return _subscribeToLayerEvents$3.apply(this, arguments);
}

function ownKeys$l(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$l(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$l(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$l(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * @typedef {module:ol/control/Control~Control|Object} ControlLike
 */

/**
 * Controls collection
 */

var controlsContainer = {
  mixins: [identMap, rxSubs],
  computed: {
    /**
     * @type {string|undefined}
     */
    controlsCollectionIdent: function controlsCollectionIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'controls_collection');
    }
  },
  watch: _objectSpread$l({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['controlsCollectionIdent'])),
  created: function created() {
    /**
     * @type {module:ol/Collection~Collection<module:ol/control/Control~Control>}
     * @private
     */
    this._controlsCollection = this.instanceFactoryCall(this.controlsCollectionIdent, function () {
      return new Collection();
    });
    this._controlSubs = {};
    defineServices$g.call(this);
  },
  methods: {
    /**
     * @returns {{readonly controlsContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get controlsContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      subscribeToCollectionEvents$4.call(this);
    },

    /**
     * @param {ControlLike[]|module:ol/Collection~Collection<ControlLike>} defaultControls
     */
    initDefaultControls: function initDefaultControls(defaultControls) {
      var _this = this;

      this.getControls().forEach(function (control) {
        if (control.get('vl_default')) {
          _this.removeControl(control);
        }
      });
      var controls;

      if (isArray(defaultControls) || defaultControls instanceof Collection) {
        controls = defaultControls;
      } else if (defaultControls !== false) {
        controls = defaults(isPlainObject(defaultControls) ? defaultControls : undefined);
      }

      if (controls) {
        controls.forEach(function (control) {
          return control.set('vl_default', true);
        });
        this.addControls(controls);
      }
    },

    /**
     * @param {ControlLike} control
     * @return {Control}
     */
    initializeControl: function initializeControl$1(control) {
      var _control;

      control = ((_control = control) === null || _control === void 0 ? void 0 : _control.$control) || control;
      instanceOf(control, Control);
      return initializeControl(control);
    },

    /**
     * @param {ControlLike[]|module:ol/Collection~Collection<ControlLike>} controls
     */
    addControls: function addControls(controls) {
      forEach(controls, this.addControl.bind(this));
    },

    /**
     * @param {ControlLike} control
     */
    addControl: function addControl(control) {
      control = this.initializeControl(control);
      if (this.getControlById(getControlId(control))) return;
      this.$controlsCollection.push(control);
    },

    /**
     * @param {ControlLike[]|module:ol/Collection~Collection<ControlLike>} controls
     */
    removeControls: function removeControls(controls) {
      forEach(controls, this.removeControl.bind(this));
    },

    /**
     * @param {ControlLike} control
     */
    removeControl: function removeControl(control) {
      var _control2;

      control = this.getControlById(getControlId(((_control2 = control) === null || _control2 === void 0 ? void 0 : _control2.$control) || control));
      if (!control) return;
      this.$controlsCollection.remove(control);
    },

    /**
     * @return {void}
     */
    clearControls: function clearControls() {
      this.$controlsCollection.clear();
    },

    /**
     * @returns {Array<module:ol/control/Control~Control>}
     */
    getControls: function getControls() {
      return this.$controlsCollection.getArray().slice();
    },

    /**
     * @returns {module:ol/Collection~Collection<module:ol/control/Control~Control>}
     */
    getControlsCollection: function getControlsCollection() {
      return this._controlsCollection;
    },

    /**
     * @param {string|number} controlId
     * @returns {ControlLike}
     */
    getControlById: function getControlById(controlId) {
      return find(this.getControls(), function (control) {
        return getControlId(control) === controlId;
      });
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    controlsCollectionIdentChanged: function controlsCollectionIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.$controlsCollection) {
        this.setInstance(value, this.$controlsCollection);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    }
  }
};

function defineServices$g() {
  Object.defineProperties(this, {
    $controlsCollection: {
      enumerable: true,
      get: this.getControlsCollection
    }
  });
}

function subscribeToCollectionEvents$4() {
  var _this2 = this;

  var adds = fromOlEvent(this.$controlsCollection, CollectionEventType.ADD).pipe(map(function (evt) {
    return _objectSpread$l(_objectSpread$l({}, evt), {}, {
      element: _this2.initializeControl(evt.element)
    });
  }), tap(function (_ref) {
    var element = _ref.element;
    var uid = getUid(element);
    var propChanges = fromOlChangeEvent(element, 'id', true);
    _this2._controlSubs[uid] = _this2.subscribeTo(propChanges, _this2.scheduleRefresh.bind(_this2));
  }));
  var removes = fromOlEvent(this.$controlsCollection, CollectionEventType.REMOVE).pipe(tap(function (_ref2) {
    var element = _ref2.element;
    var uid = getUid(element);

    if (_this2._controlSubs[uid]) {
      _this2.unsubscribe(_this2._controlSubs[uid]);

      delete _this2._controlSubs[uid];
    }
  }));
  var events = merge(adds, removes).pipe(bufferDebounceTime(FRAME_TIME));
  this.subscribeTo(events, /*#__PURE__*/function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee(events) {
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _context.next = 2;
              return _this2.debounceChanged();

            case 2:
              forEach(events, function (_ref4) {
                var type = _ref4.type,
                    element = _ref4.element;

                _this2.$emit(type + 'control', element); // todo remove in v0.13.x


                _this2.$emit(type + ':control', element);
              });

            case 3:
            case "end":
              return _context.stop();
          }
        }
      }, _callee);
    }));

    return function (_x) {
      return _ref3.apply(this, arguments);
    };
  }());
}

/**
 * @typedef {Geometry|Object} GeometryLike
 */

/**
 * @typedef {Object} GeometryTarget
 * @property {function(): Geometry|undefined} getGeometry
 * @property {function(Geometry|undefined): void} setGeometry
 */

/**
 * Geometry container
 */

var geometryContainer = {
  mixins: [projTransforms],
  created: function created() {
    this._geometry = undefined;
    this._geometryVm = undefined;
    defineService.call(this);
  },
  methods: {
    /**
     * @returns {{readonly geometryContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get geometryContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {GeometryTarget|undefined}
     * @protected
     */
    getGeometryTarget: function getGeometryTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getGeometryTarget()"));
    },

    /**
     * @return {module:ol/geom/Geometry~Geometry|undefined}
     */
    getGeometry: function getGeometry() {
      var _this$getGeometryTarg;

      return coalesce((_this$getGeometryTarg = this.getGeometryTarget()) === null || _this$getGeometryTarg === void 0 ? void 0 : _this$getGeometryTarg.getGeometry(), this._geometry);
    },

    /**
     * @return {Object}
     */
    getGeometryVm: function getGeometryVm() {
      return this._geometryVm;
    },

    /**
     * @param {GeometryLike|undefined} geom
     * @param {boolean} [viewProj=false]
     */
    setGeometry: function setGeometry(geom) {
      var _geom;

      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      geom = ((_geom = geom) === null || _geom === void 0 ? void 0 : _geom.$geometry) || geom;

      if (isGeoJSONGeometry(geom)) {
        if (viewProj) {
          geom = this.readGeometryInViewProj(geom);
        } else {
          geom = this.readGeometryInDataProj(geom);
        }
      }

      geom || (geom = undefined);
      assert(!geom || geom instanceof Geometry, 'Invalid geometry');
      var geomTarget = this.getGeometryTarget();

      if (geomTarget && geom !== geomTarget.getGeometry()) {
        geomTarget.setGeometry(geom);
        this.scheduleRefresh();
      }

      if (geom !== this._geometry) {
        var _geom2;

        this._geometry = geom;
        this._geometryVm = ((_geom2 = geom) === null || _geom2 === void 0 ? void 0 : _geom2.vm) && geom.vm[0];
        this.scheduleRefresh();
      }
    }
  }
};

function defineService() {
  Object.defineProperties(this, {
    $geometry: {
      enumerable: true,
      get: this.getGeometry
    },
    $geometryVm: {
      enumerable: true,
      get: this.getGeometryVm
    }
  });
}

/**
 * @typedef {
 *            module:ol/style/Style~Style |
 *            Array<module:ol/style/Style~Style> |
 *            module:ol/style/Style~StyleFunction
 *          } OlStyleLike
 */

/**
 * @typedef {Object} StyleTarget
 * @property {function(OlStyleLike|undefined): void} setStyle
 * @property {function(): OlStyleLike|undefined} getStyle
 */

/**
 * @typedef {OlStyleLike|Object} StyleLike
 */

/**
 * Style container mixin.
 */

var styleContainer = {
  created: function created() {
    this._style = undefined;
    defineServices$f.call(this);
  },
  methods: {
    /**
     * @returns {{readonly styleContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get styleContainer() {
          return vm;
        }

      };
    },

    /**
     * Default style factory
     * @return {OlStyleLike|undefined}
     * @protected
     */
    getDefaultStyle: function getDefaultStyle() {},

    /**
     * Returns OL object that can be styled (i.e. has setStyle/getStyle methods) or undefined
     * @return {StyleTarget|undefined}
     * @protected
     * @abstract
     */
    getStyleTarget: function getStyleTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getStyleTarget()"));
    },

    /**
     * @return {StyleLike|undefined}
     */
    getStyle: function getStyle() {
      var _this$getStyleTarget;

      return coalesce((_this$getStyleTarget = this.getStyleTarget()) === null || _this$getStyleTarget === void 0 ? void 0 : _this$getStyleTarget.getStyle(), this._style);
    },

    /**
     * @param {StyleLike} style
     */
    addStyle: function addStyle(style) {
      if (!style) return;
      var olStyle = (style === null || style === void 0 ? void 0 : style.$style) || style;
      var currentStyle = this._style;

      if (isFunction(olStyle)) {
        if (currentStyle) {
          if (process.env.NODE_ENV !== 'production') {
            this.$logger.warn('Component already has style components among its descendants. ' + 'Avoid use of multiple vl-style-func or combining vl-style-func with vl-style-box on the same level.');
          }
        }

        currentStyle = style;
      } else {
        if (!isArray(currentStyle)) {
          if (currentStyle) {
            if (process.env.NODE_ENV !== 'production') {
              this.$logger.warn('Component already has style components among its descendants. ' + 'Avoid use of multiple vl-style-func or combining vl-style-func with vl-style-box on the same level.');
            }
          }

          currentStyle = [];
        }

        if (!currentStyle.includes(olStyle)) {
          currentStyle.push(olStyle);
        }
      }

      this.setStyle(currentStyle);
    },

    /**
     * @param {StyleLike|undefined} style
     */
    removeStyle: function removeStyle(style) {
      var _style;

      if (!style) return;
      style = ((_style = style) === null || _style === void 0 ? void 0 : _style.$style) || style;
      var currentStyle = this._style;

      if (currentStyle === style) {
        currentStyle = undefined;
      } else if (isArray(currentStyle)) {
        currentStyle = currentStyle.filter(function (s) {
          return s !== style;
        });

        if (currentStyle.length === 0) {
          currentStyle = undefined;
        }
      }

      this.setStyle(currentStyle);
    },

    /**
     * @param {StyleLike|undefined} style
     */
    setStyle: function setStyle(style) {
      style || (style = undefined);

      if (style) {
        if (hasProp(style, '$style') || hasProp(style, '$styleFunction')) {
          style = style.$style || style.$styleFunction;
        } else if (isArray(style)) {
          style = style.map(function (style) {
            return (style === null || style === void 0 ? void 0 : style.$style) || style;
          });
        }

        if (isFunction(style)) {
          style = this.createStyleFunc(style, this.getDefaultStyle());
        } else {
          isArray(style) || (style = [style]);
          style.length > 0 || (style = undefined);
        }
      }

      var styleTarget = this.getStyleTarget();

      if (styleTarget && !isEqual(style, styleTarget.getStyle())) {
        styleTarget.setStyle(style);
        this.scheduleRefresh();
      }

      if (!isEqual(style, this._style)) {
        this._style = style;
        this.scheduleRefresh();
      }
    },

    /**
     * Style function factory
     * @param {StyleLike|undefined} style
     * @param {StyleLike|undefined} defaultStyle
     * @returns {module:ol/style/Style~StyleFunction}
     * @protected
     */
    createStyleFunc: function createStyleFunc(style, defaultStyle) {
      return function __styleFunc(feature, resolution) {
        if (!feature.getGeometry()) return;
        var compiledStyle;

        if (style && isFunction(style)) {
          // style - custom ol/style/Style~StyleFunction
          compiledStyle = style(feature, resolution);
        } else if (isArray(style)) {
          // style - array of ol/style/Style objects
          compiledStyle = style.slice();
        } // not empty or null style


        if (compiledStyle == null || isArray(compiledStyle) && compiledStyle.length || compiledStyle instanceof Style) {
          return compiledStyle;
        } // fallback to default style


        if (defaultStyle) {
          return isFunction(defaultStyle) ? defaultStyle(feature, resolution) : defaultStyle;
        }

        return null;
      };
    }
  }
};

function defineServices$f() {
  Object.defineProperties(this, {
    $style: {
      enumerable: true,
      get: this.getStyle
    }
  });
}

function ownKeys$k(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$k(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$k(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$k(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * A vector object for geographic features with a geometry and other attribute properties,
 * similar to the features in vector file formats like **GeoJSON**.
 */

var feature = {
  mixins: [stubVNode, projTransforms, geometryContainer, styleContainer, olCmp, waitForMap],
  stubVNode: {
    empty: false,
    attrs: function attrs() {
      return {
        id: this.vmId,
        class: this.vmClass
      };
    }
  },
  props: {
    properties: {
      type: Object,
      default: stubObject
    }
  },
  data: function data() {
    return {
      viewProjection: EPSG_3857,
      dataProjection: EPSG_3857,
      currentProperties: clonePlainObject(this.properties),
      currentGeometryName: 'geometry'
    };
  },
  computed: {
    geometryDataProj: function geometryDataProj() {
      if (!(this.rev && this.$geometry)) return;
      return this.writeGeometryInDataProj(this.$geometry);
    },
    geometryViewProj: function geometryViewProj() {
      if (!(this.rev && this.$geometry)) return;
      return this.writeGeometryInViewProj(this.$geometry);
    },
    pointDataProj: function pointDataProj() {
      return this.pointToDataProj(this.pointViewProj);
    },
    pointViewProj: function pointViewProj() {
      if (!(this.rev && this.$geometry)) return;
      return findPointOnSurface(this.$geometry);
    },
    style: function style() {
      var _this = this;

      if (!(this.rev && this.$style)) return;
      var style = this.$style;
      if (isFunction(style)) return style;
      if (!style) return;
      isArray(style) || (style = [style]);
      return style.map(function (style) {
        return dumpStyle(style, function (geom) {
          return _this.writeGeometryInDataProj(geom);
        });
      });
    }
  },
  watch: _objectSpread$k({
    rev: function rev() {
      if (!this.$feature) return;

      if (!isEqual(this.currentProperties, getFeatureProperties(this.$feature))) {
        this.currentProperties = getFeatureProperties(this.$feature);
      }

      if (this.currentGeometryName !== this.$feature.getGeometryName()) {
        this.currentGeometryName = this.$feature.getGeometryName();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['properties', 'currentProperties', 'currentGeometryName', 'geometryDataProj', 'pointDataProj', 'style'], ['properties', 'currentProperties', 'geometryDataProj', 'pointDataProj', 'style'])),
  created: function created() {
    defineServices$e.call(this);
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeInit: function beforeInit() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all([olCmp.methods.beforeInit.call(_this2), waitForMap.methods.beforeInit.call(_this2)]);

              case 2:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * Create feature without inner style applying, feature level style
     * will be applied in the layer level style function.
     * @return {module:ol/Feature~Feature}
     * @protected
     */
    createOlObject: function createOlObject() {
      var feature = initializeFeature(this.createFeature(), this.currentId);
      feature.setGeometryName(this.currentGeometryName);
      feature.setGeometry(this.$geometry);
      feature.setStyle(this.$style);
      return feature;
    },

    /**
     * @returns {Feature}
     */
    createFeature: function createFeature() {
      return new Feature(this.currentProperties);
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeMount: function beforeMount() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.prev = 0;
                _context2.next = 3;
                return waitFor(function () {
                  return _this3.$geometryVm != null;
                }, race(_this3.$olObjectEvents.pipe(filter(function (_ref) {
                  var name = _ref.name,
                      args = _ref.args;
                  return name === OlObjectEvent.ERROR && args[0] instanceof CanceledError;
                })), fromVueEvent(_this3.$eventBus, ObjectEventType.ERROR).pipe(filter(function (_ref2) {
                  var _vm$$vq;

                  var _ref3 = _slicedToArray(_ref2, 2),
                      err = _ref3[0],
                      vm = _ref3[1];

                  return (isCreateError(err) || isMountError(err)) && hasProp(vm, '$geometry') && ((_vm$$vq = vm.$vq) === null || _vm$$vq === void 0 ? void 0 : _vm$$vq.closest(_this3));
                }))).pipe(mapTo(stubTrue())));

              case 3:
                return _context2.abrupt("return", olCmp.methods.beforeMount.call(_this3));

              case 6:
                _context2.prev = 6;
                _context2.t0 = _context2["catch"](0);
                _context2.t0.message = "".concat(_this3.vmName, " wait for $geometry failed: ").concat(_context2.t0.message);
                throw _context2.t0;

              case 10:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, null, [[0, 6]]);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this4 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        var _this4$$featuresConta;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                (_this4$$featuresConta = _this4.$featuresContainer) === null || _this4$$featuresConta === void 0 ? void 0 : _this4$$featuresConta.addFeature(_this4);
                return _context3.abrupt("return", olCmp.methods.mount.call(_this4));

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this5 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
        var _this5$$featuresConta;

        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                (_this5$$featuresConta = _this5.$featuresContainer) === null || _this5$$featuresConta === void 0 ? void 0 : _this5$$featuresConta.removeFeature(_this5);
                return _context4.abrupt("return", olCmp.methods.unmount.call(_this5));

              case 2:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4);
      }))();
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      olCmp.methods.subscribeAll.call(this);
      subscribeToEvents.call(this);
    },

    /**
     * @return {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(olCmp.methods.getServices.call(this), geometryContainer.methods.getServices.call(this), styleContainer.methods.getServices.call(this), {
        get featureVm() {
          return vm;
        }

      });
    },

    /**
     * @return {string|number}
     */
    getIdInternal: function getIdInternal() {
      return getFeatureId(this.$feature);
    },

    /**
     * @param {string|number} id
     * @return {void}
     */
    setIdInternal: function setIdInternal(id) {
      if (id === this.getIdInternal()) return;
      setFeatureId(this.$feature, id);
    },

    /**
     * @return {Promise<Feature>}
     */
    resolveFeature: olCmp.methods.resolveOlObject,

    /**
     * @return {Feature}
     * @protected
     */
    getGeometryTarget: function getGeometryTarget() {
      return this.$feature;
    },

    /**
     * @return {Feature}
     * @protected
     */
    getStyleTarget: function getStyleTarget() {
      return this.$feature;
    },

    /**
     * @return {string}
     */
    getGeometryName: function getGeometryName() {
      var _this$$feature;

      return coalesce((_this$$feature = this.$feature) === null || _this$$feature === void 0 ? void 0 : _this$$feature.getGeometryName(), this.currentGeometryName);
    },

    /**
     * @param {string} geometryName
     */
    setGeometryName: function setGeometryName(geometryName) {
      assert(isString(geometryName) && !isEmpty(geometryName), 'Invalid geometry name');

      if (geometryName !== this.currentGeometryName) {
        this.currentGeometryName = geometryName;
      }

      if (this.$feature && geometryName !== this.$feature.getGeometryName()) {
        this.$feature.setGeometryName(geometryName);
      }
    },

    /**
     * @return {Object}
     */
    getProperties: function getProperties() {
      return coalesce(this.$feature && getFeatureProperties(this.$feature), this.currentProperties);
    },

    /**
     * @param {Object} properties
     */
    setProperties: function setProperties(properties) {
      properties = getFeatureProperties({
        properties: properties
      });

      if (!isEqual(properties, this.currentProperties)) {
        this.currentProperties = properties;
      }

      if (this.$feature && !isEqual(properties, getFeatureProperties(this.$feature))) {
        setFeatureProperties(this.$feature, properties);
      }
    },

    /**
     * Checks if feature lies at `pixel`.
     * @param {number[]} pixel
     * @return {Promise<boolean>}
     */
    isAtPixel: function isAtPixel(pixel) {
      var _this6 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
        var selfFeature, layerFilter, selfLayer;
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return _this6.resolveFeature();

              case 2:
                selfFeature = _context5.sent;

                if (!_this6.$layerVm) {
                  _context5.next = 8;
                  break;
                }

                _context5.next = 6;
                return _this6.$layerVm.resolveLayer();

              case 6:
                selfLayer = _context5.sent;

                layerFilter = function layerFilter(layer) {
                  return layer === selfLayer;
                };

              case 8:
                return _context5.abrupt("return", _this6.$mapVm.forEachFeatureAtPixel(pixel, function (feature) {
                  return feature === selfFeature;
                }, {
                  layerFilter: layerFilter
                }));

              case 9:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5);
      }))();
    },

    /**
     * @param {Object|undefined} value
     * @protected
     */
    propertiesChanged: function propertiesChanged(value) {
      this.setProperties(value);
    },

    /**
     * @param {Object|undefined} value
     * @protected
     */
    currentPropertiesChanged: function currentPropertiesChanged(value) {
      if (isEqual(value, this.properties)) return;
      this.$emit('update:properties', value && clonePlainObject(value));
    },

    /**
     * @param {string} value
     * @param {string} prev
     * @protected
     */
    currentGeometryNameChanged: function currentGeometryNameChanged(value, prev) {
      if (value === prev) return;
      this.$emit('update:geometryName', value);
    },

    /**
     * @param {Object|undefined} value
     * @param {Object|undefined} prev
     * @protected
     */
    geometryDataProjChanged: function geometryDataProjChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:geometry', value && clonePlainObject(value));
    },

    /**
     * @param {Object|undefined} value
     * @param {Object|undefined} prev
     * @protected
     */
    pointDataProjChanged: function pointDataProjChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:point', value && clonePlainObject(value));
    },

    /**
     * @param {Object|Function|Array|undefined} value
     * @param {Object|Function|Array|undefined} prev
     * @protected
     */
    styleChanged: function styleChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:style', isObjectLike(value) ? clonePlainObject(value) : value);
    }
  }
};

function defineServices$e() {
  var _this7 = this;

  Object.defineProperties(this, {
    $feature: {
      enumerable: true,
      get: function get() {
        return _this7.$olObject;
      }
    },
    $layerVm: {
      enumerable: true,
      get: function get() {
        var _this7$$services;

        return (_this7$$services = _this7.$services) === null || _this7$$services === void 0 ? void 0 : _this7$$services.layerVm;
      }
    },
    $mapVm: {
      enumerable: true,
      get: function get() {
        var _this7$$services2;

        return (_this7$$services2 = _this7.$services) === null || _this7$$services2 === void 0 ? void 0 : _this7$$services2.mapVm;
      }
    },
    $featuresContainer: {
      enumerable: true,
      get: function get() {
        var _this7$$services3;

        return (_this7$$services3 = _this7.$services) === null || _this7$$services3 === void 0 ? void 0 : _this7$$services3.featuresContainer;
      }
    }
  });
}

function subscribeToEvents() {
  return _subscribeToEvents.apply(this, arguments);
}

function _subscribeToEvents() {
  _subscribeToEvents = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee6() {
    var _this8 = this;

    var propChanges;
    return _regeneratorRuntime.wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            propChanges = fromOlEvent(this.$feature, ObjectEventType.PROPERTYCHANGE, function (_ref4) {
              var key = _ref4.key;

              switch (key) {
                case _this8.$feature.getGeometryName():
                  return {
                    prop: 'geometry',
                    value: _this8.$feature.getGeometry(),
                    setter: function setter(geom) {
                      return _this8.setGeometry(geom, true);
                    }
                  };

                default:
                  return {
                    prop: 'properties',
                    value: getFeatureProperties(_this8.$feature),
                    setter: _this8.setProperties
                  };
              }
            }).pipe(distinctUntilChanged(isEqual));
            this.subscribeTo(propChanges, function (_ref5) {
              var setter = _ref5.setter,
                  value = _ref5.value;
              return setter(value);
            });

          case 2:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee6, this);
  }));
  return _subscribeToEvents.apply(this, arguments);
}

var featureHelper = {
  methods: {
    /**
     * @param {FeatureLike} feature
     * @param {boolean} [viewProj=false]
     * @return {Feature}
     * @protected
     */
    initializeFeature: function initializeFeature$1(feature) {
      var _feature;

      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      feature = ((_feature = feature) === null || _feature === void 0 ? void 0 : _feature.$feature) || feature;

      if (isPlainObject(feature)) {
        if (viewProj) {
          feature = this.readFeatureInViewProj(feature);
        } else {
          feature = this.readFeatureInDataProj(feature);
        }
      }

      return initializeFeature(feature);
    },

    /**
     * @param {module:ol/Feature~Feature} feature
     * @param {module:ol/Feature~Feature} newFeature
     * @protected
     */
    updateFeature: function updateFeature(feature, newFeature) {
      var _this$$mapVm;

      if ((_this$$mapVm = this.$mapVm) !== null && _this$$mapVm !== void 0 && _this$$mapVm.isInteracting()) return;
      var featureJson = this.writeFeatureInViewProj(feature);
      var newFeatureJson = this.writeFeatureInViewProj(newFeature);
      if (isEqual(featureJson, newFeatureJson)) return;

      if (getFeatureId(feature) !== getFeatureId(newFeature)) {
        setFeatureId(feature, getFeatureId(newFeature));
      }

      var properties = getFeatureProperties(newFeature);
      var currentProperties = getFeatureProperties(feature);

      if (!isEqual(properties, currentProperties)) {
        setFeatureProperties(feature, properties);
      }

      var geomJson = get$1(newFeatureJson, "properties.".concat(CIRCLE_SERIALIZE_PROP)) || newFeatureJson.geometry || null;
      var currentGeomJson = get$1(featureJson, "properties.".concat(CIRCLE_SERIALIZE_PROP)) || featureJson.geometry || null;

      if (!isEqual(geomJson, currentGeomJson)) {
        feature.setGeometry(newFeature.getGeometry() || null);
      }

      var style = newFeature.getStyle();
      var currentStyle = feature.getStyle();

      if (isFunction(currentStyle) && isFunction(style)) {
        if (currentStyle !== style) {
          feature.setStyle(style);
        }
      } else if (isFunction(style)) {
        feature.setStyle(style);
      }

      var styleJson = get$1(newFeatureJson, "properties.".concat(STYLE_SERIALIZE_PROP)) || null;
      var currentStyleJson = get$1(featureJson, "properties.".concat(STYLE_SERIALIZE_PROP)) || null;

      if (!isEqual(styleJson, currentStyleJson)) {
        feature.setStyle(style);
      }
    }
  }
};

function ownKeys$j(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$j(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$j(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$j(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * @typedef {module:ol/Feature~Feature|Object} FeatureLike
 */

/**
 * Features container
 */

var featuresContainer = {
  mixins: [identMap, rxSubs, projTransforms, featureHelper],
  computed: {
    /**
     * @returns {string|undefined}
     */
    featuresCollectionIdent: function featuresCollectionIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'features_collection');
    }
  },
  watch: _objectSpread$j({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['featuresCollectionIdent'])),
  created: function created() {
    /**
     * @type {module:ol/Collection~Collection<module:ol/Feature~Feature>}
     * @private
     */
    this._featuresCollection = this.instanceFactoryCall(this.featuresCollectionIdent, function () {
      return new Collection();
    });
    this._featureSubs = {};
    defineServices$d.call(this);
  },
  methods: {
    /**
     * @returns {{readonly featuresContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get featuresContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      subscribeToCollectionEvents$3.call(this);
    },

    /**
     * @param {FeatureLike[]|module:ol/Collection~Collection<FeatureLike>} features
     * @param {boolean} [viewProj=false]
     */
    addFeatures: function addFeatures(features) {
      var _this = this;

      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      forEach(features, function (feature) {
        return _this.addFeature(feature, viewProj);
      });
    },

    /**
     * @param {FeatureLike} feature
     * @param {boolean} [viewProj=false]
     */
    addFeature: function addFeature(feature) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      feature = this.initializeFeature(feature, viewProj); // todo add hash {featureId => featureIdx, ....}

      var foundFeature = this.getFeatureById(getFeatureId(feature));

      if (foundFeature == null) {
        this.$featuresCollection.push(feature);
      } else {
        this.updateFeature(foundFeature, feature);
      }
    },

    /**
     * @param {FeatureLike[]|module:ol/Collection~Collection<FeatureLike>} features
     */
    removeFeatures: function removeFeatures(features) {
      forEach(features, this.removeFeature.bind(this));
    },

    /**
     * @param {FeatureLike} feature
     */
    removeFeature: function removeFeature(feature) {
      var _feature;

      feature = this.getFeatureById(getFeatureId(((_feature = feature) === null || _feature === void 0 ? void 0 : _feature.$feature) || feature));
      if (!feature) return;
      this.$featuresCollection.remove(feature);
    },

    /**
     * @return {void}
     */
    clearFeatures: function clearFeatures() {
      this.$featuresCollection.clear();
    },

    /**
     * @param {string|number} featureId
     * @return {module:ol/Feature~Feature|undefined}
     */
    getFeatureById: function getFeatureById(featureId) {
      // todo add hash {featureId => featureIdx, ....}
      return find(this.getFeatures(), function (feature) {
        return getFeatureId(feature) === featureId;
      });
    },

    /**
     * @return {Array<module:ol/Feature~Feature>}
     */
    getFeatures: function getFeatures() {
      return this.$featuresCollection.getArray().slice();
    },

    /**
     * @return {module:ol/Collection~Collection<module:ol/Feature~Feature>}
     */
    getFeaturesCollection: function getFeaturesCollection() {
      return this._featuresCollection;
    },

    /**
     * @param {function} callback
     * @return {*}
     */
    forEachFeature: function forEachFeature(callback) {
      var res;
      this.$featuresCollection.forEach(function () {
        res = callback.apply(void 0, arguments);
      });
      return res;
    },

    /**
     * @param {number[]} extent
     * @param {function} callback
     * @param {boolean} [viewProj=false]
     * @returns {*}
     */
    forEachFeatureInExtent: function forEachFeatureInExtent(extent, callback) {
      var viewProj = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      extent = viewProj ? roundExtent(extent) : this.extentToViewProj(extent);
      return this.forEachFeature(function (feature) {
        var geometry = feature.getGeometry();

        if (geometry && geometry.intersectsExtent(extent)) {
          var result = callback(feature);

          if (result) {
            return result;
          }
        }
      });
    },

    /**
     * @param {number[]} extent
     * @param {function} callback
     * @param {boolean} [viewProj=false]
     * @returns {*}
     */
    forEachFeatureIntersectingExtent: function forEachFeatureIntersectingExtent(extent, callback) {
      var viewProj = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      return this.forEachFeatureInExtent(extent, callback, viewProj);
    },

    /**
     * @param {number[]} coordinate
     * @param {boolean} [viewProj=false]
     * @returns {Array<module:ol/Feature~Feature>}
     */
    getFeaturesAtCoordinate: function getFeaturesAtCoordinate(coordinate) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      return this.getFeaturesInExtent([coordinate[0], coordinate[1], coordinate[0], coordinate[1]], viewProj);
    },

    /**
     * @param {number[]} extent
     * @param {boolean} [viewProj=false]
     * @returns {Array<module:ol/Feature~Feature>}
     */
    getFeaturesInExtent: function getFeaturesInExtent(extent) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      var features = [];
      this.forEachFeatureIntersectingExtent(extent, function (feature) {
        features.push(feature);
      }, viewProj);
      return features;
    },

    /**
     * @param {number[]} coordinate
     * @param {function} [filter]
     * @param {boolean} [viewProj=false]
     * @returns {module:ol/Feature~Feature}
     */
    getClosestFeatureToCoordinate: function getClosestFeatureToCoordinate(coordinate) {
      var filter = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : identity;
      var viewProj = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      coordinate = viewProj ? roundPointCoords(coordinate) : this.pointToViewProj(coordinate);
      var closestFeature;
      var minSquaredDistance = Infinity;
      this.forEachFeature(function (feature) {
        if (!filter(feature)) return;
        var geometry = feature.getGeometry();
        if (!geometry) return;
        var closestPoint = geometry.getClosestPoint(coordinate);
        var prevSquaredDistance = minSquaredDistance;
        minSquaredDistance = Math.pow(closestPoint[0] - coordinate[0], 2) + Math.pow(closestPoint[1] - coordinate[1]);

        if (minSquaredDistance < prevSquaredDistance) {
          closestFeature = feature;
        }
      });
      return closestFeature;
    },

    /**
     * @param {boolean} [viewProj=false]
     * @returns {number[]}
     */
    getFeaturesExtent: function getFeaturesExtent() {
      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      var extent;
      this.forEachFeature(function (feature) {
        var geometry = feature.getGeometry();
        if (!geometry) return;
        var geomExtent = geometry.getExtent();

        if (!extent) {
          extent = geomExtent;
        } else {
          extent = [geomExtent[0] < extent[0] ? geomExtent[0] : extent[0], geomExtent[1] < extent[1] ? geomExtent[1] : extent[1], geomExtent[2] > extent[2] ? geomExtent[2] : extent[2], geomExtent[3] > extent[3] ? geomExtent[3] : extent[3]];
        }
      });
      return viewProj ? extent : this.extentToDataProj(extent);
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    featuresCollectionIdentChanged: function featuresCollectionIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.$featuresCollection) {
        this.setInstance(value, this.$featuresCollection);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    }
  }
};

function defineServices$d() {
  Object.defineProperties(this, {
    $featuresCollection: {
      enumerable: true,
      get: this.getFeaturesCollection
    }
  });
}

function subscribeToCollectionEvents$3() {
  var _this2 = this;

  var adds = fromOlEvent(this.$featuresCollection, CollectionEventType.ADD).pipe(map(function (evt) {
    return _objectSpread$j(_objectSpread$j({}, evt), {}, {
      element: _this2.initializeFeature(evt.element)
    });
  }), tap(function (_ref) {
    var element = _ref.element;
    var uid = getUid(element);
    var propChanges = fromOlEvent(element, ObjectEventType.PROPERTYCHANGE);
    var changes = fromOlEvent(element, EventType.CHANGE);
    var events = merge(propChanges, changes).pipe(distinctUntilChanged(isEqual));
    _this2._featureSubs[uid] = _this2.subscribeTo(events, _this2.debounceChanged.bind(_this2));
  }));
  var removes = fromOlEvent(this.$featuresCollection, CollectionEventType.REMOVE).pipe(tap(function (_ref2) {
    var element = _ref2.element;
    var uid = getUid(element);

    if (_this2._featureSubs[uid]) {
      _this2.unsubscribe(_this2._featureSubs[uid]);

      delete _this2._featureSubs[uid];
    }
  }));
  var events = merge(adds, removes).pipe(map(function (_ref3) {
    var type = _ref3.type,
        element = _ref3.element;
    var viewProj = _this2.resolvedViewProjection;
    var dataProj = _this2.resolvedDataProjection;
    return {
      type: type,
      feature: element,

      get json() {
        if (!this._json) {
          this._json = writeGeoJsonFeature(this.feature, viewProj, dataProj, COORD_PRECISION);
        }

        return this._json;
      }

    };
  }), bufferDebounceTime(FRAME_TIME));
  this.subscribeTo(events, /*#__PURE__*/function () {
    var _ref4 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee(events) {
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _context.next = 2;
              return _this2.debounceChanged();

            case 2:
              forEach(events, function (evt) {
                _this2.$emit(evt.type + 'feature', evt); // todo remove in v0.13.x


                _this2.$emit(evt.type + ':feature', evt.feature);
              });

            case 3:
            case "end":
              return _context.stop();
          }
        }
      }, _callee);
    }));

    return function (_x) {
      return _ref4.apply(this, arguments);
    };
  }());
}

/**
 * @typedef {module:ol/style/Fill~Fill|Object|undefined} FillStyleLike
 */

/**
 * @typedef {Object} FillStyleTarget
 * @property {function(): module:ol/style/Fill~Fill|undefined} getFill
 * @property {function(module:ol/style/Fill~Fill|undefined): void} setFill
 */

/**
 * Fill style container.
 */

var fillStyleContainer = {
  created: function created() {
    this._fill = undefined;
    this._fillVm = undefined;
    defineServices$c.call(this);
  },
  methods: {
    /**
     * @returns {{readonly fillStyleContainer: Object}}
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get fillStyleContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {FillStyleTarget}
     */
    getFillStyleTarget: function getFillStyleTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getFillStyleTarget()"));
    },

    /**
     * @returns {module:ol/style/Fill~Fill|undefined}
     */
    getFill: function getFill() {
      var _this$getFillStyleTar;

      return coalesce((_this$getFillStyleTar = this.getFillStyleTarget()) === null || _this$getFillStyleTar === void 0 ? void 0 : _this$getFillStyleTar.getFill(), this._fill);
    },

    /**
     * @return {Object}
     */
    getFillVm: function getFillVm() {
      return this._fillVm;
    },

    /**
     * @param {module:ol/style/Fill~Fill|undefined} fill
     */
    setFill: function setFill(fill) {
      var _fill;

      fill = ((_fill = fill) === null || _fill === void 0 ? void 0 : _fill.$style) || fill;
      fill || (fill = undefined);
      assert(!fill || fill instanceof Fill, 'Invalid fill style');

      if (fill !== this._fill) {
        var _fill2;

        this._fill = fill;
        this._fillVm = ((_fill2 = fill) === null || _fill2 === void 0 ? void 0 : _fill2.vm) && fill.vm[0];
        this.scheduleRefresh();
      }

      var fillTarget = this.getFillStyleTarget();

      if (fillTarget && fill !== fillTarget.getFill()) {
        fillTarget.setFill(fill);
        this.scheduleRefresh();
      }
    }
  }
};

function defineServices$c() {
  Object.defineProperties(this, {
    $fill: {
      enumerable: true,
      get: this.getFill
    },
    $fillVm: {
      enumerable: true,
      get: this.getFillVm
    }
  });
}

function ownKeys$i(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$i(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$i(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$i(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base geometry mixin.
 */

var geometry = {
  mixins: [stubVNode, projTransforms, olCmp, waitForMap],
  stubVNode: {
    empty: function empty() {
      return this.vmId;
    }
  },
  data: function data() {
    return {
      viewProjection: EPSG_3857,
      dataProjection: EPSG_3857,
      extentViewProj: undefined
    };
  },
  computed: {
    type: function type() {
      if (!(this.rev && this.$geometry)) return;
      return this.getType();
    },
    extentDataProj: function extentDataProj() {
      return this.extentToDataProj(this.extentViewProj);
    }
  },
  watch: _objectSpread$i({
    rev: function rev() {
      if (!this.$geometry) return;

      if (!isEqual(this.extentViewProj, this.$geometry.getExtent())) {
        this.extentViewProj = this.$geometry.getExtent().slice();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['extentDataProj'], ['extentDataProj'])),
  created: function created() {
    defineServices$b.call(this);
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeInit: function beforeInit() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all([olCmp.methods.beforeInit.call(_this), waitForMap.methods.beforeInit.call(_this)]);

              case 2:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {Promise<module:ol/geom/Geometry~Geometry>}
     * @protected
     */
    createOlObject: function createOlObject() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.t0 = initializeGeometry;
                _context2.next = 3;
                return _this2.createGeometry();

              case 3:
                _context2.t1 = _context2.sent;
                _context2.t2 = _this2.currentId;
                return _context2.abrupt("return", (0, _context2.t0)(_context2.t1, _context2.t2));

              case 6:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @return {module:ol/geom/Geometry~Geometry|Promise<module:ol/geom/Geometry~Geometry>}
     * @protected
     * @abstract
     */
    createGeometry: function createGeometry() {
      throw new Error("".concat(this.vmName, " not implemented method: createGeometry()"));
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        var _this3$$geometryConta;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                (_this3$$geometryConta = _this3.$geometryContainer) === null || _this3$$geometryConta === void 0 ? void 0 : _this3$$geometryConta.setGeometry(_this3);
                return _context3.abrupt("return", olCmp.methods.mount.call(_this3));

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this4 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
        var _this4$$geometryConta;

        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                if (((_this4$$geometryConta = _this4.$geometryContainer) === null || _this4$$geometryConta === void 0 ? void 0 : _this4$$geometryConta.getGeometryVm()) === _this4) {
                  _this4.$geometryContainer.setGeometry(null);
                }

                return _context4.abrupt("return", olCmp.methods.unmount.call(_this4));

              case 2:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4);
      }))();
    },

    /**
     * @return {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(olCmp.methods.getServices.call(this), {
        get geometryVm() {
          return vm;
        }

      });
    },

    /**
     * @returns {void}
     */
    subscribeAll: function subscribeAll() {
      olCmp.methods.subscribeAll.call(this);
      subscribeToGeometryEvents.call(this);
    },

    /**
     * @returns {string|number}
     */
    getIdInternal: function getIdInternal() {
      return getGeometryId(this.$geometry);
    },

    /**
     * @param {string|number} id
     */
    setIdInternal: function setIdInternal(id) {
      if (id === this.getIdInternal()) return;
      setGeometryId(this.$geometry, id);
    },

    /**
     * @return {Promise<module:ol/geom/Geometry~Geometry>}
     */
    resolveGeometry: olCmp.methods.resolveOlObject,

    /**
     * @returns {string}
     */
    getType: function getType() {
      var _this$$geometry;

      return coalesce((_this$$geometry = this.$geometry) === null || _this$$geometry === void 0 ? void 0 : _this$$geometry.getType(), this.type);
    },

    /**
     * @param {boolean} [viewProj=false]
     * @returns {number[]}
     */
    getExtent: function getExtent() {
      var _this$$geometry2;

      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      var extent = coalesce((_this$$geometry2 = this.$geometry) === null || _this$$geometry2 === void 0 ? void 0 : _this$$geometry2.getExtent(), this.extentViewProj);
      return viewProj ? roundExtent(extent) : this.extentToDataProj(extent);
    },

    /**
     * @param {number[]} point
     * @param {boolean} [viewProj=false]
     * @returns {Promise<number[]>}
     */
    getClosestPoint: function getClosestPoint(point) {
      var _arguments = arguments,
          _this5 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
        var viewProj, closestPoint;
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                viewProj = _arguments.length > 1 && _arguments[1] !== undefined ? _arguments[1] : false;
                point = viewProj ? roundPointCoords(point) : _this5.pointToViewProj(point);
                _context5.next = 4;
                return _this5.resolveGeometry();

              case 4:
                closestPoint = _context5.sent.getClosestPoint(point);
                return _context5.abrupt("return", viewProj ? roundPointCoords(closestPoint) : _this5.pointToDataProj(closestPoint));

              case 6:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5);
      }))();
    },

    /**
     * @param {number[]} coordinate
     * @param {boolean} [viewProj=false]
     * @returns {Promise<boolean>}
     */
    intersectsCoordinate: function intersectsCoordinate(coordinate) {
      var _arguments2 = arguments,
          _this6 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee6() {
        var viewProj;
        return _regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                viewProj = _arguments2.length > 1 && _arguments2[1] !== undefined ? _arguments2[1] : false;
                coordinate = viewProj ? roundPointCoords(coordinate) : _this6.pointToViewProj(coordinate);
                _context6.next = 4;
                return _this6.resolveGeometry();

              case 4:
                return _context6.abrupt("return", _context6.sent.intersectsCoordinate(coordinate));

              case 5:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6);
      }))();
    },

    /**
     * @param {number[]} extent
     * @param {boolean} [viewProj=false]
     * @returns {Promise<boolean>}
     */
    intersectsExtent: function intersectsExtent(extent) {
      var _arguments3 = arguments,
          _this7 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee7() {
        var viewProj;
        return _regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                viewProj = _arguments3.length > 1 && _arguments3[1] !== undefined ? _arguments3[1] : false;
                extent = viewProj ? roundExtent(extent) : _this7.extentToViewProj(extent);
                _context7.next = 4;
                return _this7.resolveGeometry();

              case 4:
                return _context7.abrupt("return", _context7.sent.intersectsExtent(extent));

              case 5:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7);
      }))();
    },

    /**
     * @param {number} angle Angle in radians
     * @param {number[]} anchor
     * @param {boolean} [viewProj=false]
     * @returns {Promise<void>}
     */
    rotate: function rotate(angle, anchor) {
      var _arguments4 = arguments,
          _this8 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee8() {
        var viewProj;
        return _regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                viewProj = _arguments4.length > 2 && _arguments4[2] !== undefined ? _arguments4[2] : false;
                anchor = viewProj ? roundPointCoords(anchor) : _this8.pointToViewProj(anchor);
                _context8.next = 4;
                return _this8.resolveGeometry();

              case 4:
                _context8.sent.rotate(angle, anchor);

              case 5:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8);
      }))();
    },

    /**
     * @param {number} sx
     * @param {number} [sy]
     * @param {number[]} [anchor]
     * @param {boolean} [viewProj=false]
     * @returns {Promise<void>}
     */
    scale: function scale(sx, sy, anchor) {
      var _arguments5 = arguments,
          _this9 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee9() {
        var viewProj, _ref, _ref2, _ref3, _ref4;

        return _regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                viewProj = _arguments5.length > 3 && _arguments5[3] !== undefined ? _arguments5[3] : false;
                anchor = viewProj ? roundPointCoords(anchor) : _this9.pointToViewProj(anchor);
                _ref = viewProj ? roundPointCoords([sx, 0]) : _this9.pointToViewProj([sx, 0]);
                _ref2 = _slicedToArray(_ref, 1);
                sx = _ref2[0];
                _ref3 = viewProj ? roundPointCoords([0, sy]) : _this9.pointToViewProj([0, sy]);
                _ref4 = _slicedToArray(_ref3, 2);
                sy = _ref4[1];
                _context9.next = 10;
                return _this9.resolveGeometry();

              case 10:
                _context9.sent.scale(sx, sy, anchor);

              case 11:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9);
      }))();
    },

    /**
     * @param {number} tolerance
     * @returns {Promise<module:ol/geom/Geometry~Geometry>}
     */
    simplify: function simplify(tolerance) {
      var _this10 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee10() {
        return _regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                _context10.next = 2;
                return _this10.resolveGeometry();

              case 2:
                return _context10.abrupt("return", _context10.sent.simplify(tolerance));

              case 3:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10);
      }))();
    },

    /**
     * @param dx
     * @param dy
     * @param {boolean} [viewProj=false]
     * @returns {Promise<*>}
     */
    translate: function translate(dx, dy) {
      var _arguments6 = arguments,
          _this11 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee11() {
        var viewProj, _ref5, _ref6, _ref7, _ref8;

        return _regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                viewProj = _arguments6.length > 2 && _arguments6[2] !== undefined ? _arguments6[2] : false;
                _ref5 = viewProj ? roundPointCoords([dx, 0]) : _this11.pointToViewProj([dx, 0]);
                _ref6 = _slicedToArray(_ref5, 1);
                dx = _ref6[0];
                _ref7 = viewProj ? roundPointCoords([0, dy]) : _this11.pointToViewProj([0, dy]);
                _ref8 = _slicedToArray(_ref7, 2);
                dy = _ref8[1];
                _context11.next = 9;
                return _this11.resolveGeometry();

              case 9:
                return _context11.abrupt("return", _context11.sent.translate(dx, dy));

              case 10:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11);
      }))();
    },

    /**
     * @param {number[]} value
     * @param {number[]} prev
     * @protected
     */
    extentDataProjChanged: function extentDataProjChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:extent', value === null || value === void 0 ? void 0 : value.slice());
    }
  }
};

function defineServices$b() {
  var _this12 = this;

  Object.defineProperties(this, {
    /**
     * @type {module:ol/geom/Geometry~Geometry|undefined}
     */
    $geometry: {
      enumerable: true,
      get: function get() {
        return _this12.$olObject;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $mapVm: {
      enumerable: true,
      get: function get() {
        var _this12$$services;

        return (_this12$$services = _this12.$services) === null || _this12$$services === void 0 ? void 0 : _this12$$services.mapVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $viewVm: {
      enumerable: true,
      get: function get() {
        var _this12$$services2;

        return (_this12$$services2 = _this12.$services) === null || _this12$$services2 === void 0 ? void 0 : _this12$$services2.viewVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $geometryContainer: {
      enumerable: true,
      get: function get() {
        var _this12$$services3;

        return (_this12$$services3 = _this12.$services) === null || _this12$$services3 === void 0 ? void 0 : _this12$$services3.geometryContainer;
      }
    }
  });
}

function subscribeToGeometryEvents() {
  return _subscribeToGeometryEvents.apply(this, arguments);
}

function _subscribeToGeometryEvents() {
  _subscribeToGeometryEvents = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee12() {
    return _regeneratorRuntime.wrap(function _callee12$(_context12) {
      while (1) {
        switch (_context12.prev = _context12.next) {
          case 0:
          case "end":
            return _context12.stop();
        }
      }
    }, _callee12);
  }));
  return _subscribeToGeometryEvents.apply(this, arguments);
}

/**
 * @typedef {Source|Object} SourceLike
 */

/**
 * @typedef {Object} SourceTarget
 * @property {function(): Source|undefined} getSource
 * @property {function(Source|undefined): void} setSource
 */

/**
 * Source container mixin.
 */

var sourceContainer = {
  created: function created() {
    /**
     * @type {module:ol/source/Source~Source|undefined}
     * @private
     */
    this._source = undefined;
    /**
     * @type {Object|undefined}
     * @private
     */

    this._sourceVm = undefined;
    defineServices$a.call(this);
  },
  methods: {
    /**
     * @returns {{readonly sourceContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get sourceContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {SourceTarget|undefined}
     * @protected
     */
    getSourceTarget: function getSourceTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getSourceTarget()"));
    },

    /**
     * @return {module:ol/source/Source~Source|undefined}
     */
    getSource: function getSource() {
      var _this$getSourceTarget;

      return coalesce((_this$getSourceTarget = this.getSourceTarget()) === null || _this$getSourceTarget === void 0 ? void 0 : _this$getSourceTarget.getSource(), this._source);
    },

    /**
     * @return {Object}
     */
    getSourceVm: function getSourceVm() {
      return this._sourceVm;
    },

    /**
     * @param {SourceLike|undefined} source
     */
    setSource: function setSource(source) {
      var _source;

      source = ((_source = source) === null || _source === void 0 ? void 0 : _source.$source) || source;
      source || (source = undefined);
      assert(!source || source instanceof Source, 'Invalid source');
      var sourceTarget = this.getSourceTarget();

      if (sourceTarget && source !== sourceTarget.getSource()) {
        sourceTarget.setSource(source);
        this.scheduleRefresh();
      }

      if (source !== this._source) {
        var _source2;

        this._source = source;
        this._sourceVm = ((_source2 = source) === null || _source2 === void 0 ? void 0 : _source2.vm) && source.vm[0];
        this.scheduleRefresh();
      }
    }
  }
};

function defineServices$a() {
  Object.defineProperties(this, {
    $source: {
      enumerable: true,
      get: this.getSource
    },
    $sourceVm: {
      enumerable: true,
      get: this.getSourceVm
    }
  });
}

function ownKeys$h(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$h(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$h(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$h(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base simple layer mixin.
 */

var layer = {
  mixins: [sourceContainer, baseLayer],
  props: {
    // ol/layer/Layer

    /**
     * @type {function|undefined}
     */
    render: Function,
    // custom

    /**
     * @type {boolean}
     */
    overlay: {
      type: Boolean,
      default: false
    }
  },
  computed: {
    source: function source() {
      if (!(this.rev && this.$source)) return;
      return {
        id: getSourceId(this.$source),
        type: this.$source.constructor.name
      };
    }
  },
  watch: _objectSpread$h({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['source', 'render', 'overlay'], ['source'])),
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (!(_this.overlay && _this.$mapVm)) {
                  _context.next = 3;
                  break;
                }

                _this.setMap(_this.$mapVm);

                return _context.abrupt("return");

              case 3:
                return _context.abrupt("return", baseLayer.methods.mount.call(_this));

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (!_this2.overlay) {
                  _context2.next = 3;
                  break;
                }

                _this2.setMap(null);

                return _context2.abrupt("return");

              case 3:
                return _context2.abrupt("return", baseLayer.methods.unmount.call(_this2));

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      return mergeDescriptors(baseLayer.methods.getServices.call(this), sourceContainer.methods.getServices.call(this));
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      baseLayer.methods.subscribeAll.call(this);
      subscribeToLayerEvents$2.call(this);
    },

    /**
     * @return {module:ol/layer/Base~BaseLayer}
     * @protected
     */
    getSourceTarget: function getSourceTarget() {
      return this.$layer;
    },

    /**
     * @returns {module:ol/renderer/Layer~LayerRenderer}
     */
    getRenderer: function getRenderer() {
      var _this$$layer;

      return (_this$$layer = this.$layer) === null || _this$$layer === void 0 ? void 0 : _this$$layer.getRenderer();
    },

    /**
     * @param {module:ol/Map~Map|Object|undefined} map
     */
    setMap: function setMap(map) {
      var _map, _this$$layer2;

      map = ((_map = map) === null || _map === void 0 ? void 0 : _map.$map) || map;
      (_this$$layer2 = this.$layer) === null || _this$$layer2 === void 0 ? void 0 : _this$$layer2.setMap(map);
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prev
     * @protected
     */
    sourceChanged: function sourceChanged(value, prev) {
      if (value === prev) return;
      this.$emit('update:source', value);
    }
  }
};

function subscribeToLayerEvents$2() {
  return _subscribeToLayerEvents$2.apply(this, arguments);
}

function _subscribeToLayerEvents$2() {
  _subscribeToLayerEvents$2 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
    var _this3 = this;

    var setterKey, sourceChanges, renderEvents;
    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            setterKey = addPrefix('set');
            sourceChanges = fromOlChangeEvent(this.$layer, 'source', true).pipe(tap(function (_ref) {
              var source = _ref.value;

              if (_this3._sourceSubs) {
                _this3.unsubscribe(_this3._sourceSubs);
              }

              if (source) {
                _this3._sourceSubs = _this3.subscribeTo(fromOlChangeEvent(source, 'id', true), _this3.scheduleRefresh.bind(_this3));
              }
            }), map(function (evt) {
              return _objectSpread$h(_objectSpread$h({}, evt), {}, {
                setter: _this3[setterKey(evt.prop)]
              });
            }));
            this.subscribeTo(sourceChanges, function (_ref2) {
              var setter = _ref2.setter,
                  value = _ref2.value;
              return setter(value);
            });
            renderEvents = fromOlEvent(this.$layer, [RenderEventType.PRERENDER, RenderEventType.POSTRENDER]);
            this.subscribeTo(renderEvents, function (evt) {
              return _this3.$emit(evt.type, evt);
            });

          case 5:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));
  return _subscribeToLayerEvents$2.apply(this, arguments);
}

var imageLayer = {
  mixins: [layer]
};

function ownKeys$g(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$g(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$g(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$g(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var validateAttributions = /*#__PURE__*/or(isString, isFunction, function (value) {
  return isArray(value) && value.every(isString);
});
/**
 * Base source mixin.
 */

var source = {
  mixins: [stubVNode, projTransforms, olCmp, waitForMap],
  stubVNode: {
    empty: function empty() {
      return this.vmId;
    }
  },
  props: {
    // ol/source/Source

    /**
     * @type {string|string[]|undefined}
     */
    attributions: {
      type: [String, Array, Function],
      validator: validateAttributions
    },

    /**
     * @type {boolean}
     */
    attributionsCollapsible: {
      type: Boolean,
      default: true
    },

    /**
     * @type {string|undefined}
     */
    projection: {
      type: String,
      validator: function validator(value) {
        return !!get(value);
      }
    },

    /**
     * @type {boolean}
     */
    wrapX: {
      type: Boolean,
      default: true
    },

    /**
     * @type {string|undefined}
     */
    state: {
      type: String,
      default: SourceState.READY,
      validator: function validator(value) {
        return Object.values(SourceState).includes(value);
      }
    }
  },
  data: function data() {
    return {
      viewProjection: EPSG_3857,
      dataProjection: EPSG_3857,
      currentAttributions: adaptAttributions(this.attributions),
      currentState: this.state
    };
  },
  computed: {
    inputAttributions: function inputAttributions() {
      return adaptAttributions(this.attributions);
    },
    currentResolutions: function currentResolutions() {
      return this.rev ? this.getResolutions() : [];
    },
    resolvedDataProjection: function resolvedDataProjection() {
      var _this$$options;

      return coalesce(this.projection, this.dataProjection, // may or may not be present
      (_this$$options = this.$options) === null || _this$$options === void 0 ? void 0 : _this$$options.dataProjection, // may or may not be present
      this.resolvedViewProjection);
    }
  },
  watch: _objectSpread$g({
    rev: function rev() {
      if (!this.$source) return;

      if (this.currentAttributions !== this.$source.getAttributions()) {
        this.currentAttributions = this.$source.getAttributions();
      }

      if (this.currentState !== this.$source.getState()) {
        this.currentState = this.$source.getState();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['inputAttributions', 'currentAttributions', 'state', 'currentState', 'attributionsCollapsible', 'projection', 'wrapX', 'currentResolutions'], ['inputAttributions', 'currentAttributions', 'currentResolutions'])),
  created: function created() {
    defineServices$9.call(this);
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeInit: function beforeInit() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all([olCmp.methods.beforeInit.call(_this), waitForMap.methods.beforeInit.call(_this)]);

              case 2:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {Promise<module:ol/source/Source~Source>}
     * @protected
     */
    createOlObject: function createOlObject() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.t0 = initializeSource;
                _context2.next = 3;
                return _this2.createSource();

              case 3:
                _context2.t1 = _context2.sent;
                _context2.t2 = _this2.currentId;
                return _context2.abrupt("return", (0, _context2.t0)(_context2.t1, _context2.t2));

              case 6:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @return {module:ol/source/Source~Source|Promise<module:ol/source/Source~Source>}
     * @protected
     * @abstract
     */
    createSource: function createSource() {
      throw new Error("".concat(this.vmName, " not implemented method: createSource()"));
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        var _this3$$sourceContain;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                (_this3$$sourceContain = _this3.$sourceContainer) === null || _this3$$sourceContain === void 0 ? void 0 : _this3$$sourceContain.setSource(_this3);
                return _context3.abrupt("return", olCmp.methods.mount.call(_this3));

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this4 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
        var _this4$$sourceContain;

        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                if (((_this4$$sourceContain = _this4.$sourceContainer) === null || _this4$$sourceContain === void 0 ? void 0 : _this4$$sourceContain.getSourceVm()) === _this4) {
                  _this4.$sourceContainer.setSource(null);
                }

                return _context4.abrupt("return", olCmp.methods.unmount.call(_this4));

              case 2:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4);
      }))();
    },

    /**
     * @returns {Promise<void>}
     */
    refresh: function refresh() {
      var _this5 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
        var source;
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return _this5.resolveSource();

              case 2:
                source = _context5.sent;
                return _context5.abrupt("return", new Promise(function (resolve) {
                  source.once('change', function () {
                    return resolve();
                  });
                  source.refresh();
                }));

              case 4:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5);
      }))();
    },

    /**
     * @return {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(olCmp.methods.getServices.call(this), {
        get sourceVm() {
          return vm;
        }

      });
    },

    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      olCmp.methods.subscribeAll.call(this);
      subscribeToSourceEvents$3.call(this);
    },

    /**
     * @returns {string|number}
     * @protected
     */
    getIdInternal: function getIdInternal() {
      return getSourceId(this.$source);
    },

    /**
     * @param {string|number} id
     * @protected
     */
    setIdInternal: function setIdInternal(id) {
      if (id === this.getIdInternal()) return;
      setSourceId(this.$source, id);
    },

    /**
     * @return {Promise<module:ol/source/Source~Source>}
     */
    resolveSource: olCmp.methods.resolveOlObject,

    /**
     * @returns {string|string[]|undefined}
     */
    getAttributions: function getAttributions() {
      var _this$$source;

      return coalesce((_this$$source = this.$source) === null || _this$$source === void 0 ? void 0 : _this$$source.getAttributions(), this.currentAttributions);
    },

    /**
     * @param {string} attributions
     */
    setAttributions: function setAttributions(attributions) {
      assert(!attributions || validateAttributions(attributions), 'Invalid attributions');
      attributions = adaptAttributions(attributions);

      if (!isEqual(attributions, this.currentAttributions)) {
        this.currentAttributions = attributions;
      }

      if (this.$source && !isEqual(attributions, this.$source.getAttributions())) {
        this.$source.setAttributions(attributions);
      }
    },

    /**
     * @returns {boolean}
     */
    getAttributionsCollapsible: function getAttributionsCollapsible() {
      var _this$$source2;

      return coalesce((_this$$source2 = this.$source) === null || _this$$source2 === void 0 ? void 0 : _this$$source2.getAttributionsCollapsible(), this.attributionsCollapsible);
    },

    /**
     * @returns {module:ol/proj/Projection~Projection}
     */
    getProjection: function getProjection() {
      var _this$$source3;

      return coalesce((_this$$source3 = this.$source) === null || _this$$source3 === void 0 ? void 0 : _this$$source3.getProjection(), get(this.resolvedDataProjection));
    },

    /**
     * @returns {string}
     */
    getState: function getState() {
      var _this$$source4;

      return coalesce((_this$$source4 = this.$source) === null || _this$$source4 === void 0 ? void 0 : _this$$source4.getState(), this.currentState);
    },

    /**
     * @returns {boolean}
     */
    getWrapX: function getWrapX() {
      var _this$$source5;

      return coalesce((_this$$source5 = this.$source) === null || _this$$source5 === void 0 ? void 0 : _this$$source5.getWrapX(), this.wrapX);
    },

    /**
     * @returns {number[]}
     */
    getResolutions: function getResolutions() {
      try {
        var _this$$source6;

        // can be not implemented in source class
        return coalesce((_this$$source6 = this.$source) === null || _this$$source6 === void 0 ? void 0 : _this$$source6.getResolutions(), []);
      } catch (err) {
        return [];
      }
    },

    /**
     * @param {string|string[]|Function|undefined} value
     * @protected
     */
    inputAttributionsChanged: function inputAttributionsChanged(value) {
      this.setAttributions(value);
    },

    /**
     * @param {string|string[]|Function|undefined} value
     * @protected
     */
    currentAttributionsChanged: function currentAttributionsChanged(value) {
      if (isEqual(value, this.inputAttributions)) return;
      this.$emit('update:attributions', value);
    },

    /**
     * @param {string} value
     * @protected
     */
    stateChanged: function stateChanged(value) {
      if (value === this.currentState) return;

      if (process.env.VUELAYERS_DEBUG) {
        this.$logger.log('state changed, scheduling recreate... %O ===> %O', this.currentState, value);
      }

      this.currentState = value;
      return this.scheduleRecreate();
    },

    /**
     * @param {string} value
     * @protected
     */
    currentStateChanged: function currentStateChanged(value) {
      if (value === this.state) return;
      this.$emit('update:state', value);
    },

    /**
     * @param {number[]} value
     * @param {number[]} prev
     * @protected
     */
    currentResolutionsChanged: function currentResolutionsChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:resolutions', value === null || value === void 0 ? void 0 : value.slice());
    }
  }
};

function defineServices$9() {
  var _this6 = this;

  Object.defineProperties(this, {
    /**
     * @type {module:ol/source/Source~Source|undefined}
     */
    $source: {
      enumerable: true,
      get: function get() {
        return _this6.$olObject;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $mapVm: {
      enumerable: true,
      get: function get() {
        var _this6$$services;

        return (_this6$$services = _this6.$services) === null || _this6$$services === void 0 ? void 0 : _this6$$services.mapVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $viewVm: {
      enumerable: true,
      get: function get() {
        var _this6$$services2;

        return (_this6$$services2 = _this6.$services) === null || _this6$$services2 === void 0 ? void 0 : _this6$$services2.viewVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $sourceContainer: {
      enumerable: true,
      get: function get() {
        var _this6$$services3;

        return (_this6$$services3 = _this6.$services) === null || _this6$$services3 === void 0 ? void 0 : _this6$$services3.sourceContainer;
      }
    }
  });
}

function subscribeToSourceEvents$3() {}

function adaptAttributions(attributions) {
  if (!attributions) return;
  if (isFunction(attributions)) return attributions;
  return function () {
    return isArray(attributions) ? attributions : [attributions];
  };
}

function ownKeys$f(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$f(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$f(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$f(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var ImageSourceEventType = {
  IMAGELOADSTART: 'imageloadstart',
  IMAGELOADEND: 'imageloadend',
  IMAGELOADERROR: 'imageloaderror'
};
/**
 * Base image source mixin.
 */

var imageSource = {
  mixins: [source],
  props: {
    // ol/source/Image

    /**
     * @type {string}
     */
    projection: _objectSpread$f(_objectSpread$f({}, source.props.projection), {}, {
      default: EPSG_3857
    }),

    /**
     * @type {number[]|undefined}
     */
    resolutions: Array
  },
  computed: {
    inputResolutions: function inputResolutions() {
      var _this$resolutions;

      return (_this$resolutions = this.resolutions) === null || _this$resolutions === void 0 ? void 0 : _this$resolutions.slice();
    }
  },
  watch: _objectSpread$f({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['inputResolutions'], ['inputResolutions'])),
  methods: {
    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      source.methods.subscribeAll.call(this);
      subscribeToSourceEvents$2.call(this);
    },
    attributionsCollapsibleChanged: noop,
    wrapXChanged: noop
  }
};

function subscribeToSourceEvents$2() {
  return _subscribeToSourceEvents$1.apply(this, arguments);
}

function _subscribeToSourceEvents$1() {
  _subscribeToSourceEvents$1 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
    var _this = this;

    var events;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            events = fromOlEvent(this.$source, [ImageSourceEventType.IMAGELOADSTART, ImageSourceEventType.IMAGELOADEND, ImageSourceEventType.IMAGELOADERROR]);
            this.subscribeTo(events, function (evt) {
              return _this.$emit(evt.type, evt);
            });

          case 2:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));
  return _subscribeToSourceEvents$1.apply(this, arguments);
}

/**
 * Basic style mixin.
 */

var style = {
  mixins: [stubVNode, olCmp],
  stubVNode: {
    empty: function empty() {
      return this.vmId;
    }
  },
  created: function created() {
    defineServices$8.call(this);
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeInit: function beforeInit() {
      return olCmp.methods.beforeInit.call(this);
    },

    /**
     * @return {OlStyle|Promise<OlStyle>}
     * @protected
     */
    createOlObject: function createOlObject() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.t0 = initializeStyle;
                _context.next = 3;
                return _this.createStyle();

              case 3:
                _context.t1 = _context.sent;
                _context.t2 = _this.currentId;
                return _context.abrupt("return", (0, _context.t0)(_context.t1, _context.t2));

              case 6:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {OlStyle|Promise<OlStyle>}
     * @protected
     * @abstract
     */
    createStyle: function createStyle() {
      throw new Error("".concat(this.vmName, " not implemented method: createStyle()"));
    },

    /**
     * @return {Promise<void>}
     */
    mount: function mount() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return olCmp.methods.mount.call(_this2);

              case 2:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    unmount: function unmount() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return olCmp.methods.unmount.call(_this3);

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    remount: function remount() {
      var _this4 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.next = 2;
                return olCmp.methods.remount.call(_this4);

              case 2:
                _context4.next = 4;
                return _this4.refresh();

              case 4:
                if (!_this4.$mapVm) {
                  _context4.next = 7;
                  break;
                }

                _context4.next = 7;
                return _this4.$mapVm.render();

              case 7:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    refresh: function refresh() {
      var _this5 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return olCmp.methods.refresh.call(_this5);

              case 2:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5);
      }))();
    },

    /**
     * @return {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(olCmp.methods.getServices.call(this), {
        get styleVm() {
          return vm;
        }

      });
    },

    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      olCmp.methods.subscribeAll.call(this);
    },

    /**
     * @param {string|number} id
     */
    setId: function setId(id) {
      assert(id != null && id !== '', 'Invalid id');

      if (this.currentId !== id) {
        this.currentId = id;
        this.scheduleRefresh();
      }

      this.$olObject && this.setIdInternal(id);
    },

    /**
     * @returns {string|number}
     */
    getIdInternal: function getIdInternal() {
      return getStyleId(this.$style);
    },

    /**
     * @param {string|number} id
     * @returns {void}
     */
    setIdInternal: function setIdInternal(id) {
      if (id === this.getIdInternal()) return;
      setStyleId(this.$style, id);
      this.scheduleRefresh();
    },

    /**
     * @return {Promise<OlStyle>}
     */
    resolveStyle: olCmp.methods.resolveOlObject,

    /**
     * @protected
     */
    syncNonObservable: function syncNonObservable() {
      olCmp.methods.syncNonObservable.call(this);
    }
  }
};

function defineServices$8() {
  var _this6 = this;

  Object.defineProperties(this, {
    /**
     * @type {OlStyle|undefined}
     */
    $style: {
      enumerable: true,
      get: function get() {
        return _this6.$olObject;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $styleContainer: {
      enumerable: true,
      get: function get() {
        var _this6$$services;

        return (_this6$$services = _this6.$services) === null || _this6$$services === void 0 ? void 0 : _this6$$services.styleContainer;
      }
    }
  });
}

function ownKeys$e(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$e(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$e(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$e(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var imageStyle = {
  mixins: [style],
  stubVNode: {
    empty: false,
    attrs: function attrs() {
      return {
        id: this.vmId,
        class: this.vmClass
      };
    }
  },
  props: {
    // ol/style/Image

    /**
     * @type {number}
     */
    opacity: {
      type: Number,
      default: 1
    },

    /**
     * @type {boolean}
     */
    rotateWithView: Boolean,

    /**
     * @type {number}
     */
    rotation: {
      type: Number,
      default: 0
    },

    /**
     * @type {number}
     */
    scale: {
      type: Number,
      default: 1
    },
    displacement: {
      type: Array,
      default: function _default() {
        return [0, 0];
      }
    }
  },
  data: function data() {
    return {
      currentOpacity: this.opacity,
      currentRotateWithView: this.rotateWithView,
      currentRotation: this.rotation,
      currentScale: this.scale
    };
  },
  computed: {
    inputDisplacement: function inputDisplacement() {
      var _this$displacement;

      return (_this$displacement = this.displacement) === null || _this$displacement === void 0 ? void 0 : _this$displacement.slice();
    }
  },
  watch: _objectSpread$e({
    rev: function rev() {
      if (!this.$style) return;
      this.setOpacity(this.getOpacity());
      this.setRotateWithView(this.getRotateWithView());
      this.setRotation(this.getRotation());
      this.setScale(this.getScale());
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['opacity', 'currentOpacity', 'rotateWithView', 'currentRotateWithView', 'rotation', 'currentRotation', 'scale', 'currentScale', 'inputDisplacement'], ['inputDisplacement'])),
  created: function created() {
    defineServices$7.call(this);
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        var _this$$imageStyleCont;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                (_this$$imageStyleCont = _this.$imageStyleContainer) === null || _this$$imageStyleCont === void 0 ? void 0 : _this$$imageStyleCont.setImage(_this);
                return _context.abrupt("return", style.methods.mount.call(_this));

              case 2:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        var _this2$$imageStyleCon;

        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (!(((_this2$$imageStyleCon = _this2.$imageStyleContainer) === null || _this2$$imageStyleCon === void 0 ? void 0 : _this2$$imageStyleCon.getImageVm()) === _this2)) {
                  _context2.next = 3;
                  break;
                }

                _context2.next = 3;
                return _this2.$imageStyleContainer.setImage(null);

              case 3:
                return _context2.abrupt("return", style.methods.unmount.call(_this2));

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @return {Promise<void>}
     */
    refresh: function refresh() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        var _this3$$imageStyleCon;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return Promise.all([style.methods.refresh.call(_this3), (_this3$$imageStyleCon = _this3.$imageStyleContainer) === null || _this3$$imageStyleCon === void 0 ? void 0 : _this3$$imageStyleCon.refresh()]);

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @protected
     */
    syncNonObservable: function syncNonObservable() {
      style.methods.syncNonObservable.call(this);
      this.setOpacity(this.getOpacity());
      this.setRotateWithView(this.getRotateWithView());
      this.setRotation(this.getRotation());
      this.setScale(this.getScale());
    },

    /**
     * @returns {number}
     */
    getOpacity: function getOpacity() {
      var _this$$style;

      return coalesce((_this$$style = this.$style) === null || _this$$style === void 0 ? void 0 : _this$$style.getOpacity(), this.currentOpacity);
    },

    /**
     * @param {number} opacity
     */
    setOpacity: function setOpacity(opacity) {
      if (opacity !== this.currentOpacity) {
        this.currentOpacity = opacity;
        this.scheduleRefresh();
      }

      if (this.$style && opacity !== this.$style.getOpacity()) {
        this.$style.setOpacity(opacity);
        this.scheduleRefresh();
      }
    },

    /**
     * @returns {boolean}
     */
    getRotateWithView: function getRotateWithView() {
      var _this$$style2;

      return coalesce((_this$$style2 = this.$style) === null || _this$$style2 === void 0 ? void 0 : _this$$style2.getRotateWithView(), this.currentRotateWithView);
    },

    /**
     * @param {boolean} rotateWithView
     */
    setRotateWithView: function setRotateWithView(rotateWithView) {
      if (rotateWithView !== this.currentRotateWithView) {
        this.currentRotateWithView = rotateWithView;
        this.scheduleRefresh();
      }

      if (this.$style && rotateWithView !== this.$style.getRotateWithView()) {
        this.$style.setRotateWithView(rotateWithView);
        this.scheduleRefresh();
      }
    },

    /**
     * @returns {number}
     */
    getRotation: function getRotation() {
      var _this$$style3;

      return coalesce((_this$$style3 = this.$style) === null || _this$$style3 === void 0 ? void 0 : _this$$style3.getRotation(), this.currentRotation);
    },

    /**
     * @param {number} rotation
     */
    setRotation: function setRotation(rotation) {
      if (rotation !== this.currentRotation) {
        this.currentRotation = rotation;
        this.scheduleRefresh();
      }

      if (this.$style && rotation !== this.$style.getRotation()) {
        this.$style.setRotation(rotation);
        this.scheduleRefresh();
      }
    },

    /**
     * @returns {number}
     */
    getScale: function getScale() {
      var _this$$style4;

      return coalesce((_this$$style4 = this.$style) === null || _this$$style4 === void 0 ? void 0 : _this$$style4.getScale(), this.currentScale);
    },

    /**
     * @param {number} scale
     */
    setScale: function setScale(scale) {
      if (scale !== this.currentScale) {
        this.currentScale = scale;
        this.scheduleRefresh();
      }

      if (this.$style && scale !== this.$style.getScale()) {
        this.$style.setScale(scale);
        this.scheduleRefresh();
      }
    },

    /**
     * @return {number[]}
     */
    getDisplacement: function getDisplacement() {
      var _this$$style5;

      return coalesce((_this$$style5 = this.$style) === null || _this$$style5 === void 0 ? void 0 : _this$$style5.getDisplacement(), this.inputDisplacement);
    },

    /**
     * @param {number} value
     * @protected
     */
    opacityChanged: function opacityChanged(value) {
      this.setOpacity(value);
    },

    /**
     * @param {number} value
     * @protected
     */
    currentOpacityChanged: function currentOpacityChanged(value) {
      if (value === this.opacity) return;
      this.$emit('update:opacity', value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    rotateWithViewChanged: function rotateWithViewChanged(value) {
      this.setRotateWithView(value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    currentRotateWithViewChanged: function currentRotateWithViewChanged(value) {
      if (value === this.rotateWithView) return;
      this.$emit('update:rotateWithView', value);
    },

    /**
     * @param {number} value
     * @protected
     */
    rotationChanged: function rotationChanged(value) {
      this.setRotation(value);
    },

    /**
     * @param {number} value
     * @protected
     */
    currentRotationChanged: function currentRotationChanged(value) {
      if (value === this.rotation) return;
      this.$emit('update:rotation', value);
    },

    /**
     * @param {number} value
     * @protected
     */
    scaleChanged: function scaleChanged(value) {
      this.setScale(value);
    },

    /**
     * @param {number} value
     * @protected
     */
    currentScaleChanged: function currentScaleChanged(value) {
      if (value === this.scale) return;
      this.$emit('update:scale', value);
    }
  }
};

function defineServices$7() {
  var _this4 = this;

  Object.defineProperties(this, {
    /**
     * @type {Object|undefined}
     */
    $imageStyleContainer: {
      enumerable: true,
      get: function get() {
        var _this4$$services;

        return (_this4$$services = _this4.$services) === null || _this4$$services === void 0 ? void 0 : _this4$$services.imageStyleContainer;
      }
    }
  });
}

/**
 * @typedef {module:ol/style/Image~ImageStyle|Object|undefined} ImageStyleLike
 */

/**
 * @typedef {Object} ImageStyleTarget
 * @property {function(): module:ol/style/Image~ImageStyle|undefined} getImage
 * @property {function(module:ol/style/Image~ImageStyle|undefined): void} setImage
 */

/**
 * Image style container.
 */

var imageStyleContainer = {
  created: function created() {
    this._image = undefined;
    this._imageVm = undefined;
    defineServices$6.call(this);
  },
  methods: {
    /**
     * @returns {{readonly imageStyleContainer: Object}}
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get imageStyleContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {ImageStyleTarget|undefined}
     */
    getImageStyleTarget: function getImageStyleTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getImageStyleTarget()"));
    },

    /**
     * @returns {module:ol/style/Image~ImageStyle|undefined}
     */
    getImage: function getImage() {
      var _this$getImageStyleTa;

      return coalesce((_this$getImageStyleTa = this.getImageStyleTarget()) === null || _this$getImageStyleTa === void 0 ? void 0 : _this$getImageStyleTa.getImage(), this._image);
    },

    /**
     * @return {Object}
     */
    getImageVm: function getImageVm() {
      return this._imageVm;
    },

    /**
     * @param {module:ol/style/Image~ImageStyle|undefined} image
     */
    setImage: function setImage(image) {
      var _image;

      image = ((_image = image) === null || _image === void 0 ? void 0 : _image.$style) || image;
      image || (image = undefined);
      assert(!image || image instanceof Image, 'Invalid image style');
      var imageTarget = this.getImageStyleTarget();

      if (imageTarget && image !== imageTarget.getImage()) {
        imageTarget.setImage(image);
        this.scheduleRefresh();
      }

      if (image !== this._image) {
        var _image2;

        this._image = image;
        this._imageVm = ((_image2 = image) === null || _image2 === void 0 ? void 0 : _image2.vm) && image.vm[0];
        this.scheduleRefresh();
      }
    }
  }
};

function defineServices$6() {
  Object.defineProperties(this, {
    $image: {
      enumerable: true,
      get: this.getImage
    },
    $imageVm: {
      enumerable: true,
      get: this.getImageVm
    }
  });
}

function ownKeys$d(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$d(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$d(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$d(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base interaction mixin.
 */

var interaction = {
  mixins: [stubVNode, projTransforms, olCmp, waitForMap],
  stubVNode: {
    empty: function empty() {
      return this.vmId;
    }
  },
  props: {
    /**
     * @type {boolean}
     */
    active: {
      type: Boolean,
      default: true
    },

    /**
     * Priority of interactions in the event handling stream.
     * The higher the value, the sooner it will handle map event.
     * @type {number}
     */
    priority: {
      type: Number,
      default: 0
    }
  },
  data: function data() {
    return {
      viewProjection: EPSG_3857,
      dataProjection: EPSG_3857,
      currentActive: this.active,
      currentPriority: this.priority,
      interacting: false
    };
  },
  watch: _objectSpread$d({
    rev: function rev() {
      if (!this.$interaction) return;

      if (this.currentActive !== this.$interaction.getActive()) {
        this.currentActive = this.$interaction.getActive();
      }

      if (this.currentPriority !== getInteractionPriority(this.$interaction)) {
        this.currentPriority = getInteractionPriority(this.$interaction);
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['active', 'currentActive', 'priority', 'currentPriority'])),
  created: function created() {
    defineServices$5.call(this);
  },
  methods: {
    /**
     * @return {Promise<void>}
     * @protected
     */
    beforeInit: function beforeInit() {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all([olCmp.methods.beforeInit.call(_this), waitForMap.methods.beforeInit.call(_this)]);

              case 2:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @return {Promise<module:ol/interaction/Interaction~Interaction>}
     * @protected
     */
    createOlObject: function createOlObject() {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        var interaction;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.t0 = initializeInteraction;
                _context2.next = 3;
                return _this2.createInteraction();

              case 3:
                _context2.t1 = _context2.sent;
                _context2.t2 = _this2.currentId;
                _context2.t3 = _this2.currentPriority;
                interaction = (0, _context2.t0)(_context2.t1, _context2.t2, _context2.t3);
                interaction.setActive(_this2.currentActive);
                interaction.set('interacting', _this2.interacting);
                return _context2.abrupt("return", interaction);

              case 10:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @return {module:ol/interaction/Interaction~Interaction|Promise<module:ol/interaction/Interaction~Interaction>}
     * @protected
     * @abstract
     */
    createInteraction: function createInteraction() {
      throw new Error("".concat(this.vmName, " not implemented method: createInteraction()"));
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    mount: function mount() {
      var _this3 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
        var _this3$$interactionsC;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                (_this3$$interactionsC = _this3.$interactionsContainer) === null || _this3$$interactionsC === void 0 ? void 0 : _this3$$interactionsC.addInteraction(_this3);
                return _context3.abrupt("return", olCmp.methods.mount.call(_this3));

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3);
      }))();
    },

    /**
     * @return {Promise<void>}
     * @protected
     */
    unmount: function unmount() {
      var _this4 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
        var _this4$$interactionsC;

        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                (_this4$$interactionsC = _this4.$interactionsContainer) === null || _this4$$interactionsC === void 0 ? void 0 : _this4$$interactionsC.removeInteraction(_this4);
                return _context4.abrupt("return", olCmp.methods.unmount.call(_this4));

              case 2:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4);
      }))();
    },

    /**
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(olCmp.methods.getServices.call(this), {
        get interactionVm() {
          return vm;
        }

      });
    },

    /**
     * @returns {void}
     */
    subscribeAll: function subscribeAll() {
      olCmp.methods.subscribeAll.call(this);
      subscribeToInteractionEvents.call(this);
    },

    /**
     * @return {Promise<module:ol/interaction/Interaction~Interaction>}
     */
    resolveInteraction: olCmp.methods.resolveOlObject,

    /**
     * @returns {string|number}
     */
    getIdInternal: function getIdInternal() {
      return getInteractionId(this.$interaction);
    },

    /**
     * @param {string|number} id
     * @returns {void}
     */
    setIdInternal: function setIdInternal(id) {
      if (id === this.getIdInternal()) return;
      setInteractionId(this.$interaction, id);
    },

    /**
     * @returns {boolean}
     */
    getActive: function getActive() {
      var _this$$interaction;

      return coalesce((_this$$interaction = this.$interaction) === null || _this$$interaction === void 0 ? void 0 : _this$$interaction.getActive(), this.currentActive);
    },

    /**
     * @param {boolean} active
     */
    setActive: function setActive(active) {
      active = !!active;

      if (active !== this.currentActive) {
        this.currentActive = active;
      }

      if (this.$interaction && active !== this.$interaction.getActive()) {
        this.$interaction.setActive(active);
      }
    },

    /**
     * @returns {number}
     */
    getPriority: function getPriority() {
      return coalesce(this.$interaction && getInteractionPriority(this.$interaction), this.currentPriority);
    },

    /**
     * @param {number} priority
     */
    setPriority: function setPriority(priority) {
      var _this$$interactionsCo;

      assert(isNumber(priority), 'Invalid priority');

      if (priority !== this.currentPriority) {
        this.currentPriority = priority;
      }

      if (this.$interaction && priority !== getInteractionPriority(this.$interaction)) {
        setInteractionPriority(this.$interaction, priority);
      } // eslint-disable-next-line no-unused-expressions


      (_this$$interactionsCo = this.$interactionsContainer) === null || _this$$interactionsCo === void 0 ? void 0 : _this$$interactionsCo.sortInteractions();
    },

    /**
     * @return {module:ol/Map~Map|undefined}
     */
    getMap: function getMap() {
      var _this$$interaction2;

      return coalesce((_this$$interaction2 = this.$interaction) === null || _this$$interaction2 === void 0 ? void 0 : _this$$interaction2.getMap(), this.$map);
    },
    // /**
    //  * @param {module:ol/Map~Map} map
    //  */
    // setMap (map) {
    //   this.$interaction?.setMap(map)
    // },
    setInteracting: function setInteracting(flag) {
      flag = !!flag;

      if (flag !== this.interacting) {
        this.interacting = flag;
      }

      if (this.$interaction && flag !== this.$interaction.get('interacting')) {
        this.$interaction.set('interacting', flag);
      }
    },
    isInteracting: function isInteracting() {
      var _this$$interaction3;

      return coalesce((_this$$interaction3 = this.$interaction) === null || _this$$interaction3 === void 0 ? void 0 : _this$$interaction3.get('interacting'), this.interacting);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    activeChanged: function activeChanged(value) {
      this.setActive(value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    currentActiveChanged: function currentActiveChanged(value) {
      if (value === this.active) return;
      this.$emit('update:active', value);
    },

    /**
     * @param {number} value
     * @protected
     */
    priorityChanged: function priorityChanged(value) {
      this.setPriority(value);
    },

    /**
     * @param {number} value
     * @protected
     */
    currentPriorityChanged: function currentPriorityChanged(value) {
      if (value === this.priority) return;
      this.$emit('update:priority', value);
    }
  }
};

function defineServices$5() {
  var _this5 = this;

  Object.defineProperties(this, {
    /**
     * @type {module:ol/interaction/Interaction~Interaction|undefined}
     */
    $interaction: {
      enumerable: true,
      get: function get() {
        return _this5.$olObject;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $mapVm: {
      enumerable: true,
      get: function get() {
        var _this5$$services;

        return (_this5$$services = _this5.$services) === null || _this5$$services === void 0 ? void 0 : _this5$$services.mapVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $viewVm: {
      enumerable: true,
      get: function get() {
        var _this5$$services2;

        return (_this5$$services2 = _this5.$services) === null || _this5$$services2 === void 0 ? void 0 : _this5$$services2.viewVm;
      }
    },

    /**
     * @type {Object|undefined}
     */
    $interactionsContainer: {
      enumerable: true,
      get: function get() {
        var _this5$$services3;

        return (_this5$$services3 = _this5.$services) === null || _this5$$services3 === void 0 ? void 0 : _this5$$services3.interactionsContainer;
      }
    }
  });
}

function subscribeToInteractionEvents() {
  return _subscribeToInteractionEvents.apply(this, arguments);
}

function _subscribeToInteractionEvents() {
  _subscribeToInteractionEvents = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5() {
    var _this6 = this;

    var setterKey, propChanges;
    return _regeneratorRuntime.wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            setterKey = addPrefix('current');
            propChanges = fromOlChangeEvent(this.$interaction, ['active', 'priority'], true, function (evt) {
              return _objectSpread$d(_objectSpread$d({}, evt), {}, {
                setter: _this6[setterKey(evt.prop)]
              });
            });
            this.subscribeTo(propChanges, function (_ref) {
              var setter = _ref.setter,
                  value = _ref.value;
              return setter(value);
            });

          case 3:
          case "end":
            return _context5.stop();
        }
      }
    }, _callee5, this);
  }));
  return _subscribeToInteractionEvents.apply(this, arguments);
}

function ownKeys$c(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$c(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$c(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$c(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * @typedef {module:ol/interaction/Interaction~Interaction|Object} InteractionLike
 */

/**
 * Interactions container
 */

var interactionsContainer = {
  mixins: [identMap, rxSubs],
  computed: {
    /**
     * @returns {string|undefined}
     */
    interactionsCollectionIdent: function interactionsCollectionIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'interactions_collection');
    }
  },
  watch: _objectSpread$c({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['interactionsCollectionIdent'])),
  created: function created() {
    /**
     * @type {module:ol/Collection~Collection<module:ol/interaction/Interaction~Interaction>}
     * @private
     */
    this._interactionsCollection = this.instanceFactoryCall(this.interactionsCollectionIdent, function () {
      return new Collection();
    });
    this._interactionSubs = {};
    defineServices$4.call(this);
  },
  methods: {
    /**
     * @returns {{readonly interactionsContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get interactionsContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      subscribeToCollectionEvents$2.call(this);
    },

    /**
     * @param {InteractionLike[]|module:ol/Collection~Collection<InteractionLike>} defaultInteractions
     */
    initDefaultInteractions: function initDefaultInteractions(defaultInteractions) {
      var _this = this;

      this.getInteractions().forEach(function (interaction) {
        if (interaction.get('vl_default')) {
          _this.removeInteraction(interaction);
        }
      });
      var interactions;

      if (isArray(defaultInteractions) || defaultInteractions instanceof Collection) {
        interactions = defaultInteractions;
      } else if (defaultInteractions !== false) {
        interactions = defaults$1(isPlainObject(defaultInteractions) ? this.defaultInteractions : undefined);
      }

      if (interactions) {
        interactions.forEach(function (interaction) {
          return interaction.set('vl_default', true);
        });
        this.addInteractions(interactions);
      }
    },

    /**
     * @param {InteractionLike} interaction
     * @return {Interaction}
     */
    initializeInteraction: function initializeInteraction$1(interaction) {
      var _interaction;

      interaction = ((_interaction = interaction) === null || _interaction === void 0 ? void 0 : _interaction.$interaction) || interaction;
      instanceOf(interaction, Interaction);
      return initializeInteraction(interaction);
    },

    /**
     * @param {InteractionLike[]|module:ol/Collection~Collection<InteractionLike>} interactions
     */
    addInteractions: function addInteractions(interactions) {
      forEach(interactions, this.addInteraction.bind(this));
    },

    /**
     * @param {InteractionLike} interaction
     */
    addInteraction: function addInteraction(interaction) {
      interaction = this.initializeInteraction(interaction);
      if (this.getInteractionById(getInteractionId(interaction))) return;
      this.$interactionsCollection.push(interaction);
      this.sortInteractions();
    },

    /**
     * @param {InteractionLike[]|module:ol/Collection~Collection<InteractionLike>} interactions
     */
    removeInteractions: function removeInteractions(interactions) {
      forEach(interactions, this.removeInteraction.bind(this));
    },

    /**
     * @param {InteractionLike} interaction
     */
    removeInteraction: function removeInteraction(interaction) {
      var _interaction2;

      interaction = this.getInteractionById(getInteractionId(((_interaction2 = interaction) === null || _interaction2 === void 0 ? void 0 : _interaction2.$interaction) || interaction));
      if (!interaction) return;
      this.$interactionsCollection.remove(interaction);
      this.sortInteractions();
    },

    /**
     * @return {void}
     */
    clearInteractions: function clearInteractions() {
      this.$interactionsCollection.clear();
    },

    /**
     * @return {module:ol/interaction/Interaction~Interaction[]}
     */
    getInteractions: function getInteractions() {
      return this.$interactionsCollection.getArray().slice();
    },

    /**
     * @return {module:ol/Collection~Collection<module:ol/interaction/Interaction~Interaction>}
     */
    getInteractionsCollection: function getInteractionsCollection() {
      return this._interactionsCollection;
    },

    /**
     * @param {string|number} interactionId
     * @return {module:ol/interaction/Interaction~Interaction|undefined}
     */
    getInteractionById: function getInteractionById(interactionId) {
      return find(this.getInteractions(), function (interaction) {
        return getInteractionId(interaction) === interactionId;
      });
    },

    /**
     * @param {function} [sorter]
     */
    sortInteractions: function sortInteractions(sorter) {
      sorter || (sorter = this.getDefaultInteractionsSorter());
      this.$interactionsCollection.getArray().sort(sorter);
    },

    /**
     * @return {function(): number}
     * @protected
     */
    getDefaultInteractionsSorter: function getDefaultInteractionsSorter() {
      // sort interactions by priority in asc order
      // the higher the priority, the earlier the interaction handles the event
      return function (a, b) {
        var ap = getInteractionPriority(a) || 0;
        var bp = getInteractionPriority(b) || 0;
        return ap === bp ? 0 : ap - bp;
      };
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    interactionsCollectionIdentChanged: function interactionsCollectionIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.$interactionsCollection) {
        this.setInstance(value, this.$interactionsCollection);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    }
  }
};

function defineServices$4() {
  Object.defineProperties(this, {
    $interactionsCollection: {
      enumerable: true,
      get: this.getInteractionsCollection
    }
  });
}

function subscribeToCollectionEvents$2() {
  var _this2 = this;

  var adds = fromOlEvent(this.$interactionsCollection, CollectionEventType.ADD).pipe(map(function (evt) {
    return _objectSpread$c(_objectSpread$c({}, evt), {}, {
      element: _this2.initializeInteraction(evt.element)
    });
  }), tap(function (_ref) {
    var element = _ref.element;
    var uid = getUid(element);
    var propChanges = fromOlChangeEvent(element, 'id', true);
    _this2._interactionSubs[uid] = _this2.subscribeTo(propChanges, _this2.scheduleRefresh.bind(_this2));
  }));
  var removes = fromOlEvent(this.$interactionsCollection, CollectionEventType.REMOVE).pipe(tap(function (_ref2) {
    var element = _ref2.element;
    var uid = getUid(element);

    if (_this2._interactionSubs[uid]) {
      _this2.unsubscribe(_this2._interactionSubs[uid]);

      delete _this2._interactionSubs[uid];
    }
  }));
  var events = merge(adds, removes).pipe(bufferDebounceTime(FRAME_TIME));
  this.subscribeTo(events, /*#__PURE__*/function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee(events) {
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _context.next = 2;
              return _this2.debounceChanged();

            case 2:
              forEach(events, function (_ref4) {
                var type = _ref4.type,
                    element = _ref4.element;

                _this2.$emit(type + 'interaction', element); // todo remove in v0.13.x


                _this2.$emit(type + ':interaction', element);
              });

            case 3:
            case "end":
              return _context.stop();
          }
        }
      }, _callee);
    }));

    return function (_x) {
      return _ref3.apply(this, arguments);
    };
  }());
}

function ownKeys$b(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$b(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$b(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$b(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * @typedef {module:ol/layer/Base~BaseLayer|Object} LayerLike
 */

/**
 * Layers container mixin.
 */

var layersContainer = {
  mixins: [identMap, rxSubs],
  computed: {
    /**
     * @returns {string|undefined}
     */
    layersCollectionIdent: function layersCollectionIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'layers_collection');
    }
  },
  watch: _objectSpread$b({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['layersCollectionIdent'])),
  created: function created() {
    /**
     * @type {module:ol/Collection~Collection<module:ol/layer/Base~BaseLayer>}
     * @private
     */
    this._layersCollection = this.instanceFactoryCall(this.layersCollectionIdent, function () {
      return new Collection();
    });
    this._layerSubs = {};
    defineServices$3.call(this);
  },
  methods: {
    /**
     * @returns {{readonly layersContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get layersContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      subscribeToCollectionEvents$1.call(this);
    },

    /**
     * @param {LayerLike} layer
     * @return {BaseLayer}
     */
    initializeLayer: function initializeLayer$1(layer) {
      var _layer;

      layer = ((_layer = layer) === null || _layer === void 0 ? void 0 : _layer.$layer) || layer;
      instanceOf(layer, BaseLayer);
      return initializeLayer(layer);
    },

    /**
     * @param {LayerLike[]|module:ol/Collection~Collection<LayerLike>} layers
     */
    addLayers: function addLayers(layers) {
      forEach(layers, this.addLayer.bind(this));
    },

    /**
     * @param {LayerLike} layer
     */
    addLayer: function addLayer(layer) {
      layer = this.initializeLayer(layer);
      if (this.getLayerById(getLayerId(layer))) return;
      this.$layersCollection.push(layer);
    },

    /**
     * @param {LayerLike[]|module:ol/Collection~Collection<LayerLike>} layers
     */
    removeLayers: function removeLayers(layers) {
      forEach(layers, this.removeLayer.bind(this));
    },

    /**
     * @param {LayerLike} layer
     */
    removeLayer: function removeLayer(layer) {
      var _layer2;

      layer = this.getLayerById(getLayerId(((_layer2 = layer) === null || _layer2 === void 0 ? void 0 : _layer2.$layer) || layer));
      if (!layer) return;
      this.$layersCollection.remove(layer);
    },

    /**
     * @return {void}
     */
    clearLayers: function clearLayers() {
      this.$layersCollection.clear();
    },

    /**
     * @return {Array<module:ol/layer/Base~BaseLayer>}
     */
    getLayers: function getLayers() {
      return this.$layersCollection.getArray().slice();
    },

    /**
     * @return {module:ol/Collection~Collection<module:ol/layer/Base~BaseLayer>}
     */
    getLayersCollection: function getLayersCollection() {
      return this._layersCollection;
    },

    /**
     * @param {string|number} layerId
     * @return {module:ol/layer/Base~BaseLayer|undefined}
     */
    getLayerById: function getLayerById(layerId) {
      return find(this.getLayers(), function (layer) {
        return getLayerId(layer) === layerId;
      });
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    layersCollectionIdentChanged: function layersCollectionIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.$layersCollection) {
        this.setInstance(value, this.$layersCollection);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    }
  }
};

function defineServices$3() {
  Object.defineProperties(this, {
    $layersCollection: {
      enumerable: true,
      get: this.getLayersCollection
    }
  });
}

function subscribeToCollectionEvents$1() {
  var _this = this;

  var adds = fromOlEvent(this.$layersCollection, CollectionEventType.ADD).pipe(map(function (evt) {
    return _objectSpread$b(_objectSpread$b({}, evt), {}, {
      element: _this.initializeLayer(evt.element)
    });
  }), tap(function (_ref) {
    var element = _ref.element;
    var uid = getUid(element);
    var propChanges = fromOlChangeEvent(element, 'id', true);
    _this._layerSubs[uid] = _this.subscribeTo(propChanges, _this.scheduleRefresh.bind(_this));
  }));
  var removes = fromOlEvent(this.$layersCollection, CollectionEventType.REMOVE).pipe(tap(function (_ref2) {
    var element = _ref2.element;
    var uid = getUid(element);

    if (_this._layerSubs[uid]) {
      _this.unsubscribe(_this._layerSubs[uid]);

      delete _this._layerSubs[uid];
    }
  }));
  var events = merge(adds, removes).pipe(bufferDebounceTime(FRAME_TIME));
  this.subscribeTo(events, /*#__PURE__*/function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee(events) {
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _context.next = 2;
              return _this.debounceChanged();

            case 2:
              forEach(events, function (_ref4) {
                var type = _ref4.type,
                    element = _ref4.element;

                _this.$emit(type + 'layer', element); // todo remove in v0.13.x


                _this.$emit(type + ':layer', element);
              });

            case 3:
            case "end":
              return _context.stop();
          }
        }
      }, _callee);
    }));

    return function (_x) {
      return _ref3.apply(this, arguments);
    };
  }());
}

function ownKeys$a(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$a(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$a(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$a(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * @typedef {module:ol/Overlay~Overlay|Object} OverlayLike
 */

/**
 * Overlays container mixin.
 */

var overlaysContainer = {
  mixins: [identMap, rxSubs],
  computed: {
    /**
     * @returns {string|undefined}
     */
    overlaysCollectionIdent: function overlaysCollectionIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'overlays_collection');
    }
  },
  watch: _objectSpread$a({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['overlaysCollectionIdent'])),
  created: function created() {
    /**
     * @type {module:ol/Collection~Collection<module:ol/Overlay~Overlay>}
     * @private
     */
    this._overlaysCollection = this.instanceFactoryCall(this.overlaysCollectionIdent, function () {
      return new Collection();
    });
    this._overlaySubs = {};
    defineServices$2.call(this);
  },
  methods: {
    /**
     * @returns {readonly overlaysContainer: Object}}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get overlaysContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {void}
     * @protected
     */
    subscribeAll: function subscribeAll() {
      subscribeToCollectionEvents.call(this);
    },

    /**
     * @param {OverlayLike} overlay
     * @return {Overlay}
     */
    initializeOverlay: function initializeOverlay$1(overlay) {
      var _overlay;

      overlay = ((_overlay = overlay) === null || _overlay === void 0 ? void 0 : _overlay.$overlay) || overlay;
      instanceOf(overlay, Overlay);
      return initializeOverlay(overlay);
    },

    /**
     * @param {OverlayLike[]|module:ol/Collection~Collection<OverlayLike>} overlays
     */
    addOverlays: function addOverlays(overlays) {
      forEach(overlays, this.addOverlay.bind(this));
    },

    /**
     * @param {OverlayLike} overlay
     */
    addOverlay: function addOverlay(overlay) {
      overlay = this.initializeOverlay(overlay);
      if (this.getOverlayById(getOverlayId(overlay))) return;
      this.$overlaysCollection.push(overlay);
    },

    /**
     * @param {OverlayLike[]|module:ol/Collection~Collection<OverlayLike>} overlays
     */
    removeOverlays: function removeOverlays(overlays) {
      forEach(overlays, this.removeOverlay.bind(this));
    },

    /**
     * @param {OverlayLike} overlay
     */
    removeOverlay: function removeOverlay(overlay) {
      var _overlay2;

      overlay = this.getOverlayById(getOverlayId(((_overlay2 = overlay) === null || _overlay2 === void 0 ? void 0 : _overlay2.$overlay) || overlay));
      if (!overlay) return;
      this.$overlaysCollection.remove(overlay);
    },

    /**
     * @return {void}
     */
    clearOverlays: function clearOverlays() {
      this.$overlaysCollection.clear();
    },

    /**
     * @return {Array<module:ol/Overlay~Overlay>}
     */
    getOverlays: function getOverlays() {
      return this.$overlaysCollection.getArray().slice();
    },

    /**
     * @return {module:ol/Collection~Collection<module:ol/Overlay~Overlay>}
     */
    getOverlaysCollection: function getOverlaysCollection() {
      return this._overlaysCollection;
    },

    /**
     * @param {string|number} overlayId
     * @return {module:ol/Overlay~Overlay|undefined}
     */
    getOverlayById: function getOverlayById(overlayId) {
      return find(this.getOverlays(), function (overlay) {
        return getOverlayId(overlay) === overlayId;
      });
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    overlaysCollectionIdentChanged: function overlaysCollectionIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.$overlaysCollection) {
        this.setInstance(value, this.$overlaysCollection);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    }
  }
};

function defineServices$2() {
  Object.defineProperties(this, {
    $overlaysCollection: {
      enumerable: true,
      get: this.getOverlaysCollection
    }
  });
}

function subscribeToCollectionEvents() {
  var _this = this;

  var adds = fromOlEvent(this.$overlaysCollection, CollectionEventType.ADD).pipe(map(function (evt) {
    return _objectSpread$a(_objectSpread$a({}, evt), {}, {
      element: _this.initializeOverlay(evt.element)
    });
  }), tap(function (_ref) {
    var element = _ref.element;
    var uid = getUid(element);
    var propChanges = fromOlChangeEvent(element, 'id', true);
    _this._overlaySubs[uid] = _this.subscribeTo(propChanges, _this.scheduleRefresh.bind(_this));
  }));
  var removes = fromOlEvent(this.$overlaysCollection, CollectionEventType.REMOVE).pipe(tap(function (_ref2) {
    var element = _ref2.element;
    var uid = getUid(element);

    if (_this._overlaySubs[uid]) {
      _this.unsubscribe(_this._overlaySubs[uid]);

      delete _this._overlaySubs[uid];
    }
  }));
  var events = merge(adds, removes).pipe(bufferDebounceTime(FRAME_TIME));
  this.subscribeTo(events, /*#__PURE__*/function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee(events) {
      return _regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _context.next = 2;
              return _this.debounceChanged();

            case 2:
              forEach(events, function (_ref4) {
                var type = _ref4.type,
                    element = _ref4.element;

                _this.$emit(type + 'overlay', element); // todo remove in v0.13.x


                _this.$emit(type + ':overlay', element);
              });

            case 3:
            case "end":
              return _context.stop();
          }
        }
      }, _callee);
    }));

    return function (_x) {
      return _ref3.apply(this, arguments);
    };
  }());
}

/**
 * @typedef {module:ol/style/Stroke~Stroke|Object|undefined} StrokeStyleLike
 */

/**
 * @typedef {Object} StrokeStyleTarget
 * @property {function(): module:ol/style/Stroke~Stroke|undefined} getStroke
 * @property {function(module:ol/style/Stroke~Stroke|undefined): void} setStroke
 */

/**
 * Stroke style container.
 */

var strokeStyleContainer = {
  created: function created() {
    this._stroke = undefined;
    this._strokeVm = undefined;
    defineServices$1.call(this);
  },
  methods: {
    /**
     * @returns {{readonly strokeStyleContainer: Object}}
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get strokeStyleContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {StrokeStyleTarget}
     */
    getStrokeStyleTarget: function getStrokeStyleTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getStrokeStyleTarget()"));
    },

    /**
     * @returns {module:ol/style/Stroke~Stroke|undefined}
     */
    getStroke: function getStroke() {
      var _this$getStrokeStyleT;

      return coalesce((_this$getStrokeStyleT = this.getStrokeStyleTarget()) === null || _this$getStrokeStyleT === void 0 ? void 0 : _this$getStrokeStyleT.getStroke(), this._stroke);
    },

    /**
     * @return {Object}
     */
    getStrokeVm: function getStrokeVm() {
      return this._strokeVm;
    },

    /**
     * @param {module:ol/style/Stroke~Stroke|undefined} stroke
     */
    setStroke: function setStroke(stroke) {
      var _stroke;

      stroke = ((_stroke = stroke) === null || _stroke === void 0 ? void 0 : _stroke.$style) || stroke;
      stroke || (stroke = undefined);
      assert(!stroke || stroke instanceof Stroke, 'Invalid stroke');
      var strokeTarget = this.getStrokeStyleTarget();

      if (strokeTarget && stroke !== strokeTarget.getStroke()) {
        strokeTarget.setStroke(stroke);
        this.scheduleRefresh();
      }

      if (stroke !== this._stroke) {
        var _stroke2;

        this._stroke = stroke;
        this._strokeVm = ((_stroke2 = stroke) === null || _stroke2 === void 0 ? void 0 : _stroke2.vm) && stroke.vm[0];
        this.scheduleRefresh();
      }
    }
  }
};

function defineServices$1() {
  Object.defineProperties(this, {
    $stroke: {
      enumerable: true,
      get: this.getStroke
    },
    $strokeVm: {
      enumerable: true,
      get: this.getStrokeVm
    }
  });
}

function ownKeys$9(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$9(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$9(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$9(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var regShapeStyle = {
  mixins: [fillStyleContainer, strokeStyleContainer, imageStyle],
  props: {
    // ol/style/RegularShape
    points: Number,
    radius: Number,
    radius1: Number,
    radius2: Number,
    angle: {
      type: Number,
      default: 0
    }
  },
  computed: {
    stroke: function stroke() {
      if (!(this.rev && this.$stroke)) return;
      return dumpStrokeStyle(this.$stroke);
    },
    fill: function fill() {
      if (!(this.rev && this.$fill)) return;
      return dumpFillStyle(this.$fill);
    }
  },
  watch: _objectSpread$9({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['points', 'radius', 'radius1', 'radius2', 'angle', 'stroke', 'fill'], ['stroke', 'fill'])),
  methods: {
    /**
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      return mergeDescriptors(style.methods.getServices.call(this), fillStyleContainer.methods.getServices.call(this), strokeStyleContainer.methods.getServices.call(this));
    },

    /**
     * @protected
     */
    syncNonObservable: function syncNonObservable() {
      imageStyle.methods.syncNonObservable.call(this);
    },

    /**
     * @return {FillStyleTarget}
     * @protected
     */
    getFillStyleTarget: function getFillStyleTarget() {
      var _this = this;

      return {
        getFill: function getFill() {
          var _this$$style;

          return (_this$$style = _this.$style) === null || _this$$style === void 0 ? void 0 : _this$$style.getFill();
        },
        setFill: function setFill() {
          if (process.env.VUELAYERS_DEBUG) {
            _this.$logger.log('fill changed, scheduling recreate...');
          }

          _this.scheduleRecreate();
        }
      };
    },

    /**
     * @return {StrokeStyleTarget}
     * @protected
     */
    getStrokeStyleTarget: function getStrokeStyleTarget() {
      var _this2 = this;

      return {
        getStroke: function getStroke() {
          var _this2$$style;

          return (_this2$$style = _this2.$style) === null || _this2$$style === void 0 ? void 0 : _this2$$style.getStroke();
        },
        setStroke: function setStroke() {
          if (process.env.VUELAYERS_DEBUG) {
            _this2.$logger.log('stroke changed, scheduling recreate...');
          }

          _this2.scheduleRecreate();
        }
      };
    },
    getAnchor: function getAnchor() {
      var _this$$style2;

      return (_this$$style2 = this.$style) === null || _this$$style2 === void 0 ? void 0 : _this$$style2.getAnchor();
    },
    getAngle: function getAngle() {
      var _this$$style3;

      return coalesce((_this$$style3 = this.$style) === null || _this$$style3 === void 0 ? void 0 : _this$$style3.getAngle(), this.angle);
    },
    getImage: function getImage() {
      var _this$$style4;

      return (_this$$style4 = this.$style) === null || _this$$style4 === void 0 ? void 0 : _this$$style4.getImage();
    },
    getOrigin: function getOrigin() {
      var _this$$style5;

      return (_this$$style5 = this.$style) === null || _this$$style5 === void 0 ? void 0 : _this$$style5.getOrigin();
    },
    getPoints: function getPoints() {
      var _this$$style6;

      return coalesce((_this$$style6 = this.$style) === null || _this$$style6 === void 0 ? void 0 : _this$$style6.getPoints(), this.points);
    },
    getRadius: function getRadius() {
      var _this$$style7;

      return coalesce((_this$$style7 = this.$style) === null || _this$$style7 === void 0 ? void 0 : _this$$style7.getRadius(), this.radius);
    },
    getRadius2: function getRadius2() {
      var _this$$style8;

      return coalesce((_this$$style8 = this.$style) === null || _this$$style8 === void 0 ? void 0 : _this$$style8.getRadius2(), this.radius2);
    },
    getSize: function getSize() {
      var _this$$style9;

      return (_this$$style9 = this.$style) === null || _this$$style9 === void 0 ? void 0 : _this$$style9.getSize();
    },

    /**
     * @param {Object|undefined} value
     * @param {Object|undefined} prev
     * @protected
     */
    strokeChanged: function strokeChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:stroke', value && clonePlainObject(value));
    },

    /**
     * @param {Object|undefined} value
     * @param {Object|undefined} prev
     * @protected
     */
    fillChanged: function fillChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:fill', value && clonePlainObject(value));
    }
  }
};

function ownKeys$8(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$8(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$8(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$8(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base simple geometry with coordinates mixin.
 */

var simpleGeometry = {
  mixins: [geometry],
  props: {
    // ol/geom/SimpleGeometry

    /**
     * @type {number[]} Coordinates in map data projection
     */
    coordinates: {
      type: Array,
      required: true,
      validator: /*#__PURE__*/negate(isEmpty)
    } // todo add support of coord layout
    // /**
    //  * @type {string}
    //  */
    // layout: {
    //   type: String,
    //   default: GeometryLayout.XY,
    //   validator: value => Object.values(GeometryLayout).includes(value.toUpperCase()),
    // },

  },
  data: function data() {
    return {
      currentCoordinatesViewProj: clonePlainObject(this.coordinates)
    };
  },
  computed: {
    coordinatesDataProj: function coordinatesDataProj() {
      return roundCoords(this.type, this.coordinates);
    },
    coordinatesViewProj: function coordinatesViewProj() {
      return this.coordinatesToViewProj(this.coordinates);
    },
    currentCoordinatesDataProj: function currentCoordinatesDataProj() {
      return this.coordinatesToDataProj(this.currentCoordinatesViewProj);
    },
    pointDataProj: function pointDataProj() {
      return this.rev ? this.findPointOnSurface() : findPointOnSurface({
        type: this.type,
        coordinates: this.coordinatesDataProj
      });
    },
    pointViewProj: function pointViewProj() {
      return this.rev ? this.findPointOnSurface(true) : findPointOnSurface({
        type: this.type,
        coordinates: this.coordinatesViewProj
      });
    }
  },
  watch: _objectSpread$8({
    rev: function rev() {
      if (!this.$geometry) return;

      if (!isEqualCoord({
        coordinates: this.currentCoordinatesViewProj,
        extent: boundingExtent(this.currentCoordinatesViewProj)
      }, {
        coordinates: this.$geometry.getCoordinates(),
        extent: this.$geometry.getExtent()
      })) {
        this.currentCoordinatesViewProj = this.$geometry.getCoordinates();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['coordinatesViewProj', 'currentCoordinatesDataProj', 'pointDataProj'], ['coordinatesViewProj', 'currentCoordinatesDataProj', 'pointDataProj'])),
  created: function created() {
    this.currentCoordinatesViewProj = clonePlainObject(this.coordinatesViewProj);
    this.extentViewProj = boundingExtent(this.currentCoordinatesViewProj);
  },
  methods: {
    /**
     * @returns {function}
     */
    getCoordinatesTransformFunction: function getCoordinatesTransformFunction() {
      return transforms[this.getType()].transform;
    },

    /**
     * @param {number[]} coordinates
     * @returns {Promise<number[]>}
     */
    coordinatesToDataProj: function coordinatesToDataProj(coordinates) {
      var transform = this.getCoordinatesTransformFunction();
      return transform(coordinates, this.resolvedViewProjection, this.resolvedDataProjection);
    },

    /**
     * @param {number[]} coordinates
     * @returns {Promise<number[]>}
     */
    coordinatesToViewProj: function coordinatesToViewProj(coordinates) {
      var transform = this.getCoordinatesTransformFunction();
      return transform(coordinates, this.resolvedDataProjection, this.resolvedViewProjection);
    },

    /**
     * @param {boolean} [viewProj=false]
     * @return {number[]}
     */
    getCoordinates: function getCoordinates() {
      var _this$$geometry;

      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      var coordinates = coalesce((_this$$geometry = this.$geometry) === null || _this$$geometry === void 0 ? void 0 : _this$$geometry.getCoordinates(), this.currentCoordinatesViewProj);
      return viewProj ? roundCoords(this.getType(), coordinates) : this.coordinatesToDataProj(coordinates);
    },

    /**
     * @param {number[]} coordinates
     * @param {boolean} [viewProj=false]
     */
    setCoordinates: function setCoordinates(coordinates) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      assert(isArray(coordinates), 'Invalid coordinates');
      coordinates = viewProj ? roundCoords(this.getType(), coordinates) : this.coordinatesToViewProj(coordinates);
      var extent = boundingExtent(coordinates);

      if (!isEqualCoord({
        coordinates: coordinates,
        extent: extent
      }, {
        coordinates: this.currentCoordinatesViewProj,
        extent: this.extentViewProj
      })) {
        this.currentCoordinatesViewProj = coordinates;
      }

      if (this.$geometry && !isEqualCoord({
        coordinates: coordinates,
        extent: extent
      }, {
        coordinates: this.$geometry.getCoordinates(),
        extent: this.$geometry.getExtent()
      })) {
        this.$geometry.setCoordinates(coordinates);
      }
    },

    /**
     * @param {boolean} [viewProj=false]
     * @returns {number[]>}
     */
    getFirstCoordinate: function getFirstCoordinate() {
      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      var coordinate;

      if (this.$geometry) {
        coordinate = this.$geometry.getFirstCoordinate();
      } else {
        switch (this.getType()) {
          case GeometryType.POINT:
          case GeometryType.CIRCLE:
            coordinate = this.currentCoordinatesViewProj;
            break;

          case GeometryType.LINE_STRING:
          case GeometryType.MULTI_POINT:
            coordinate = this.currentCoordinatesViewProj[0];
            break;

          case GeometryType.POLYGON:
          case GeometryType.MULTI_LINE_STRING:
            coordinate = this.currentCoordinatesViewProj[0][0];
            break;

          case GeometryType.MULTI_POLYGON:
            coordinate = this.currentCoordinatesViewProj[0][0][0];
            break;
        }
      }

      return viewProj ? roundPointCoords(coordinate) : this.pointToDataProj(coordinate);
    },

    /**
     * @param {boolean} [viewProj=false]
     * @returns {number[]}
     */
    getLastCoordinate: function getLastCoordinate() {
      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      var coordinate;

      if (this.$geometry) {
        coordinate = this.$geometry.getLastCoordinate();
      } else {
        var arr;

        switch (this.getType()) {
          case GeometryType.POINT:
          case GeometryType.CIRCLE:
            coordinate = this.currentCoordinatesViewProj;
            break;

          case GeometryType.LINE_STRING:
          case GeometryType.MULTI_POINT:
            coordinate = this.currentCoordinatesViewProj[this.currentCoordinatesViewProj.length - 1];
            break;

          case GeometryType.POLYGON:
          case GeometryType.MULTI_LINE_STRING:
            arr = this.currentCoordinatesViewProj[this.currentCoordinatesViewProj.length - 1];
            coordinate = arr[arr.length - 1];
            break;

          case GeometryType.MULTI_POLYGON:
            arr = this.currentCoordinatesViewProj[this.currentCoordinatesViewProj.length - 1];
            arr = arr[arr.length - 1];
            coordinate = arr[arr.length - 1];
            break;
        }
      }

      return viewProj ? roundPointCoords(coordinate) : this.pointToDataProj(coordinate);
    },

    /**
     * @returns {string}
     */
    getLayout: function getLayout() {
      var _this$$geometry2;

      return coalesce((_this$$geometry2 = this.$geometry) === null || _this$$geometry2 === void 0 ? void 0 : _this$$geometry2.getLayout(), GeometryLayout.XY);
    },

    /**
     * @param {boolean} [viewProj=false]
     * @return {number[]|undefined}
     */
    findPointOnSurface: function findPointOnSurface$1() {
      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      return findPointOnSurface({
        type: this.getType(),
        coordinates: this.getCoordinates(viewProj)
      });
    },

    /**
     * @param {number[]} value
     * @protected
     */
    coordinatesViewProjChanged: function coordinatesViewProjChanged(value) {
      this.setCoordinates(value, true);
    },

    /**
     * @param {number[]} value
     * @protected
     */
    currentCoordinatesDataProjChanged: function currentCoordinatesDataProjChanged(value) {
      if (isEqual(value, this.coordinatesDataProj)) return;
      this.$emit('update:coordinates', clonePlainObject(value));
    },

    /**
     * @param {number[]} value
     * @param {number[]} prev
     * @protected
     */
    pointDataProjChanged: function pointDataProjChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:point', value === null || value === void 0 ? void 0 : value.slice());
    }
  }
};

function boundingExtent(coordinates) {
  if (!coordinates) return;

  if (!isArray(coordinates[0])) {
    coordinates = [coordinates];
  }

  return boundingExtent$1(coordinates);
}

/**
 * @typedef {module:ol/style/Text~Text|Object|undefined} TextStyleLike
 */

/**
 * @typedef {Object} TextStyleTarget
 * @property {function(): module:ol/style/Text~Text|undefined} getText
 * @property {function(module:ol/style/Text~Text|undefined): void} setText
 */

/**
 * Text style container.
 */

var textStyleContainer = {
  created: function created() {
    this._text = undefined;
    this._textVm = undefined;
    defineServices.call(this);
  },
  methods: {
    /**
     * @returns {{readonly textStyleContainer: Object}}
     */
    getServices: function getServices() {
      var vm = this;
      return {
        get textStyleContainer() {
          return vm;
        }

      };
    },

    /**
     * @return {TextStyleTarget|undefined}
     */
    getTextStyleTarget: function getTextStyleTarget() {
      throw new Error("".concat(this.vmName, " not implemented method: getTextStyleTarget()"));
    },

    /**
     * @returns {module:ol/style/Text~Text|null}
     */
    getText: function getText() {
      var _this$getTextStyleTar;

      return coalesce((_this$getTextStyleTar = this.getTextStyleTarget()) === null || _this$getTextStyleTar === void 0 ? void 0 : _this$getTextStyleTar.getText(), this._text);
    },

    /**
     * @return {Object}
     */
    getTextVm: function getTextVm() {
      return this._textVm;
    },

    /**
     * @param {module:ol/style/Text~Text|undefined} text
     */
    setText: function setText(text) {
      var _text;

      text = ((_text = text) === null || _text === void 0 ? void 0 : _text.$style) || text;
      text || (text = undefined);
      assert(!text || text instanceof Text, 'Invalid text style');
      var textTarget = this.getTextStyleTarget();

      if (textTarget && text !== textTarget.getText()) {
        textTarget.setText(text);
        this.scheduleRefresh();
      }

      if (text !== this._text) {
        var _text2;

        this._text = text;
        this._textVm = ((_text2 = text) === null || _text2 === void 0 ? void 0 : _text2.vm) && text.vm[0];
        this.scheduleRefresh();
      }
    }
  }
};

function defineServices() {
  Object.defineProperties(this, {
    $text: {
      enumerable: true,
      get: this.getText
    },
    $textVm: {
      enumerable: true,
      get: this.getTextVm
    }
  });
}

function ownKeys$7(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$7(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$7(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$7(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var tileSource = {
  mixins: [source],
  props: {
    // ol/source/Tile

    /**
     * @type {number|undefined}
     */
    cacheSize: Number,

    /**
     * @type {boolean|undefined}
     */
    opaque: Boolean,

    /**
     * @type {number}
     */
    tilePixelRatio: {
      type: Number,
      default: 1
    },

    /**
     * @type {function|undefined}
     */
    tileGridFactory: Function,

    /**
     * @type {number}
     */
    transition: Number,

    /**
     * @type {string|undefined}
     */
    tileKey: String,

    /**
     * @type {number}
     */
    zDirection: {
      type: Number,
      default: 0
    }
  },
  data: function data() {
    return {
      tileGrid: undefined
    };
  },
  computed: {
    /**
     * @type {string|undefined}
     */
    tileGridIdent: function tileGridIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'tile_grid');
    },

    /**
     * @returns {function|undefined}
     */
    derivedTileGridFactory: function derivedTileGridFactory() {
      return this.tileGridFactory;
    },

    /**
     * @returns {function|undefined}
     */
    inputTileGridFactory: function inputTileGridFactory() {
      if (!isFunction(this.derivedTileGridFactory)) return;
      return sealFactory(this.derivedTileGridFactory.bind(this));
    },

    /**
     * @returns {string}
     */
    resolvedDataProjection: function resolvedDataProjection() {
      return coalesce(this.projection, this.resolvedViewProjection);
    }
  },
  watch: _objectSpread$7({
    rev: function rev() {
      if (!this.$source) return;

      if (this.tileGrid !== this.$source.getTileGrid()) {
        this.tileGrid = this.$source.getTileGrid();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['tileGridIdent', 'inputTileGridFactory', 'tileGrid', 'tileKey', 'opaque', 'tilePixelRatio', 'cacheSize', 'transition', 'zDirection'])),
  created: function created() {
    if (isFunction(this.inputTileGridFactory)) {
      this.tileGrid = this.instanceFactoryCall(this.tileGridIdent, this.inputTileGridFactory.bind(this));
    }
  },
  methods: {
    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      source.methods.subscribeAll.call(this);
    },

    /**
     * @returns {module:ol/tilegrid/TileGrid~TileGrid|undefined}
     */
    getTileGrid: function getTileGrid() {
      var _this$$source;

      return coalesce((_this$$source = this.$source) === null || _this$$source === void 0 ? void 0 : _this$$source.getTileGrid(), this.tileGrid);
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    tileGridIdentChanged: function tileGridIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.tileGrid) {
        this.setInstance(value, this.tileGrid);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    },

    /**
     * @param {function|undefined} value
     * @return {Promise<void>}
     * @protected
     */
    inputTileGridFactoryChanged: function inputTileGridFactoryChanged(value) {
      while (this.hasInstance(this.tileGridIdent)) {
        this.unsetInstance(this.tileGridIdent);
      }

      if (isFunction(value)) {
        this.tileGrid = this.instanceFactoryCall(this.tileGridIdent, value.bind(this));
      } else {
        this.tileGrid = null;
      }
    }
  }
};

function ownKeys$6(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$6(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$6(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$6(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var isNotEmptyString$1 = /*#__PURE__*/and(isString, negate(isEmpty));
var isArrayOfNotEmptyStrings = /*#__PURE__*/and(isArray, function (value) {
  return value.every(isNotEmptyString$1);
});
var urlTileSource = {
  mixins: [tileSource],
  props: {
    // ol/source/UrlTile

    /**
     * @type {function|undefined}
     */
    tileLoadFunction: Function,

    /**
     * @type {function|undefined}
     * @deprecated
     * @todo remove in v0.13.x
     */
    tileUrlFunc: Function,

    /**
     * @type {function|undefined}
     */
    tileUrlFunction: Function,

    /**
     * @type {string|undefined}
     */
    url: {
      type: String,
      validator: isNotEmptyString$1
    },

    /**
     * @type {string[]|undefined}
     */
    urls: {
      type: Array,
      validator: isArrayOfNotEmptyStrings
    }
  },
  data: function data() {
    return {
      currentTileLoadFunction: undefined,
      currentTileUrlFunction: undefined,
      currentUrls: []
    };
  },
  computed: {
    urlTokens: function urlTokens() {
      return [];
    },
    inputUrl: function inputUrl() {
      if (!this.url) return;
      return replaceTokens(this.url, pick(this, this.urlTokens));
    },
    parsedUrls: function parsedUrls() {
      var urls = [];

      if (this.urls && this.urls.length > 0) {
        urls.push.apply(urls, _toConsumableArray(this.urls));
      } else if (this.url) {
        urls.push(this.url);
      }

      var tokens = pick(this, this.urlTokens);
      return urls.map(function (url) {
        return replaceTokens(url, tokens);
      });
    },
    inputUrls: function inputUrls() {
      return this.parsedUrls.reduce(function (urls, url) {
        return urls.concat.apply(urls, _toConsumableArray(expandUrl(url)));
      }, []);
    },
    inputTileUrlFunction: function inputTileUrlFunction() {
      return coalesce(this.tileUrlFunction, this.tileUrlFunc);
    },
    inputTileLoadFunction: function inputTileLoadFunction() {
      return this.tileLoadFunction;
    }
  },
  watch: _objectSpread$6({
    rev: function rev() {
      if (!this.$source) return;

      if (this.currentTileLoadFunction !== this.$source.getTileLoadFunction()) {
        this.currentTileLoadFunction = this.$source.getTileLoadFunction();
      }

      if (this.currentTileUrlFunction !== this.$source.getTileUrlFunction()) {
        this.currentTileUrlFunction = this.$source.getTileUrlFunction();
      }

      if (!isEqual(this.currentUrls, this.$source.getUrls()) && !this.inputTileUrlFunction) {
        this.currentUrls = this.$source.getUrls();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['inputUrls', 'currentUrls', 'inputTileUrlFunction', 'currentTileUrlFunction', 'inputTileLoadFunction', 'currentTileLoadFunction'], ['inputUrls', 'currentUrls'])),
  created: function created() {
    var _this$inputUrls;

    if (process.env.NODE_ENV !== 'production') {
      if (this.tileUrlFunc) {
        this.$logger.warn("'tileUrlFunc' prop is deprecated. Use 'tileUrlFunction' prop instead.");
      }
    }

    this.currentUrls = (_this$inputUrls = this.inputUrls) === null || _this$inputUrls === void 0 ? void 0 : _this$inputUrls.slice();
    this.currentTileUrlFunction = this.inputTileUrlFunction;
    this.currentTileLoadFunction = this.inputTileLoadFunction;
  },
  updated: function updated() {
    if (process.env.NODE_ENV !== 'production') {
      if (this.tileUrlFunc) {
        this.$logger.warn("'tileUrlFunc' prop is deprecated. Use 'tileUrlFunction' prop instead.");
      }
    }
  },
  methods: {
    /**
     * @returns {void}
     */
    subscribeAll: function subscribeAll() {
      tileSource.methods.subscribeAll.call(this);
      subscribeToSourceEvents$1.call(this);
    },

    /**
     * @returns {module:ol/Tile.LoadFunction}
     */
    getTileLoadFunction: function getTileLoadFunction() {
      var _this$$source;

      return coalesce((_this$$source = this.$source) === null || _this$$source === void 0 ? void 0 : _this$$source.getTileLoadFunction(), this.currentTileLoadFunction);
    },

    /**
     * @param {module:ol/Tile.LoadFunction} tileLoadFunction
     */
    setTileLoadFunction: function setTileLoadFunction(tileLoadFunction) {
      assert(isFunction(tileLoadFunction), 'Invalid tile load function');

      if (tileLoadFunction !== this.currentTileLoadFunction) {
        this.currentTileLoadFunction = tileLoadFunction;
      }

      if (this.$source && tileLoadFunction !== this.$source.getTileLoadFunction()) {
        this.$source.setTileLoadFunction(tileLoadFunction);
      }
    },

    /**
     * @returns {module:ol/Tile.UrlFunction}
     */
    getTileUrlFunction: function getTileUrlFunction() {
      var _this$$source2;

      return coalesce((_this$$source2 = this.$source) === null || _this$$source2 === void 0 ? void 0 : _this$$source2.getTileUrlFunction(), this.currentTileUrlFunction);
    },

    /**
     * @param {module:ol/Tile.UrlFunction} tileUrlFunction
     * @param {number} [tileKey]
     */
    setTileUrlFunction: function setTileUrlFunction(tileUrlFunction, tileKey) {
      var _this$inputUrls2;

      assert(isFunction(tileUrlFunction), 'Invalid tile url function');
      tileKey || (tileKey = coalesce((_this$inputUrls2 = this.inputUrls) === null || _this$inputUrls2 === void 0 ? void 0 : _this$inputUrls2.join('\n'), v4()));

      if (tileUrlFunction !== this.currentTileUrlFunction) {
        this.currentTileUrlFunction = tileUrlFunction;
      }

      if (this.$source && tileUrlFunction !== this.$source.getTileUrlFunction()) {
        this.$source.setTileUrlFunction(tileUrlFunction, tileKey);
      }
    },

    /**
     * @returns {string[]|undefined}
     */
    getUrls: function getUrls() {
      var _this$$source3;

      return coalesce((_this$$source3 = this.$source) === null || _this$$source3 === void 0 ? void 0 : _this$$source3.getUrls(), this.currentUrls);
    },

    /**
     * @param {string[]} urls
     */
    setUrls: function setUrls(urls) {
      assert(isArrayOfNotEmptyStrings(urls), 'Invalid urls');
      urls = urls.slice();

      if (!isEqual(urls, this.currentUrls)) {
        this.currentUrls = urls;
      }

      if (this.$source) {
        if (this.inputTileUrlFunction) {
          this.$source.setTileUrlFunction(this.inputTileUrlFunction, urls.join('\n'));
        } else if (!isEqual(urls, this.$source.getUrls())) {
          this.$source.setUrls(urls);
        }
      }
    },

    /**
     * @param {string} url
     */
    setUrl: function setUrl(url) {
      assert(isNotEmptyString$1(url), 'Invalid url');
      this.setUrls(expandUrl(url));
    },

    /**
     * @param {string[]} value
     * @protected
     */
    inputUrlsChanged: function inputUrlsChanged(value) {
      this.setUrls(value);
    },

    /**
     * @param {string[]} value
     * @protected
     */
    currentUrlsChanged: function currentUrlsChanged(value) {
      if (isEqual(value, this.urls)) return;
      this.$emit('update:urls', value === null || value === void 0 ? void 0 : value.slice());
    },

    /**
     * @param {function|undefined} value
     * @protected
     */
    inputTileUrlFunctionChanged: function inputTileUrlFunctionChanged(value) {
      this.setTileUrlFunction(value);
    },

    /**
     * @param {function|undefined} value
     * @protected
     */
    currentTileUrlFunctionChanged: function currentTileUrlFunctionChanged(value) {
      if (value === this.inputTileUrlFunction) return;
      this.$emit('update:tileUrlFunction', value);
    },

    /**
     * @param {function|undefined} value
     * @protected
     */
    inputTileLoadFunctionChanged: function inputTileLoadFunctionChanged(value) {
      this.setTileLoadFunction(value);
    },

    /**
     * @param {function|undefined} value
     * @protected
     */
    currentTileLoadFunctionChanged: function currentTileLoadFunctionChanged(value) {
      if (value === this.inputTileLoadFunction) return;
      this.$emit('update:tileLoadFunction', value);
    }
  }
};

function subscribeToSourceEvents$1() {
  return _subscribeToSourceEvents.apply(this, arguments);
}

function _subscribeToSourceEvents() {
  _subscribeToSourceEvents = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
    var _this = this;

    var events;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            events = fromOlEvent(this.$source, [TileEventType.TILELOADSTART, TileEventType.TILELOADEND, TileEventType.TILELOADERROR]);
            this.subscribeTo(events, function (evt) {
              return _this.$emit(evt.type, evt);
            });

          case 2:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));
  return _subscribeToSourceEvents.apply(this, arguments);
}

function ownKeys$5(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$5(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$5(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$5(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base tile image source mixin.
 */

var tileImageSource = {
  mixins: [urlTileSource],
  props: {
    // ol/source/TileImage

    /**
     * @type {string|undefined}
     */
    crossOrigin: String,

    /**
     * @type {number|undefined}
     */
    reprojectionErrorThreshold: Number,

    /**
     * @type {string|undefined}
     */
    tileClass: String,

    /**
     * @type {boolean}
     */
    imageSmoothing: {
      type: Boolean,
      default: true
    }
  },
  watch: _objectSpread$5({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['crossOrigin', 'reprojectionErrorThreshold', 'tileClass', 'imageSmoothing'])),
  methods: {
    setRenderReprojectionEdges: function setRenderReprojectionEdges(render) {
      var _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _this.resolveSource();

              case 2:
                _context.sent.setRenderReprojectionEdges(render);

              case 3:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    }
  }
};

function ownKeys$4(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$4(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$4(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$4(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * Base tile layer mixin.
 */

var tileLayer = {
  mixins: [layer],
  props: {
    // ol/layer/BaseTile

    /**
     * @type {number}
     */
    preload: {
      type: Number,
      default: 0
    },

    /**
     * @type {boolean}
     */
    useInterimTilesOnError: {
      type: Boolean,
      default: true
    }
  },
  data: function data() {
    return {
      currentPreload: this.preload,
      currentUseInterimTilesOnError: this.useInterimTilesOnError
    };
  },
  watch: _objectSpread$4({
    rev: function rev() {
      if (!this.$layer) return;

      if (this.currentPreload !== this.$layer.getPreload()) {
        this.currentPreload = this.$layer.getPreload();
      }

      if (this.currentUseInterimTilesOnError !== this.$layer.getUseInterimTilesOnError()) {
        this.currentUseInterimTilesOnError = this.$layer.getUseInterimTilesOnError();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['preload', 'currentPreload', 'useInterimTilesOnError', 'currentUseInterimTilesOnError'])),
  methods: {
    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      layer.methods.subscribeAll.call(this);
      subscribeToLayerEvents$1.call(this);
    },

    /**
     * @returns {number}
     */
    getPreload: function getPreload() {
      var _this$$layer;

      return coalesce((_this$$layer = this.$layer) === null || _this$$layer === void 0 ? void 0 : _this$$layer.getPreload(), this.currentPreload);
    },

    /**
     * @param {number} preload
     */
    setPreload: function setPreload(preload) {
      preload = Number(preload);
      assert(isNumber(preload), 'Invalid preload');

      if (preload !== this.currentPreload) {
        this.currentPreload = preload;
      }

      if (this.$layer && preload !== this.$layer.getPreload()) {
        this.$layer.setPreload(preload);
      }
    },

    /**
     * @returns {boolean}
     */
    getUseInterimTilesOnError: function getUseInterimTilesOnError() {
      var _this$$layer2;

      return coalesce((_this$$layer2 = this.$layer) === null || _this$$layer2 === void 0 ? void 0 : _this$$layer2.getUseInterimTilesOnError(), this.currentUseInterimTilesOnError);
    },

    /**
     * @param {boolean} useInterimTilesOnError
     */
    setUseInterimTilesOnError: function setUseInterimTilesOnError(useInterimTilesOnError) {
      useInterimTilesOnError = !!useInterimTilesOnError;

      if (useInterimTilesOnError !== this.currentUseInterimTilesOnError) {
        this.currentUseInterimTilesOnError = useInterimTilesOnError;
      }

      if (this.$layer && useInterimTilesOnError !== this.$layer.getUseInterimTilesOnError()) {
        this.$layer.setUseInterimTilesOnError(useInterimTilesOnError);
      }
    },

    /**
     * @param {number} value
     * @protected
     */
    preloadChanged: function preloadChanged(value) {
      this.setPreload(value);
    },

    /**
     * @param {number} value
     * @protected
     */
    currentPreloadChanged: function currentPreloadChanged(value) {
      if (value === this.preload) return;
      this.$emit('update:preload', value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    useInterimTilesOnErrorChanged: function useInterimTilesOnErrorChanged(value) {
      this.setUseInterimTilesOnError(value);
    },

    /**
     * @param {boolean} value
     * @protected
     */
    currentUseInterimTilesOnErrorChanged: function currentUseInterimTilesOnErrorChanged(value) {
      if (value === this.useInterimTilesOnError) return;
      this.$emit('update:useInterimTilesOnError', value);
    }
  }
};

function subscribeToLayerEvents$1() {
  return _subscribeToLayerEvents$1.apply(this, arguments);
}

function _subscribeToLayerEvents$1() {
  _subscribeToLayerEvents$1 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
    var _this = this;

    var setterKey, propChanges;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setterKey = addPrefix('current');
            propChanges = fromOlChangeEvent(this.$layer, ['preload', 'useInterimTilesOnError'], true, function (evt) {
              return _objectSpread$4(_objectSpread$4({}, evt), {}, {
                setter: _this[setterKey(evt.prop)]
              });
            });
            this.subscribeTo(propChanges, function (_ref) {
              var setter = _ref.setter,
                  value = _ref.value;
              return setter(value);
            });

          case 3:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));
  return _subscribeToLayerEvents$1.apply(this, arguments);
}

function ownKeys$3(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$3(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$3(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$3(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var vectorLayer = {
  mixins: [styleContainer, layer],
  props: {
    // ol/layer/BaseVector

    /**
     * @type {function|undefined}
     */
    renderOrder: Function,

    /**
     * @type {number}
     */
    renderBuffer: {
      type: Number,
      default: 100
    },

    /**
     * @type {boolean}
     */
    declutter: Boolean,

    /**
     * When set to `true`, feature batches will be recreated during animations.
     * @type {boolean}
     */
    updateWhileAnimating: Boolean,

    /**
     * When set to `true`, feature batches will be recreated during interactions.
     * @type {boolean}
     */
    updateWhileInteracting: Boolean
  },
  computed: {
    style: function style() {
      var _this = this;

      if (!(this.rev && this.$style)) return;
      var style = this.$style;
      if (isFunction(style)) return style;
      if (!style) return;
      isArray(style) || (style = [style]);
      return style.map(function (style) {
        return dumpStyle(style, function (geom) {
          return _this.writeGeometryInDataProj(geom);
        });
      });
    }
  },
  watch: _objectSpread$3({}, /*#__PURE__*/makeChangeOrRecreateWatchers(['renderOrder', 'renderBuffer', 'declutter', 'updateWhileAnimating', 'updateWhileInteracting', 'style'], ['style'])),
  methods: {
    /**
     * @returns {Object}
     * @protected
     */
    getServices: function getServices() {
      return mergeDescriptors(layer.methods.getServices.call(this), styleContainer.methods.getServices.call(this));
    },

    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      layer.methods.subscribeAll.call(this);
      subscribeToLayerEvents.call(this);
    },

    /**
     * @return {StyleTarget|undefined}
     */
    getStyleTarget: function getStyleTarget() {
      return this.$layer;
    },

    /**
     * @param {number[]} pixel
     * @return {Promise<Array<module:ol/Feature~Feature>>}
     */
    getFeatures: function getFeatures(pixel) {
      var _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _this2.resolveLayer();

              case 2:
                return _context.abrupt("return", _context.sent.getFeatures(pixel));

              case 3:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @param {Object|Function|Array|undefined} value
     * @param {Object|Function|Array|undefined} prev
     * @protected
     */
    styleChanged: function styleChanged(value, prev) {
      if (isEqual(value, prev)) return;

      if (isPlainObject(value) || isArray(value)) {
        value = clonePlainObject(value);
      }

      this.$emit('update:style', value);
    }
  }
};

function subscribeToLayerEvents() {
  return _subscribeToLayerEvents.apply(this, arguments);
}

function _subscribeToLayerEvents() {
  _subscribeToLayerEvents = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2);
  }));
  return _subscribeToLayerEvents.apply(this, arguments);
}

function ownKeys$2(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$2(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$2(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$2(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
var isNotEmptyString = /*#__PURE__*/and(isString, negate(isEmpty));
/**
 * Basic vector source mixin.
 */

var vectorSource = {
  mixins: [featureHelper, featuresContainer, source],
  stubVNode: {
    empty: false,
    attrs: function attrs() {
      return {
        id: this.vmId,
        class: this.vmClass
      };
    }
  },
  props: {
    // ol/source/Vector

    /**
     * @type {Object[]}
     */
    features: {
      type: Array,
      default: stubArray,
      validator: function validator(value) {
        return every(value, isGeoJSONFeature);
      }
    },

    /**
     * Source format factory
     * @type {function} formatFactory
     */
    formatFactory: {
      type: Function,
      default: getGeoJsonFmt
    },

    /**
     * Feature loader.
     * Feature loader should load features from some remote service, decode them and pas to `features` prop to render.
     * @type {module:ol/featureloader~FeatureLoader|undefined} loader
     */
    loader: Function,

    /**
     * @deprecated Use `loader` prop instead
     * @todo remove in v0.13.x
     */
    loaderFactory: Function,

    /**
     * Loading strategy factory.
     * Extent here in map view projection.
     * @type {function} strategyFactory
     */
    loadingStrategy: {
      type: Function // default: loadAll,

    },

    /**
     * @deprecated Use `loadingStrategy` prop instead
     * @todo remove in v0.13.x
     */
    strategyFactory: Function,

    /**
     * String or url factory
     * @type {string|function} url
     */
    url: {
      type: [String, Function],
      validator: function validator(value) {
        return isFunction(value) || isNotEmptyString(value);
      }
    },

    /**
     * @type {boolean}
     */
    overlaps: {
      type: Boolean,
      default: true
    },

    /**
     * @type {boolean}
     */
    useSpatialIndex: {
      type: Boolean,
      default: true
    }
  },
  data: function data() {
    return {
      /**
       * @returns {module:ol/format/Feature~FeatureFormat|undefined}
       */
      format: undefined,
      currentUrl: undefined,
      currentLoader: undefined
    };
  },
  computed: {
    featuresDataProj: function featuresDataProj() {
      return map$1(this.features, function (feature) {
        return initializeFeature(clonePlainObject(feature));
      });
    },
    featuresViewProj: function featuresViewProj() {
      var _this = this;

      return map$1(this.features, function (feature) {
        return initializeFeature(_this.writeFeatureInViewProj(_this.readFeatureInDataProj(feature)));
      });
    },
    currentFeaturesDataProj: function currentFeaturesDataProj() {
      var _this2 = this;

      if (!this.rev) return [];
      return map$1(this.getFeatures(), function (feature) {
        return _this2.writeFeatureInDataProj(feature);
      });
    },
    currentFeaturesViewProj: function currentFeaturesViewProj() {
      var _this3 = this;

      if (!this.rev) return [];
      return map$1(this.getFeatures(), function (feature) {
        return _this3.writeFeatureInViewProj(feature);
      });
    },
    extentDataProj: function extentDataProj() {
      return this.rev ? this.getExtent() : undefined;
    },
    extentViewProj: function extentViewProj() {
      return this.rev ? this.getExtent(true) : undefined;
    },
    inputUrl: function inputUrl() {
      return this.createUrlFunc(this.url);
    },
    inputLoader: function inputLoader() {
      var loader = this.loader;

      if (!loader && this.loaderFactory) {
        loader = this.loaderFactory();
      }

      return this.createLoaderFunc(loader);
    },
    inputLoadingStrategy: function inputLoadingStrategy() {
      var loadingStrategy = this.loadingStrategy;

      if (!loadingStrategy && this.strategyFactory) {
        loadingStrategy = this.strategyFactory();
      }

      return loadingStrategy || all;
    },
    formatIdent: function formatIdent() {
      if (!this.olObjIdent) return;
      return this.makeIdent(this.olObjIdent, 'format');
    },
    inputFormatFactory: function inputFormatFactory() {
      return sealFactory(this.formatFactory.bind(this));
    }
  },
  watch: _objectSpread$2({
    rev: function rev() {
      if (!this.$source) return;

      if (this.currentUrl !== this.$source.getUrl()) {
        this.currentUrl = this.$source.getUrl();
      }
    }
  }, /*#__PURE__*/makeChangeOrRecreateWatchers(['featuresViewProj', 'currentFeaturesDataProj', 'extentDataProj', 'inputUrl', 'currentUrl', 'inputLoader', 'currentLoader', 'inputLoadingStrategy', 'inputFormatFactory', 'formatIdent', 'format', 'overlaps', 'useSpatialIndex'], ['featuresViewProj', 'currentFeaturesDataProj', 'extentDataProj'])),
  created: function created() {
    if (process.env.NODE_ENV !== 'production') {
      if (this.loaderFactory) {
        this.$logger.warn("'loaderFactory' prop is deprecated. Use 'loader' prop instead.");
      }

      if (this.strategyFactory) {
        this.$logger.warn("'strategyFactory' prop is deprecated. Use 'loadingStrategy' prop instead.");
      }
    }

    if (isFunction(this.inputFormatFactory)) {
      this.format = this.instanceFactoryCall(this.formatIdent, this.inputFormatFactory.bind(this));
    }

    this.currentUrl = this.inputUrl;
    this.currentLoader = this.inputLoader;
  },
  updated: function updated() {
    if (process.env.NODE_ENV !== 'production') {
      if (this.loaderFactory) {
        this.$logger.warn("'loaderFactory' prop is deprecated. Use 'loader' prop instead.");
      }

      if (this.strategyFactory) {
        this.$logger.warn("'strategyFactory' prop is deprecated. Use 'loadingStrategy' prop instead.");
      }
    }
  },
  methods: {
    /**
     * @return {module:ol/source/Vector~VectorSource}
     * @protected
     */
    createSource: function createSource() {
      return new Vector({
        // ol/source/Source
        attributions: this.currentAttributions,
        projection: this.resolvedDataProjection,
        wrapX: this.wrapX,
        // ol/source/Vector
        format: this.format,
        loader: this.currentLoader,
        strategy: this.inputLoadingStrategy,
        url: this.currentUrl,
        overlaps: this.overlaps,
        useSpatialIndex: this.useSpatialIndex,
        features: this.$featuresCollection
      });
    },

    /**
     * @return {Object}
     * @protected
     */
    getServices: function getServices() {
      var vm = this;
      return mergeDescriptors(source.methods.getServices.call(this), {
        get featuresContainer() {
          return vm;
        }

      });
    },

    /**
     * @protected
     */
    subscribeAll: function subscribeAll() {
      source.methods.subscribeAll.call(this);
      featuresContainer.methods.subscribeAll.call(this);
      subscribeToSourceEvents.call(this);
    },

    /**
     * @param {FeatureLike[]|module:ol/Collection~Collection<FeatureLike>} features
     * @param {boolean} [viewProj=false]
     */
    addFeatures: function addFeatures(features) {
      var _this4 = this;

      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

      if (!this.$source) {
        return featuresContainer.methods.addFeatures.call(this, features, viewProj);
      }

      var newFeatures = [];
      forEach(features || [], function (feature) {
        feature = _this4.initializeFeature(feature, viewProj);
        instanceOf(feature, Feature);

        var foundFeature = _this4.$source.getFeatureById(getFeatureId(feature));

        if (foundFeature) {
          _this4.updateFeature(foundFeature, feature);
        } else {
          newFeatures.push(feature);
        }
      });
      if (!newFeatures.length) return;
      this.$source.addFeatures(newFeatures);
    },

    /**
     * @param {FeatureLike} feature
     * @param {boolean} [viewProj=false]
     */
    addFeature: function addFeature(feature) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

      if (!this.$source) {
        return featuresContainer.methods.addFeature.call(this, feature, viewProj);
      }

      feature = this.initializeFeature(feature, viewProj);
      var foundFeature = this.$source.getFeatureById(getFeatureId(feature));

      if (foundFeature == null) {
        this.$source.addFeature(feature);
      } else {
        this.updateFeature(foundFeature, feature);
      }
    },

    /**
     * @param {FeatureLike[]|module:ol/Collection~Collection<FeatureLike>} features
     */
    removeFeatures: function removeFeatures(features) {
      var _this5 = this;

      if (!this.$source) {
        return featuresContainer.methods.removeFeatures.call(this, features);
      }

      var changed = false;
      forEach(features, function (feature) {
        feature = _this5.getFeatureById(getFeatureId(feature));
        if (!feature) return;
        var featureKey = getUid$1(feature);

        if (featureKey in _this5.$source.nullGeometryFeatures_) {
          delete _this5.$source.nullGeometryFeatures_[featureKey];
        } else {
          if (_this5.$source.featuresRtree_) {
            _this5.$source.featuresRtree_.remove(feature);
          }
        }

        _this5.$source.removeFeatureInternal(feature);

        changed = true;
      });
      if (!changed) return;
      this.$source.changed();
    },

    /**
     * @param {FeatureLike} feature
     */
    removeFeature: function removeFeature(feature) {
      var _feature;

      if (!this.$source) {
        return featuresContainer.methods.removeFeature.call(this, feature);
      }

      feature = this.$source.getFeatureById(getFeatureId(((_feature = feature) === null || _feature === void 0 ? void 0 : _feature.$feature) || feature));
      if (!feature) return;
      this.$source.removeFeature(feature);
    },

    /**
     * @param {string} featureId
     * @return {module:ol/Feature~Feature|undefined}
     */
    getFeatureById: function getFeatureById(featureId) {
      if (this.$source) {
        return this.$source.getFeatureById(featureId);
      }

      return featuresContainer.methods.getFeatureById.call(this, featureId);
    },

    /**
     * @return {Array<module:ol/Feature~Feature>}
     */
    getFeatures: function getFeatures() {
      var _this$$source;

      return coalesce((_this$$source = this.$source) === null || _this$$source === void 0 ? void 0 : _this$$source.getFeatures(), featuresContainer.methods.getFeatures.call(this));
    },

    /**
     * @return {module:ol/Collection~Collection|undefined}
     */
    getFeaturesCollection: function getFeaturesCollection() {
      var _this$$source2;

      return coalesce((_this$$source2 = this.$source) === null || _this$$source2 === void 0 ? void 0 : _this$$source2.getFeaturesCollection(), this._featuresCollection);
    },

    /**
     * @return {boolean}
     */
    isEmpty: function isEmpty() {
      var _this$$source3;

      return coalesce((_this$$source3 = this.$source) === null || _this$$source3 === void 0 ? void 0 : _this$$source3.isEmpty(), this.$featuresCollection.getLength() === 0);
    },

    /**
     * @param {boolean} [fast]
     * @return {Promise<void>}
     */
    clear: function clear(fast) {
      var _this6 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        var source;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (_this6.$source) {
                  _context.next = 2;
                  break;
                }

                return _context.abrupt("return", _this6.clearFeatures());

              case 2:
                _context.next = 4;
                return _this6.resolveSource();

              case 4:
                source = _context.sent;
                return _context.abrupt("return", new Promise(function (resolve) {
                  source.once('change', function () {
                    return resolve();
                  });
                  source.clear();
                }));

              case 6:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @param {function} callback
     * @returns {*}
     */
    forEachFeature: function forEachFeature(callback) {
      if (this.$source) {
        return this.$source.forEachFeature(callback);
      }

      return featuresContainer.methods.forEachFeature.call(this, callback);
    },

    /**
     * @param {number[]} extent
     * @param {function} callback
     * @param {boolean} [viewProj=false]
     * @returns {*}
     */
    forEachFeatureInExtent: function forEachFeatureInExtent(extent, callback) {
      var viewProj = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      extent = viewProj ? roundExtent(extent) : this.extentToViewProj(extent);

      if (this.$source) {
        return this.$source.forEachFeatureInExtent(extent, callback);
      }

      return featuresContainer.methods.forEachFeatureInExtent.call(this, extent, callback, true);
    },

    /**
     * @param {number[]} extent
     * @param {function} callback
     * @param {boolean} [viewProj=false]
     * @returns {*}
     */
    forEachFeatureIntersectingExtent: function forEachFeatureIntersectingExtent(extent, callback) {
      var viewProj = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      extent = viewProj ? roundExtent(extent) : this.extentToViewProj(extent);

      if (this.$source) {
        return this.$source.forEachFeatureIntersectingExtent(extent, callback);
      }

      return featuresContainer.methods.forEachFeatureIntersectingExtent.call(this, extent, callback, true);
    },

    /**
     * @param {number[]} coordinate
     * @param {boolean} [viewProj=false]
     * @returns {Array<module:ol/Feature~Feature>}
     */
    getFeaturesAtCoordinate: function getFeaturesAtCoordinate(coordinate) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      coordinate = viewProj ? roundPointCoords(coordinate) : this.pointToViewProj(coordinate);

      if (this.$source) {
        return this.$source.getFeaturesAtCoordinate(coordinate);
      }

      return featuresContainer.methods.getFeaturesAtCoordinate.call(this, coordinate, true);
    },

    /**
     * @param {number[]} extent
     * @param {boolean} [viewProj=false]
     * @returns {Array<module:ol/Feature~Feature>}
     */
    getFeaturesInExtent: function getFeaturesInExtent(extent) {
      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      extent = viewProj ? roundExtent(extent) : this.extentToViewProj(extent);

      if (this.$source) {
        return this.$source.getFeaturesInExtent(extent);
      }

      return featuresContainer.methods.getFeaturesInExtent.call(this, extent, true);
    },

    /**
     * @param {number[]} coordinate
     * @param {function} [filter]
     * @param {boolean} [viewProj=false]
     * @returns {module:ol/Feature~Feature}
     */
    getClosestFeatureToCoordinate: function getClosestFeatureToCoordinate(coordinate) {
      var filter = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : identity;
      var viewProj = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      coordinate = viewProj ? roundPointCoords(coordinate) : this.pointToViewProj(coordinate);

      if (this.$source) {
        return this.$source.getClosestFeatureToCoordinate(coordinate, filter);
      }

      return featuresContainer.methods.getClosestFeatureToCoordinate.call(this, coordinate, filter, true);
    },

    /**
     * @param {boolean} [viewProj=false]
     * @returns {number[]}
     */
    getExtent: function getExtent() {
      var viewProj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
      var extent;

      if (this.$source) {
        extent = this.$source.getExtent();
      } else {
        extent = featuresContainer.methods.getFeaturesExtent.call(this, true);
      }

      return viewProj ? extent : this.extentToDataProj(extent);
    },

    /**
     * @returns {module:ol/format/Feature~FeatureFormat|undefined}
     */
    getFormat: function getFormat() {
      var _this$$source4;

      return coalesce((_this$$source4 = this.$source) === null || _this$$source4 === void 0 ? void 0 : _this$$source4.getFormat(), this.format);
    },

    /**
     * @returns {Promise<string|function|undefined>}
     */
    getUrl: function getUrl() {
      var _this$$source5;

      return coalesce((_this$$source5 = this.$source) === null || _this$$source5 === void 0 ? void 0 : _this$$source5.getUrl(), this.currentUrl);
    },

    /**
     * @param {string|function} url
     */
    setUrl: function setUrl(url) {
      url = this.createUrlFunc(url);

      if (url !== this.currentUrl) {
        this.currentUrl = url;
      }

      if (this.$source && url !== this.$source.getUrl()) {
        this.$source.setUrl(url);
      }
    },

    /**
     * @param {function} loader
     */
    setLoader: function setLoader(loader) {
      loader = this.createLoaderFunc(loader);

      if (loader !== this.currentLoader) {
        var _this$$source6;

        this.currentLoader = loader;
        (_this$$source6 = this.$source) === null || _this$$source6 === void 0 ? void 0 : _this$$source6.setLoader(loader);
      }
    },

    /**
     * @param {number[]} extent
     * @param {boolean} [viewProj=false]
     */
    removeLoadedExtent: function removeLoadedExtent(extent) {
      var _this$$source7;

      var viewProj = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      extent = viewProj ? roundExtent(extent) : this.extentToViewProj(extent);
      (_this$$source7 = this.$source) === null || _this$$source7 === void 0 ? void 0 : _this$$source7.removeLoadedExtent(extent);
    },

    /**
     * @param {string|function|undefined} url
     * @return {function|undefined}
     * @protected
     */
    createUrlFunc: function createUrlFunc(url) {
      if (!url) return;

      if (!isFunction(url)) {
        url = constant(url);
      }

      return this.wrapUrlFunc(url);
    },

    /**
     * @param {function} urlFunc
     * @returns {function}
     * @protected
     */
    wrapUrlFunc: function wrapUrlFunc(urlFunc) {
      var _this7 = this;

      var fn = urlFunc;

      if (!fn.wrapped) {
        fn = function fn(extent, resolution, projection) {
          extent = transformExtent(extent, projection, _this7.resolvedDataProjection);
          projection = _this7.resolvedDataProjection;
          return urlFunc(extent, resolution, projection);
        };

        fn.wrapped = true;
      }

      return fn;
    },

    /**
     * @param {function|undefined} loader
     * @return {function|undefined}
     * @protected
     */
    createLoaderFunc: function createLoaderFunc(loader) {
      if (!isFunction(loader)) return;
      return this.wrapLoaderFunc(loader);
    },

    /**
     * @param {function} loaderFunc
     * @returns {Function}
     * @protected
     */
    wrapLoaderFunc: function wrapLoaderFunc(loaderFunc) {
      var _this8 = this;

      var fn = loaderFunc;

      if (!fn.wrapped) {
        fn = /*#__PURE__*/function () {
          var _ref = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2(extent, resolution, projection) {
            var features;
            return _regeneratorRuntime.wrap(function _callee2$(_context2) {
              while (1) {
                switch (_context2.prev = _context2.next) {
                  case 0:
                    _context2.next = 2;
                    return loaderFunc(transformExtent(extent, projection, _this8.resolvedDataProjection), resolution, _this8.resolvedDataProjection);

                  case 2:
                    features = _context2.sent;

                    if (!(isArray(features) && features[0] instanceof Feature)) {
                      features = _this8.readSource(features);
                    }

                    if (isArray(features)) {
                      _this8.addFeatures(features);
                    }

                  case 5:
                  case "end":
                    return _context2.stop();
                }
              }
            }, _callee2);
          }));

          return function fn(_x, _x2, _x3) {
            return _ref.apply(this, arguments);
          };
        }();

        fn.wrapped = true;
      }

      return fn;
    },

    /**
     * @param {*} data
     * @returns {Array<module:ol/Feature~Feature>}
     */
    readSource: function readSource(data) {
      return this.getFormat().readFeatures(data, {
        featureProjection: this.resolvedViewProjection,
        dataProjection: this.resolvedDataProjection
      });
    },

    /**
     * @param {GeoJSONFeature[]} value
     * @protected
     */
    featuresViewProjChanged: function featuresViewProjChanged(value) {
      // add new features
      this.addFeatures(value, true); // remove non-matched features

      this.removeFeatures(difference(this.getFeatures(), value, function (a, b) {
        return getFeatureId(a) === getFeatureId(b);
      }));
    },

    /**
     * @param {GeoJSONFeature[]} value
     * @protected
     */
    currentFeaturesDataProjChanged: function currentFeaturesDataProjChanged(value) {
      if (isEqual(value, this.featuresDataProj)) return;
      this.$emit('update:features', clonePlainObject(value));
    },

    /**
     * @param {number[]|undefined} value
     * @param {number[]|undefined} prev
     * @protected
     */
    extentDataProjChanged: function extentDataProjChanged(value, prev) {
      if (isEqual(value, prev)) return;
      this.$emit('update:extent', value === null || value === void 0 ? void 0 : value.slice());
    },

    /**
     * @param {string|function|undefined} value
     * @protected
     */
    inputUrlChanged: function inputUrlChanged(value) {
      this.setUrl(value);
    },

    /**
     * @param {string|function|undefined} value
     * @protected
     */
    currentUrlChanged: function currentUrlChanged(value) {
      if (value === this.inputUrl) return;
      this.$emit('update:url', value);
    },

    /**
     * @param {function|undefined} value
     * @protected
     */
    inputLoaderChanged: function inputLoaderChanged(value) {
      this.setLoader(value);
    },

    /**
     * @param {function|undefined} value
     * @protected
     */
    currentLoaderChanged: function currentLoaderChanged(value) {
      if (value === this.inputLoader) return;
      this.$emit('update:loader', value);
    },

    /**
     * @param {string|undefined} value
     * @param {string|undefined} prevValue
     * @protected
     */
    formatIdentChanged: function formatIdentChanged(value, prevValue) {
      if (value && prevValue) {
        this.moveInstance(value, prevValue);
      } else if (value && !prevValue && this.format) {
        this.setInstance(value, this.format);
      } else if (!value && prevValue) {
        this.unsetInstance(prevValue);
      }
    },

    /**
     * @param {function} value
     * @protected
     */
    inputFormatFactoryChanged: function inputFormatFactoryChanged(value) {
      while (this.hasInstance(this.formatIdent)) {
        this.unsetInstance(this.formatIdent);
      }

      if (isFunction(value)) {
        this.format = this.instanceFactoryCall(this.formatIdent, value.bind(this));
      } else {
        this.format = undefined;
      }
    },
    // skip not used watchers
    attributionsCollapsibleChanged: noop,
    stateChanged: noop
  }
};

function subscribeToSourceEvents() {}

function transformExtent(extent, sourceProj, destProj) {
  extent = extent.slice();

  if (isFinite(extent[0]) && isFinite(extent[1])) {
    var _transform = transform([extent[0], extent[1]], sourceProj, destProj);

    var _transform2 = _slicedToArray(_transform, 2);

    extent[0] = _transform2[0];
    extent[1] = _transform2[1];
  }

  if (isFinite(extent[2]) && isFinite(extent[3])) {
    var _transform3 = transform([extent[2], extent[3]], sourceProj, destProj);

    var _transform4 = _slicedToArray(_transform3, 2);

    extent[2] = _transform4[0];
    extent[3] = _transform4[1];
  }

  return extent;
}

function ownKeys$1(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread$1(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys$1(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys$1(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

var cleanWmsSourceParams = function cleanWmsSourceParams(params) {
  return cleanSourceParams(params, ['LAYERS', 'VERSION', 'STYLES', 'FORMAT', 'TRANSPARENT', 'BGCOLOR', 'TIME']);
};
/**
 * Basic WMS params and methods mixin.
 */


var wmsSource = {
  props: {
    /**
     * @type {boolean}
     */
    hidpi: {
      type: Boolean,
      default: true
    },

    /**
     * @type {string|undefined}
     */
    serverType: {
      type: String,
      validator: function validator(value) {
        return Object.values(WMSServerType).includes(value);
      }
    },

    /**
     * @type {string|string[]}
     */
    layers: {
      type: String,
      required: true
    },

    /**
     * WMS Request styles
     * @type {string|string[]|undefined}
     */
    styles: String,

    /**
     * @type {string}
     */
    version: {
      type: String,
      default: '1.3.0'
    },

    /**
     * @type {boolean}
     */
    transparent: {
      type: Boolean,
      default: true
    },

    /**
     * @type {string}
     */
    format: {
      type: String,
      default: 'image/png'
    },

    /**
     * @type {string|undefined}
     */
    bgColor: String,

    /**
     * @type {string|undefined}
     */
    time: String,

    /**
     * Additional WMS request parameters
     * @type {Object|undefined}
     */
    params: Object
  },
  data: function data() {
    return {
      currentParams: undefined
    };
  },
  computed: _objectSpread$1({
    /**
     * @returns {string}
     */
    inputLayers: function inputLayers() {
      return isArray(this.layers) ? this.layers.join(',') : this.layers;
    },

    /**
     * @returns {string|undefined}
     */
    inputStyles: function inputStyles() {
      return isArray(this.styles) ? this.styles.join(',') : this.styles;
    },

    /**
     * @returns {Object|null}
     */
    customParams: function customParams() {
      return this.params ? cleanWmsSourceParams(this.params) : undefined;
    },

    /**
     * @returns {Object}
     */
    inputParams: function inputParams() {
      return _objectSpread$1(_objectSpread$1({}, this.customParams || {}), {}, {
        LAYERS: this.inputLayers,
        STYLES: this.inputStyles,
        VERSION: this.version,
        FORMAT: this.format,
        TRANSPARENT: this.transparent,
        BGCOLOR: this.bgColor,
        TIME: this.time
      });
    }
  }, /*#__PURE__*/reduce(['inputLayers', 'inputStyles', 'version', 'format', 'transparent', 'bgColor', 'time'], function (props, inProp) {
    var prop = inProp.slice(0, 5) === 'input' ? lowerFirst(inProp.slice(5)) : inProp;
    var curProp = 'current' + upperFirst(prop);

    props[curProp] = function () {
      return coalesce((this.currentParams || {})[prop.toUpperCase()], (this.inputParams || {})[prop.toUpperCase()]);
    };

    return props;
  }, {})),
  watch: _objectSpread$1(_objectSpread$1({
    rev: function rev() {
      if (!this.$source) return;

      if (!isEqual(this.currentParams, this.$source.getParams())) {
        this.currentParams = this.$source.getParams();
      }
    }
  }, /*#__PURE__*/makeWatchers(['currentLayers', 'currentStyles', 'currentVersion', 'currentFormat', 'currentTransparent', 'currentBgColor', 'currentTime'], function (curProp) {
    return function (value) {
      var prop = lowerFirst(curProp.slice(7));
      var inProp = hasProp(this, 'input' + upperFirst(prop)) ? 'input' + upperFirst(prop) : prop;
      if (isEqual(value, this[inProp])) return;
      this.$emit('update:' + prop, isObjectLike(value) ? clonePlainObject(value) : value);
    };
  })), /*#__PURE__*/makeChangeOrRecreateWatchers(['hidpi', 'serverType', 'inputParams', 'currentParams'], ['inputParams', 'currentParams'])),
  created: function created() {
    this.currentParams = this.inputParams && clonePlainObject(this.inputParams);
  },
  methods: {
    /**
     * @param {number[]} coordinate
     * @param {number} [resolution]
     * @param {string} [projection]
     * @param {Object|undefined} [params] GetFeatureInfo params. `info_format` at least should be provided.
     *                          If `query_layers` is not provided then the layers specified in the `layers` prop will be used.
     *                          `version` should not be specified here (value from `version` prop will be used).
     * @return {Promise<string|undefined>}
     */
    getFeatureInfoUrl: function getFeatureInfoUrl(coordinate, resolution, projection) {
      var _arguments = arguments,
          _this = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
        var _this$$viewVm;

        var params;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                params = _arguments.length > 3 && _arguments[3] !== undefined ? _arguments[3] : {};
                resolution || (resolution = (_this$$viewVm = _this.$viewVm) === null || _this$$viewVm === void 0 ? void 0 : _this$$viewVm.getResolution());
                projection || (projection = _this.resolvedDataProjection);
                _context.next = 5;
                return _this.resolveSource();

              case 5:
                return _context.abrupt("return", _context.sent.getFeatureInfoUrl(coordinate, resolution, projection, params));

              case 6:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }))();
    },

    /**
     * @param {number} resolution
     * @param {Object|undefined} [params]
     * @return {Promise<string|undefined>}
     */
    getLegendUrl: function getLegendUrl(resolution) {
      var _arguments2 = arguments,
          _this2 = this;

      return _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
        var params;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                params = _arguments2.length > 1 && _arguments2[1] !== undefined ? _arguments2[1] : {};
                _context2.next = 3;
                return _this2.resolveSource();

              case 3:
                return _context2.abrupt("return", _context2.sent.getLegendUrl(resolution, params));

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }))();
    },

    /**
     * @returns {Object}
     */
    getParams: function getParams() {
      var _this$$source;

      return coalesce((_this$$source = this.$source) === null || _this$$source === void 0 ? void 0 : _this$$source.getParams(), this.currentParams);
    },

    /**
     * @param {Object} params
     */
    updateParams: function updateParams(params) {
      params = reduce(_objectSpread$1(_objectSpread$1({}, this.currentParams), params), function (params, value, name) {
        return _objectSpread$1(_objectSpread$1({}, params), {}, _defineProperty({}, name.toUpperCase(), value));
      }, {});

      if (!isEqual(params, this.currentParams)) {
        this.currentParams = params;
      }

      if (this.$source && !isEqual(params, this.$source.getParams())) {
        this.$source.updateParams(params);
      }
    },

    /**
     * @param {string} param
     * @param {*} value
     */
    updateParam: function updateParam(param, value) {
      this.updateParams(_defineProperty({}, param.toUpperCase(), value));
    },

    /**
     * @param {Object|undefined} value
     * @protected
     */
    inputParamsChanged: function inputParamsChanged(value) {
      this.updateParams(value);
    },

    /**
     * @param {Object|undefined} value
     * @protected
     */
    currentParamsChanged: function currentParamsChanged(value) {
      value = value ? cleanWmsSourceParams(value) : undefined;
      if (isEqual(value, this.customParams)) return;
      this.$emit('update:params', value && clonePlainObject(value));
    }
  }
};

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

var validateMinZoom = function validateMinZoom(value) {
  return value >= 0;
};

var validateTileSize = /*#__PURE__*/or(isNumber, and(isArray, function (value) {
  return value.length === 2 && value.every(isNumber);
}));
/**
 * Base XYZ source mixin.
 */

var xyzSource = {
  mixins: [tileImageSource],
  props: {
    /* eslint-disable vue/require-prop-types */
    // ol/source/Source
    projection: _objectSpread(_objectSpread({}, source.props.projection), {}, {
      default: EPSG_3857
    }),

    /* eslint-enable vue/require-prop-types */
    // ol/source/XYZ
    maxZoom: {
      type: Number,
      default: 42
    },
    minZoom: {
      type: Number,
      default: 0,
      validator: validateMinZoom
    },
    maxResolution: Number,
    tileSize: {
      type: [Number, Array],
      default: function _default() {
        return [256, 256];
      },
      validator: validateTileSize
    }
  },
  computed: {
    inputTileSize: function inputTileSize() {
      return isArray(this.tileSize) ? this.tileSize.slice() : [this.tileSize, this.tileSize];
    },
    derivedTileGridFactory: function derivedTileGridFactory() {
      if (isFunction(this.tileGridFactory)) {
        return this.tileGridFactory;
      }

      var extent = extentFromProjection(this.resolvedDataProjection);
      var maxZoom = this.maxZoom;
      var minZoom = this.minZoom;
      var maxResolution = this.maxResolution;
      var tileSize = this.inputTileSize;
      return function () {
        return createXYZ({
          extent: extent,
          maxZoom: maxZoom,
          minZoom: minZoom,
          maxResolution: maxResolution,
          tileSize: tileSize
        });
      };
    },
    inputTileUrlFunction: function inputTileUrlFunction() {
      var urlFunc = coalesce(this.tileUrlFunction, this.tileUrlFunc);
      if (isFunction(urlFunc)) return urlFunc;
      if (this.currentUrls.length === 0) return;
      return createTileUrlFunctionFromTemplates(this.currentUrls, this.tileGrid);
    }
  },
  methods: {
    /**
     * @return {module:ol/source/XYZ~XYZSource}
     * @protected
     */
    createSource: function createSource() {
      return new XYZ({
        // ol/source/Source
        attributions: this.currentAttributions,
        attributionsCollapsible: this.attributionsCollapsible,
        projection: this.resolvedDataProjection,
        wrapX: this.wrapX,
        // ol/source/Tile
        cacheSize: this.cacheSize,
        opaque: this.opaque,
        tilePixelRatio: this.tilePixelRatio,
        transition: this.transition,
        zDirection: this.zDirection,
        tileGrid: this.tileGrid,
        // ol/source/UrlTile
        tileLoadFunction: this.currentTileLoadFunction,
        tileUrlFunction: this.currentTileUrlFunction,
        // ol/source/TileImage
        crossOrigin: this.crossOrigin,
        reprojectionErrorThreshold: this.reprojectionErrorThreshold,
        tileClass: this.tileClass,
        imageSmoothing: this.imageSmoothing
      });
    },
    tileKeyChanged: noop,
    // input tileKey is not allowed in XYZ constructor
    stateChanged: noop // input state is not allowed in XYZ constructor

  }
};

export { CanceledError, EVENT_BUS_PROP, FRAME_TIME, IDENTITY_MAP_PROP, INSTANCES_POOL, LifecycleError, OlObjectAction, OlObjectEvent, OlObjectState, SERVICES_PROP, VM_PROP, arcgisSource, baseLayer, controlsContainer, eventBus, feature, featureHelper, featuresContainer, fillStyleContainer, geometry, geometryContainer, identMap, imageLayer, imageSource, imageStyle, imageStyleContainer, interaction, interactionsContainer, isCreateError, isDestroyError, isMountError, isUnmountError, layer, layersContainer, makeChangeOrRecreateWatchers, olCmp, overlaysContainer, projTransforms, regShapeStyle, rxSubs, services, simpleGeometry, source, sourceContainer, strokeStyleContainer, stubVNode, style, styleContainer, textStyleContainer, tileImageSource, tileLayer, tileSource, urlTileSource, vectorLayer, vectorSource, waitForMap, wmsSource, xyzSource };
//# sourceMappingURL=mixins.js.map
